/* Command-line parser.
 *
 * PegSoft command-line parser library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file cmdline/cmdline.h
 * \brief Header file for the command line parser.
 */

#ifndef CMDLINE_CMDLINE_H
#define CMDLINE_CMDLINE_H

/*************************************************************************/

#include <map>
#include <string>

/*************************************************************************/

/// Namespace for all command-line parser declarations.

namespace cmdline
{

/*************************************************************************/

/// A parser for command-line parameters and environment variables.

/** The CmdLineParser class is used to parse and evaluate command-line
 *  parameters and environment variables as given at program launch (the argc,
 *  argv and envp parameters of main(). Parameters are stored in a map so they
 *  can be easily retrieved later.
 */

class CmdLineParser
{
	public:
		/// Multimap type used to store parameters.
		typedef std::multimap<std::string, std::string> ParamsList;
		/// Map type used to store environment variables.
		typedef std::map<std::string, std::string> VarsList;

		/// Default constructor.
		CmdLineParser() : m_error(), m_progname(), m_params(), m_vars(),
				m_argc(0), m_argv(0), m_envp(0) { }
		/// Constructor setting argc, argv and envp.
		/** This constructor will set the count and array of parameters
		 *  used by the parse() function (usually the same as
		 *  those of the main() function) Optionally, an array of
		 *  environment variables (also available in main()) can
		 *  be supplied, and they will be parsed by the parse() function
		 *  too in that case.
		 *
		 *  \param[in] argc The number of parameters in \a argv.
		 *  \param[in] argv The array of parameters.
		 *  \param[in] envp The array of environment variables (defaults to
		 *                  NULL).
		 *  \sa parse()
		 */
		CmdLineParser(int argc, char const *const *argv,
				char const *const *envp = NULL): m_error(), m_progname(),
				m_params(), m_vars(), m_argc(argc), m_argv(argv),
				m_envp(envp) {}
		/// Copy constructor.
		CmdLineParser(CmdLineParser const &clp) : m_error(clp.m_error),
				m_progname(clp.m_progname), m_params(clp.m_params),
				m_vars(clp.m_vars), m_argc(clp.m_argc), m_argv(clp.m_argv),
				m_envp(clp.m_envp) { }
		/// Destructor.
		virtual ~CmdLineParser() { }

		/// Parses the command-line parameters and environment variables.
		/** Parses command-line parameters and environment variables (if
		  * available), and calls the eval() function.
		  *
		  * \return \b true if everything has been parsed successfully
		  * and the eval() function succeeded, \b else it returns false
		  * and will generate an error message that can be retrieved
		  * through error().
		  * \sa error() eval()
		  */
		virtual bool parse();
		
		/// Returns the number of parameters in the argv() array.
		int argc() const { return m_argc; }
		/// Returns the raw parameters list.
		char const *const *argv() const { return m_argv; }
		/// Returns the array of environment variables.
		char const *const *envp() const { return m_envp; }
		
		/// Returns a description of the error that occured.
		/** If parse() fails, this function will return a
		  * description of the error that occured.
		  * \return The description of the error that occured.
		  * \sa parse()
		  */
		std::string const &error() const { return m_error; }
		/// Same as error() but returns a char const *
		char const *c_error() const { return m_error.c_str(); }
		
		/// Returns the program path.
		/** Returns the program path. This is roughly equivalent
		  * to string(argv()[0]).
		  * \return The program path.
		  * \sa argv()
		  */
		std::string const program() const { return std::string(m_argv[0]); }
		/// Same as program() but returns a char const *
		char const *c_program() const { return m_argv[0]; }
		/// Returns the program name.
		/** Returns the program file name, with any directory
		  * information removed.
		  * \return The program file name.
		  */ 
		std::string const &progname() const { return m_progname; }
		/// Same as progname() but returns a char const *
		char const *c_progname() const { return m_progname.c_str(); }
		
		/// Returns the multimap of parameters.
		/** Returns the multimap of parameters. The keys are the
		  * parameter names, while the values are the parameter
		  * values.
		  * \return The map of parameters.
		  */
		ParamsList const &params() const { return m_params; }
		
		/// Returns the map of environment variables.
		/** Returns the map of environment variables. The keys are the
		  * variable names, while the values are the variable values.
		  * \return The map of environment variables. */
		VarsList const &vars() const { return m_vars; }
	protected:
		/// Processes the command line parameters and environment variables.
		/** Processes the command line parameters and environment variables.
		  * This function is automatically called from parse(). It should
		  * be overriden in derived classes to do whatever needed.
		  * \return \b true if the processing succeeded, else it
		  * must return \b false and set m_error as appropriate.
		  * \sa m_error parse()
		  */
		virtual bool eval() { return true; };
		
		/// Last error.
		std::string m_error;
		/// Program name.
		std::string m_progname;
		/// Parameters list.
		ParamsList m_params;
		/// Variables list.
		VarsList m_vars;
	private:
		int m_argc;
		char const *const *m_argv;
		char const *const *m_envp;
};

/*************************************************************************/

} /* namespace cmdline */

/*************************************************************************/

#endif /* CMDLINE_CMDLINE_H */
