/* Data collection.
 *
 * PegSoft databases library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include "databases/collection.h"
#include "databases/database.h"
#include "databases/field.h"

using namespace std;

/*************************************************************************/

namespace databases
{

/*************************************************************************/

class RecHash : public containers::Hash<Record *, Record::identifier_type>
{
	public:
		virtual int compare(Record::identifier_type const &key,
				Record *const &value) const 
				{ return ((key < value->id()) ? -1 : ((key > value->id()) ? 1 : 0)); }
		
		virtual uint32_t hash(Record::identifier_type const &key,
				uint32_t max) const { return (key % max); }
		
		virtual bool match(Record::identifier_type const &key,
				Record *const &value) const { return (key == value->id()); }
};

/*************************************************************************/

static RecHash rh;

/*************************************************************************/

bool ArrayCollection::add(Record *record, int *err)
{
	if (m_named)
	{
		/* Ensure there are no other records with the same name. */
		NameRecord *nr = (NameRecord *)record;
		
		const_iterator cit, cend;
		NameRecord const *nr2;
	
		for (cit = begin(), cend = end(); cit != cend; ++cit)
		{
			nr2 = (NameRecord *)(*cit);
			
			if (*nr == nr2->name())
			{
				if (err)
					*err = ceName;
				return false;
			}
		}
	}
	
	return Collection::add(record, err);
}

/*************************************************************************/

bool ArrayCollection::add_list(Record *record)
{
	array_type::iterator it = m_array.end(), end;
	
	for (it = m_array.begin(), end = m_array.end(); it != end; ++it)
	{
		if (*it == 0)
		{
			*it = record;
			break;
		}
	}

	if (it == m_array.end())
		it = m_array.insert(it, record);
	
	if (!Collection::add_list(record))
	{
		*it = 0;
		return false;
	}
	
	return true;
}

/*************************************************************************/

void ArrayCollection::clear_list()
{
	/* Clear the base class first. */
	Collection::clear_list();
	
	m_array.clear();
}

/*************************************************************************/

size_t ArrayCollection::memory_usage() const
{
	size_t result = Collection::memory_usage();
	/* Guess how much memory std::list adds for each record... =) */
	result += sizeof(m_array);
	result += (m_array.capacity() * sizeof(Record *));
	
	return result;
}

/*************************************************************************/

void ArrayCollection::remove_list(Record *record)
{
	Collection::remove_list(record);
	
	array_type::iterator it, end;
	
	for (it = m_array.begin(), end = m_array.end(); it != end; ++it)
	{
		if (*it == record)
		{
			*it = 0;
			break;
		}
	}
}

/*************************************************************************/

Collection::Collection() : m_members(rh), m_db(0), m_status(0), m_last_id(0),
		m_owned()
{
}

/*************************************************************************/

Collection::~Collection()
{
	id_hashlist::iterator it, end;
	
	for (it = m_members.begin(), end = m_members.end(); it != end; ++it)
	{
		(*it)->db_status_remove(Record::dbsMember);
		delete *it;
	}
}

/*************************************************************************/

bool Collection::add(Record *record, int *err)
{
	if (!(record->db_status() & Record::dbsMember))
	{
		if (record->id() == 0)
		{
			if (++m_last_id >= external_id)
			{
				if (err)
					*err = ceID;
				return false;
			}
			
			record->id(m_last_id);
		} else {
			if (record->id() > m_last_id && record->id() < external_id)
				m_last_id = record->id();
		}
		
		if (!add_list(record))
		{
			if (err)
				*err = ceList;
			return false;
		}
		
		record->db_status_add(Record::dbsMember);
	}
	
	/* Add the record to the database */
	if (!(record->db_status() & Record::dbsInsert))
	{
		if (!add_db(record))
		{
			/* Remove the record from the lists. */
			remove_list(record);
			record->db_status_remove(Record::dbsMember);
			
			if (err)
				*err = ceDatabase;
			return false;
		}
		
		record->db_status_add(Record::dbsInsert);
	}
	
	status_add(csChanges);
	return true;
}

/*************************************************************************/

bool Collection::add_db(Record *record)
{	
	if (!m_db)
		return true;
	
	return m_db->insert(record);
}

/*************************************************************************/

bool Collection::add_list(Record *record)
{
	if (m_members.add_sorted(record->id(), record) == m_members.end())
		return false;
	
	return true;
}

/*************************************************************************/

bool Collection::change_i(Record *record,
		Field const *field, int32_t value)
{
	if (!m_db || !(record->db_status() & Record::dbsInsert) ||
			(record->db_status() & Record::dbsUpdating) || 
			!m_db->field_find(field->name(), Database::vtExport))
	{
		status_add(csChanges);
		return true;
	}
	
	bool res = m_db->update_i(record, field, value);
	
	if (res)
		status_add(csChanges);
	
	return res;
}

/*************************************************************************/

bool Collection::change_multi(Record *record,
		unsigned int valcount, ...)
{
	if (!m_db || !(record->db_status() & Record::dbsInsert) ||
			(record->db_status() & Record::dbsUpdating))
	{
		status_add(csChanges);
		return true;
	}
	
	va_list args;
	va_start(args, valcount);
	
	bool res = m_db->update_multi(record, valcount, args);
	va_end(args);	

	if (res)
		status_add(csChanges);
	
	return res;
}

/*************************************************************************/

bool Collection::change_owned(Field const *field,
		Record::identifier_type id, Record::identifier_type new_id)
{
	if (!m_db || !m_db->field_find(field->name(), Database::vtExport))
	{
		status_add(csChanges);
		return true;
	}
	
	bool res = m_db->update_owned(field, id, new_id);
	
	if (res)
		status_add(csChanges);
	
	return res;
}

/*************************************************************************/

bool Collection::change_ri(Record *record,
		Field const *field, Record::identifier_type value)
{
	if (!m_db || !(record->db_status() & Record::dbsInsert) ||
			(record->db_status() & Record::dbsUpdating) || 
			!m_db->field_find(field->name(), Database::vtExport))
	{
		status_add(csChanges);
		return true;
	}
	
	bool res = m_db->update_ri(record, field, value);
	
	if (res)
		status_add(csChanges);
	
	return res;
}

/*************************************************************************/

bool Collection::change_s(Record *record,
		Field const *field, char const *value)
{
	if (!m_db || !(record->db_status() & Record::dbsInsert) ||
			(record->db_status() & Record::dbsUpdating) || 
			!m_db->field_find(field->name(), Database::vtExport))
	{
		status_add(csChanges);
		return true;
	}
	
	bool res = m_db->update_s(record, field, value);
	
	if (res)
		status_add(csChanges);
	
	return res;
}

/*************************************************************************/

bool Collection::change_ui(Record *record,
		Field const *field, uint32_t value)
{
	if (!m_db || !(record->db_status() & Record::dbsInsert) ||
			(record->db_status() & Record::dbsUpdating) || 
			!m_db->field_find(field->name(), Database::vtExport))
	{
		status_add(csChanges);
		return true;
	}
	
	bool res = m_db->update_ui(record, field, value);
	
	if (res)
		status_add(csChanges);
	
	return res;
}

/*************************************************************************/

bool Collection::change_ut(Record *record,
		Field const *field, time_t value)
{
	if (!m_db || !(record->db_status() & Record::dbsInsert) ||
			(record->db_status() & Record::dbsUpdating) || 
			!m_db->field_find(field->name(), Database::vtExport))
	{
		status_add(csChanges);
		return true;
	}
	
	bool res = m_db->update_ut(record, field, value);
	
	if (res)
		status_add(csChanges);
	
	return res;
}

/*************************************************************************/

bool Collection::clear(int *err)
{
	/* Clear owned collection */
	owned_list::iterator it, end;
	
	for (it = m_owned.begin(), end = m_owned.end(); it != end; ++it)
		if (!(*it)->clear_owned(this, err))
			return false;
	
	/* Clear the database */
	if (!clear_db())
	{
		if (err)
			*err = ceDatabase;
		return false;
	}
	
	/* Clear the lists */
	clear_list();
	
	status_add(csChanges);
	return true;
}

/*************************************************************************/

bool Collection::clear_db()
{
	if (!m_db)
		return true;
	
	return m_db->truncate();
}

/*************************************************************************/

void Collection::clear_list()
{
	id_hashlist::iterator it, end;
	
	for (it = m_members.begin(), end = m_members.end(); it != end; ++it)
	{
		/* Records have been erased from database and are going to
		 * not be in the collection anymore, so erase them. */
		(*it)->db_status_remove(Record::dbsInsert | Record::dbsMember);
		delete *it;
	}

	m_members.clear();
}

/*************************************************************************/

bool Collection::clear_owned(Collection const *coll, int *err)
{
	return clear(err);
}

/*************************************************************************/

Record *Collection::find(Record::identifier_type id)
{
	id_hashlist::iterator it = m_members.find_sorted(id);
	
	if (it == m_members.end())
		return 0;
	
	return *it;
}

/*************************************************************************/

Record const *Collection::find(Record::identifier_type id) const
{
	id_hashlist::const_iterator it = m_members.find_sorted(id);
	
	if (it == m_members.end())
		return 0;
	
	return *it;
}

/*************************************************************************/

size_t Collection::memory_usage() const
{
	size_t result = sizeof(Collection);
	/* Guess how much memory std::list adds for each record... =) */
	result += (m_members.size() * sizeof(Record *) * 3);
	
	id_hashlist::const_iterator it, end;
	
	for (it = m_members.begin(), end = m_members.end(); it != end; ++it)
		result += (*it)->memory_usage();
	
	return result;
}

/*************************************************************************/

bool Collection::readable() const
{
	 return (m_db ? m_db->readable() : true);
}

/*************************************************************************/

bool Collection::remove(Record *record, int *err)
{
	/* Remove owned records */
	owned_list::iterator it, end;
	
	for (it = m_owned.begin(), end = m_owned.end(); it != end; ++it)
		if (!(*it)->remove_owned(this, record, err))
			return false;
	
	/* Erase the record from the database */
	if (record->db_status() & Record::dbsInsert)
	{
		if (!remove_db(record))
		{
			if (err)
				*err = ceDatabase;
			return false;
		}
	
		record->db_status_remove(Record::dbsInsert);
	}
	
	/* Remove the record from the list */
	if (record->db_status() & Record::dbsMember)
	{
		remove_list(record);
		record->db_status_remove(Record::dbsMember);
	}
	
	/* Free the record! ;) */
	delete record;
	
	status_add(csChanges);
	return true;
}

/*************************************************************************/

bool Collection::remove_db(Record *record)
{
	if (!m_db)
		return true;
	
	return m_db->delete_(record);
}

/*************************************************************************/

void Collection::remove_list(Record *record)
{
	m_members.remove_sorted(record->id());
}

/*************************************************************************/

bool Collection::remove_owned(Collection const *coll,
		Record const *record, int *err)
{
	if (!(record->db_status() & Record::dbsMember))
		return true;
	
	if (!remove_owned_db(coll, record, err))
		return false;
	
	return remove_owned_list(coll, record, err);
}

/*************************************************************************/

bool Collection::remove_owned_db(Collection const *coll,
		Record const *record, int *err)
{
	if (m_db)
	{
		Field const *field = owner_field(coll);
	
		if (!field)
		{
			*err = ceUnknown;
			return false;
		}
		
		if (!m_db->delete_owned(field, record->id()))
		{
			*err = ceDatabase;
			return false;
		}
	}
	
	return true;
}

/*************************************************************************/

bool Collection::remove_owned_list(Collection const *coll,
		Record const *record, int *err)
{
	id_hashlist::iterator cur, it, end;
	
	for (it = m_members.begin(), end = m_members.end(); it != end; )
	{
		cur = it++;
		
		if ((*cur)->owned(coll, record))
		{
			(*cur)->db_status_remove(Record::dbsInsert);
		
		 	if (!remove(*cur, err))
				return false;
		}
	}
	
	return true;
}

/*************************************************************************/

bool Collection::writable() const
{
	 return (m_db ? m_db->writable() : true);
}

/*************************************************************************/

bool NameCollection::add_list(Record *record)
{
	NameRecord *nr = (NameRecord *)record;
	
	if (m_nmembers.add_sorted(nr->name(), nr) == m_nmembers.end())
		return false;
	
	if (!Collection::add_list(record))
	{
		m_nmembers.remove_sorted(nr->name());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool NameCollection::change_name(NameRecord *record, char const *newname)
{
	if (!(record->db_status() & Record::dbsMember))
		return true;
	
	/* If the name hasn't changed, don't bother doing this. */
	if (*record == newname)
		return true;
	
	if (m_nmembers.add_sorted(newname, record) == m_nmembers.end())
		return false;
	
	m_nmembers.remove_sorted(record->name());
	return true;
}

/*************************************************************************/

void NameCollection::clear_list()
{
	/* Clear the base class first. */
	Collection::clear_list();
	
	m_nmembers.clear();
}

/*************************************************************************/

size_t NameCollection::memory_usage() const
{
	size_t result = Collection::memory_usage();
	/* Guess how much memory std::list adds for each record... =) */
	result += (m_nmembers.size() * sizeof(Record *) * 3);
	
	return result;
}

/*************************************************************************/

NameRecord *NameCollection::nfind(char const *name)
{
	name_hashlist::iterator it = m_nmembers.find_sorted(name);
	
	if (it == m_nmembers.end())
		return 0;
	
	return *it;
}

/*************************************************************************/

NameRecord const *NameCollection::nfind(char const *name) const
{
	name_hashlist::const_iterator it = m_nmembers.find_sorted(name);
	
	if (it == m_nmembers.end())
		return 0;
	
	return *it;
}

/*************************************************************************/

void NameCollection::remove_list(Record *record)
{
	Collection::remove_list(record);
	
	NameRecord *nr = (NameRecord *)record;
	m_nmembers.remove_sorted(nr->name());
}

/*************************************************************************/

} /* namespace databases */

/*************************************************************************/
