/* Database.
 *
 * PegSoft databases library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file databases/database.h
  * \brief Database class header file. */

#ifndef DATABASES_DATABASE_H
#define DATABASES_DATABASE_H

/*************************************************************************/

#include <list>
#include <string>
#include <vector>

#include <databases/record.h>
#include <modules/interface.h>

/*************************************************************************/

namespace modules { class Core; }

/*************************************************************************/

namespace databases
{

/*************************************************************************/

class Collection;
class Field;

/*************************************************************************/

/// Database interface.

/** This interface must be used as a base class by modules providing
  * a database driver (which are often referred to as database modules).
  *
  * Basically, what database drivers do is loading and saving records using
  * the functions provided by the Record class and a list of fields that
  * describe the values contained in the record which can then be retrieved
  * without having to know how the record manages its data internally. There
  * is also good error detection provided by the records so that the 
  * database driver can correct or discard a record it is loading in case
  * there is bogus data in its values.
  *
  * So to sum it up, the goal of database drivers is to take care of all the
  * details pertaining to its storage method while complying to a generic
  * interface so that they can be reused by all other modules needing to store
  * data persistently. And since there's no need for the records to know
  * how they're stored, they can use whatever database driver they want
  * without having to write extra code that would be painful to maintain
  * in the long term.
  *
  * To use a database, you must get an instance of a database from a
  * module through its Module::factory function, then set its name,
  * version information, permissions, add fields and finally call
  * the start() function.
  *
  * The start_real() function must be overriden in derived classes.
  * The generate_expfields(), reg(), readable() and writable(), field_create(),
  * field_add(), insert(), update_s(), update_i(), update_ui(), update_ut(),
  * update_ri(), delete_(), delete_owned() and truncate() functions may be
  * overriden too.
  *
  * This interface must be exposed by database modules through the
  * Module::factory() function. */

class Database : public modules::Interface
{
	public:
		/* Types */
		
		/// Type of the list of databases
		typedef std::list<Database *> container_type;
		/// Const iterator type
		typedef container_type::const_iterator const_iterator;
		/// Const reverse iterator type
		typedef container_type::const_reverse_iterator const_reverse_iterator;
		
		/// Type of the list of fields
		typedef std::vector<Field *> fields_type;
		
		/* Constants */
		
		/// Interface identifier
		static modules::PSIID const IID = 0x42d05105;
		
		/* Static functions */
		
		/// Finds a database by name.
		/** Finds a database by name, using the global list of databases
		  * that is maintained internally with started databases.
		  *
		  * \param[in] name The name of the database to find.
		  * \return A pointer to the database if it was found, 0 otherwise. */
		static Database *find(std::string const &name);
		
		/// Returns a const iterator to the first database in the list.
		static const_iterator begin() { return g_list.begin(); }
		/// Returns a const iterator to the last database in the list.
		static const_iterator end() { return g_list.end(); }
		
		/// Returns a const reverse iterator to the first database in the list in reverse order.
		static const_reverse_iterator rbegin() { return g_list.rbegin(); }
		/// Returns a const reverse iterator to the last database in the list in reverse order.
		static const_reverse_iterator rend() { return g_list.rend(); }
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] core The core module. */
		explicit Database(modules::Core &core) :
				m_core(&core), m_coll(0), m_name(), m_impver(0),
				m_expver(0), m_minver(0), m_maxver(0), m_perms(0),
				m_fields(), m_expfields(), m_impfields(), m_status(0) { }
		/// Destructor.
		virtual ~Database();
		
		/// "Starts" the database.
		/** "Starts" the database. This function must be called by user
		  * modules in their Module::start() function (to make sure the
		  * database module has been started).
		  *
		  * This function ensures properties have been set (for those
		  * that _MUST_ be set, at least ;), calls generate_expfields(),
		  * calls generate_impfields() if the dpInitial permission is
		  * not set, then calls the reg() function to add the database
		  * to the databases list (and possibly other lists if it is
		  * overriden in derived classes).
		  *
		  * Finally, the start_real() function is called.
		  *
		  * \return true if successful, false otherwise. */
		bool start();
		
		/// Inserts a record in the database.
		/** Inserts a record in the database. This function is usually
		  * called by the Collection::add_db() function of the associated
		  * collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] record The record to insert.
		  * \return true if successful, false otherwise. */
		virtual bool insert(Record const *record);
		
		/// Updates a string value of a record.
		/** Updates a string value of a record. This function is usually
		  * called by the Collection::change_s() function of the associated
		  * collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] record The record to update.
		  * \param[in] field The record's field to update.
		  * \param[in] value The new value of the field of the record.
		  * \return true if successful, false otherwise. */
		virtual bool update_s(Record const *record, Field const *field,
				char const *value);
		
		/// Updates an integer value of a record.
		/** Updates an integer value of a record. This function is usually
		  * called by the Collection::change_i() function of the associated
		  * collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] record The record to update.
		  * \param[in] field The record's field to update.
		  * \param[in] value The new value of the field of the record.
		  * \return true if successful, false otherwise. */
		virtual bool update_i(Record const *record, Field const *field,
				int32_t value);
		
		/// Updates an unsigned integer value of a record.
		/** Updates an unsigned integer value of a record. This function is
		  * usually called by the Collection::change_ui() function of the
		  * associated collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] record The record to update.
		  * \param[in] field The record's field to update.
		  * \param[in] value The new value of the field of the record.
		  * \return true if successful, false otherwise. */
		virtual bool update_ui(Record const *record, Field const *field,
				uint32_t value);
		
		/// Updates a Unix timestamp value of a record.
		/** Updates a Unix timestamp value of a record. This function is
		  * usually called by the Collection::change_ut() function of the
		  * associated collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] record The record to update.
		  * \param[in] field The record's field to update.
		  * \param[in] value The new value of the field of the record.
		  * \return true if successful, false otherwise. */
		virtual bool update_ut(Record const *record, Field const *field,
				time_t value);
		
		/// Updates a record identifier value of a record.
		/** Updates a record identifier value of a record. This function is
		  * usually called by the Collection::change_ri() function of the
		  * associated collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] record The record to update.
		  * \param[in] field The record's field to update.
		  * \param[in] value The new value of the field of the record.
		  * \return true if successful, false otherwise. */
		virtual bool update_ri(Record const *record, Field const *field,
				Record::identifier_type value);
		
		/// Updates multiples values of a record.
		/** Updates multiples values of a record. This function is
		  * usually called by the Collection::change_multi() function of
		  * the associated collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * Note that unlike the other update_ functions, it is up
		  * to the implementers to check if the fields are indeed
		  * exportable through a call to field_find(name, vtExport);
		  *
		  * \param[in] record The record to update.
		  * \param[in] valcount The number of values that are being
		  * changed at once. There must be one Field pointer and one
		  * value of the corresponding field type (char * for
		  * Field::ftString, int32_t for Field::ftInteger, uint32_t
		  * for Field::ftUnsigned, time_t for Field::ftUnixTime and
		  * Record::identifier_type for Field::ftIdentifier) arguments
		  * for each.
		  * \param[in] args The arguments.
		  * \return true if successful, false otherwise. */
		virtual bool update_multi(Record const *record,
				unsigned int valcount, va_list args);
		
		/// Updates owned records in the database.
		/** Updates owned records in the database, i.e. records whose
		  * owner field's value is the given record identifier, to
		  * a new value.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] field The owner field.
		  * \param[in] id The record identifier of the owner.
		  * \param[in] new_id The new record identifier of the owner.
		  * \return true if successful, false otherwise. */
		virtual bool update_owned(Field const *field,
				Record::identifier_type id, Record::identifier_type new_id);
		
		/// Deletes a record from the database.
		/** Deletes a record from the database. This function is usually
		  * called by the Collection::remove_db() function of the associated
		  * collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] record The record to delete.
		  * \return true if successful, false otherwise. */
		virtual bool delete_(Record const *record);
		
		/// Deletes owned records from the database.
		/** Deletes owned records from the database, i.e. records whose
		  * owner field's value is the given record identifier.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \param[in] field The owner field.
		  * \param[in] id The record identifier of the owner.
		  * \return true if successful, false otherwise.
		  * \sa Collection::remove_owned(). */
		virtual bool delete_owned(Field const *field,
				Record::identifier_type id);
		
		/// Clears all records from the database.
		/** Clears all records from the database. This function is usually
		  * called by the Collection::clear_db() function of the associated
		  * collection.
		  *
		  * The default implementation returns false if the dpExport
		  * permission is not set or writable() yields false, else it
		  * just returns true.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool truncate();
		
		/// Returns the collection associated with the database.
		Collection *coll() { return m_coll; }
		/// Sets the collection associated with the database.
		/** Sets the collection associated with the database. The database
		  * will add, update and remove records through this collection.
		  * It must remain valid until the database is destructed (the
		  * database destructor is guaranteed not to use the collection.)
		  *
		  * \param[in] coll The collection.
		  * \return The collection associated with the database. */
		Collection *coll(Collection *coll);
		
		/// Returns the database name.
		std::string const &name() const { return m_name; }
		/// Sets the database name.
		/** Sets the database name. The name must only contain characters
		  * 0-9A-Za-z_ (no checks are made so you must make sure it
		  * is correct.)
		  *
		  * Setting the name when the dsRegistered flag is set in the database
		  * status has no effect.
		  *
		  * \param[in] name The name.
		  * \return The database name. */
		std::string const &name(std::string const &name);
		
		/// Versions type used to set and retrieve version numbers.
		enum
		{
			/// The version number of the imported data.
			vtImport = 1,
			/// The version number for the exported data.
			vtExport = 2,
			/// The minimum version number.
			vtMinimum = 3,
			/// The maximum version number.
			vtMaximum = 4
		};
		
		/// Returns version numbers.
		/** Returns version numbers that have been set for the
		  * database. Each database has its own version number
		  * that allows to handle the data differently for different
		  * versions.
		  *
		  * \param[in] type The type of version to return, which
		  * can be one of the following values:
		  *   - vtImport: The version number of imported data. After
		  *     the initial loading of the data, this version is the
		  *     same as the vtExport version if the dpExport
		  *     permission is set. This may be zero if the database
		  *     didn't exist prior to initial loading. If the dpInitial
		  *     permission is not set on the database, but dpExport
		  *     is set, this is always the same as the vtExport version.
		  *   - vtExport: The version number of exported data. This
		  *     is therefore the version number of the database after
		  *     the initial loading. This version may be zero if the
		  *     dpExport permission is not set.
		  *   - vtMinimum: The minimum version number required. If the
		  *     database has a lower version, initial loading won't
		  *     be done (the dsInitial status flag is still set in any
		  *     case though). This means that, if the dpExport permission
		  *     is set, the contents of the database is removed.
		  *     This version may be zero only if the dpInitial permission
		  *     is not set.
		  *   - vtMaximum: The minimum version number required. If the
		  *     database has a higher version, initial loading won't
		  *     be done(the dsInitial status flag is still set in any
		  *     case though). This means that, if the dpExport permission
		  *     is set, the contents of the database is removed.
		  *     If the maximum version is zero, then there won't be
		  *     any check for the maximum version number.
		  * \return The requested version number. */
		uint16_t version(int type) const;
		
		/// Sets version numbers.
		/** Sets version numbers for the database.
		  *
		  * Setting version numbers when the dsRegistered flag is
		  * set in the database status has no effect.
		  *
		  * \param[in] type The type of version to set, which
		  * can be either vtExport, vtMinimum or vtMaximum.
		  * \param[in] ver The version number.
		  * \return The new version number for the given type.
		  * \sa version(int) */
		uint16_t version(int type, uint16_t ver);
		
		/// Possible permissions
		enum
		{
			/// Initial loading allowed
			dpInitial = 0x0001,
			/// Creation of records after initial loading allowed
			dpCreate = 0x0002,
			/// Update of records allowed
			dpUpdate = 0x0004,
			/// Deletion of records allowed
			dpDelete = 0x0008,
			/// Records export allowed
			dpExport = 0x0010,
			
			/// Same as dpCreate | dpUpdate | dpDelete
			dpImport = dpCreate | dpUpdate | dpDelete
		};
		
		/// Returns database permissions
		uint16_t perms() const { return m_perms; }
		/// Adds database permissions.
		/** Adds database permissions. Note that permissions describe
		  * what the database driver is allowed to do, not what it HAS to do
		  * (whether or not it does support all features is up to the
		  * implementer).
		  *
		  * Adding permissions when the dsRegistered flag is set in the
		  * database status has no effect.
		  *
		  * \param[in] perms An or'ed set of the following permissions:
		  *   - dpInitial: The database driver is allowed to add records to the
		  *     collection when the start() function is called using the current
		  *     contents of its database. This is called initial loading. Note
		  *     that it'll be done only if the database versions match the
		  *     requirements. The dsInitial status is set on the database
		  *     when initial loading has been done.
		  *   - dpCreate: The database driver is allowed to add records to the
		  *     collection even after the start() function has been called.
		  *   - dpUpdate: The database driver is allowed to update records that
		  *     are in the collection.
		  *   - dpDelete: The database driver is allowed to remove records from
		  *     the collection.
		  *   - dpExport: The database driver is allowed to add, update and
		  *     remove records in its database. The new version of the
		  *     database becomes the version set using the vtExport
		  *     constant. If this perm is set and initial loading isn't done,
		  *     then the database contents is entirely removed.
		  * \return The database permissions.
		  * \sa version(int) */
		uint16_t perms_add(uint16_t perms);
		/// Removes database permissions.
		/** Removes database permissions.
		  *
		  * Removing permissions when the dsRegistered flag is set in the
		  * database status has no effect.
		  *
		  * \param[in] perms The permissions to remove, which are the
		  * same as those for perms_add().
		  * \return The database permissions.
		  * \sa perms_add() */
		uint16_t perms_remove(uint16_t perms);
		
		/// Creates a new field for this database.
		/** Creates a new field for this database. The created field can
		  * be freed by calling operator delete on it.
		  *
		  * The default implementation creates a field of the class Field.
		  * It is however possible to override this function to create
		  * a field of a derived class.
		  *
		  * \param[in] name The name of the field, which will be set on the
		  * newly created field through a call to Field::name().
		  * \param[in] minver The minimum version number, which will be set
		  * on the newly created field through a call to
		  * Field::version(vtMinimum, minver).
		  * \param[in] type The type of the field, which will be set on the
		  * newly created field through a call to Field::type().
		  * \param[in] size The size of the field, which will be set on the
		  * newly created field through a call to Field::size().
		  * \param[in] attrs The attributes of the field, which will be set
		  * on the newly created field through a call to Field::attrs_add().
		  * \param[in] def The default value of the field, which will be set
		  * on the newly created field through a call to
		  * Field::default_value().
		  * \param[in] tag The tag of the field, which will be set on the
		  * newly created field through a call to Field::tag().
		  * \return The new field, or 0 on error. */
		virtual Field *field_create(char const *name, uint16_t minver,
				uint16_t type, size_t size, uint16_t attrs, char const *def,
				uint16_t tag = 0);
		
		/// Adds a field to this database.
		/** Adds a field to this database, which should have been created
		  * through a call to the field_create() function. After the field
		  * is added to the database, it is owned by it; the database will
		  * therefore free it when it is destructed.
		  *
		  * This function will automatically call the Field::setup() function
		  * before it is added to the database. No additional checks are made.
		  *
		  * \param[in] field The field to add to the database.
		  * \return true if successful, false otherwise (including if field
		  * is NULL). */
		virtual bool field_add(Field *field);
		
		/// Finds a field in this database.
		/** Finds a field in this database by name.
		  *
		  * \param[in] name The name of the field to look for.
		  * \param[in] restrict The restriction to apply when looking
		  * for a field. This can be one of the following value:
		  *   - 0 : no restriction (the default). Any field with this name
		  *         will be returned.
		  *   - vtImport: limit the search to the imported fields,
		  *     that may have been generated through a call to
		  *     generate_impfields(). If this call has not been done by
		  *     the database driver, then the function always returns 0.
		  *   - vtExport: limit the search to the exported fields, that
		  *     are generated through a call to generate_expfields(). This
		  *     call is done when the start() function is called if the
		  *     dpExport permission is set; until that point, the function
		  *     always returns 0.
		  * \return true if successful, false otherwise. */
		Field *field_find(char const *name, int restrict = 0);
		
		/// Returns whether data can potentially be read from the database.
		/** Returns whether data can potentially be read from the database
		  * by the database driver. The result is totally independant of
		  * whether the permissions set for the database allow it to
		  * import records or not; it's just meant to be used as an
		  * indicator of whether the database can be read.
		  *
		  * The default implementation always returns true.
		  *
		  * \return true if the database is readable, false otherwise. */
		virtual bool readable() const { return true; }
		
		/// Returns whether data can potentially be written to the database.
		/** Returns whether data can potentially be written to the database
		  * by the database driver. The result is totally independant of
		  * whether the permissions set for the database allow
		  * it to export records or not; it's just meant to be used as an
		  * indicator of whether the database can be written.
		  *
		  * The default implementation returns true if the
		  * modules::Core::pfReadOnly program flag is not set for the core,
		  * false otherwise.
		  *
		  * \return true if the database is writable, false otherwise. */
		virtual bool writable() const;
		
		/// Possible status bits
		enum
		{
			/// reg() has been called
			dsRegistered = 0x0001,
			/// start_real() has been called
			dsStarted = 0x0002,
			/// Initial loading has been done.
			dsInitial = 0x0004
		};
		
		/// Returns database status.
		uint16_t status() const { return m_status; }
	protected:
		/// Core module.
		modules::Core *m_core;
		/// The collection associated to the database
		Collection *m_coll;
	
		/// Name of the database
		std::string m_name;
		
		/// The version number of the imported data.
		uint16_t m_impver;
		/// The version number for the exported data.
		uint16_t m_expver;
		/// Minimum version required for initial loading.
		uint16_t m_minver;
		/// Maximum version allowed for initial loading.
		uint16_t m_maxver;
		
		/// Permissions of the database.
		uint16_t m_perms;
		
		/// List of the fields.
		fields_type m_fields;
		/// List of exported fields.
		fields_type m_expfields;
		/// List of imported fields.
		fields_type m_impfields;
		
		/// Status of the database.
		uint16_t m_status;
		
		/// Generates the list of exported fields.
		/** Generates the list of exported fields (the m_expfields member).
		  * The exported fields are fields that match the vtExport version
		  * of the database. This is called from the start() function before
		  * the database is registered, but only if the dpExport perm is set.
		  *
		  * \return true if successful, false otherwise. */
		virtual void generate_expfields();
		
		/// Generates the list of imported fields.
		/** Generates the list of imported fields (the m_impfields member).
		  * The imported fields are fields that match the vtImport version
		  * of the database. This may be called by the database driver, if
		  * needed, after the version of the imported has been set. It can
		  * be called a second time once initial loading has taken place,
		  * and in this case, the generated list will be the same as the
		  * one generated through generate_expfields().
		  *
		  * The start() function calls this function right after it calls
		  * generate_expfields() when the dpInitial permission is not set.
		  *
		  * \return true if successful, false otherwise. */
		virtual void generate_impfields();
		
		/// Registers the database.
		/** Registers the database.
		  *
		  * The default implementation adds the database to the list
		  * of databases. This is _vital_ for good operation, and you
		  * should therefore always call the base version of the function
		  * if you inherit it in derived classes.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool reg();
		
		/// "Starts" the database, overridable version.
		/** "Starts" the database. This must be overriden in derived
		  * classes.
		  *
		  * This function must do initial loading if the dpInitial
		  * permission is set, and should do so according to the procedure
		  * described in the perms_add() function.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool start_real() = 0;
	private:
		static container_type g_list;
	
		Database(Database const &);
		Database &operator=(Database const &);
};

/*************************************************************************/

/// Database removal interface.

/** This interface may be implemented by database modules to allow
  * for safe removal of databases.
  *
  * This interface must be exposed through the Module::factory()
  * function. */

class DBRemove : public modules::ErrorInterface
{
	public:
		/// The interface identifier.
		static modules::PSIID const iid = 0x43f0f5f3;
		
		/// Removes a database.
		/** Removes a database.
		  *
		  * \param[in] database The name of the database to remove.
		  * \return true if the database could be removed, false
		  * otherwise. */
		virtual bool remove(std::string const &database) = 0;
		
		/// Removes all databases.
		/** Removes all databases.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool clear() = 0;
	private:
};

/*************************************************************************/

} /* namespace databases */

/*************************************************************************/

#endif /* DATABASES_DATABASE_H */
