/* A brick.
 *
 * PegSoft modules subsystem library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file modules/module.h
 * \brief Basic module header file.
 */

#ifndef MODULES_MODULE_H
#define MODULES_MODULE_H

/*************************************************************************/

#include <string>

#include <modules/interface.h>

/*************************************************************************/

namespace conf { class ConfQ; class ConfGlobalQ; }
namespace modules { class Core; }

/*************************************************************************/

namespace modules
{

/*************************************************************************/

/// Module base class.

/** This class is the base class for all modules. Each module must 
  * create a new class derived from it and use it as the class of
  * their module object.
  *
  * The version() function must be overriden in derived classes.
  * The config(), factory(), interface(), load(), memory_usage(),
  * notify(), query(), start(), stop() and unload() functions are
  * possible candidates as well.
  *
  * The module manager calls the functions in the following order
  * to load and start a module: 
  *
  *     - config() : to get configuration information;
  *     - load(): to initialize the module;
  *     - notify() : to notify the module about other modules already loaded;
  *     - start() : to start the module;
  *     - notify() : to notify the module about other modules already started.
  *
  * The module manager calls the functions in the following order
  * to stop and unload a module:
  *
  *     - stop() : to stop the module;
  *     - notify() : to tell the module all other modules are stopped;
  *       while not true, it'll allow the module to remove hooks and references
  *       to the other modules;
  *     - unload() : to unload the module.
  *
  * When the program stops (or restarts), the stop() function is called for
  * each module in the reverse order they were started. unload() is NOT called
  * and there is NO notification sent so everything vital (saving data, etc.)
  * MUST be done by the stop() function, while cleaning up (such as removing
  * hooks and references or freeing memory) may be left out of it as the
  * program will stop anyway. */

class Module
{
	public:
		/// Returns the module version number
		/** Returns the module version number, which must be
		  * a hexadecimal number of the following format:
		  *
		  * 0xMMmmRRRR
		  *
		  * where MM is the major version number, mm the minor
		  * version number and RRRR the release number.
		  *
		  * Example: 0x01050010 => 1.5.16
		  *
		  * Modules should check the version number of other
		  * modules they use to ensure that their version is
		  * the same as the one for which the module was
		  * compiled. */
		virtual int version() const = 0;
	
	public:
		/// Constructor.
		/** Module constructor that sets the module name and
		  * description.
		  * 
		  * \param[in] name The module name (must be the same as
		  * the shared library filename, with no extension).
		  * \param[in] description The module description, whose
		  * contents can be freely chosen by the module developer.
		  * \param[in] author The author of the module's name
		  * (for informative purpose only).
		  * \param[in] url An URL where information about the
		  * module and latest versions can be found (for informative
		  * purpose only). */
		Module(std::string const &name, std::string const &description,
				std::string const &author = "", std::string const &url = "") :
				m_name(name), m_description(description), m_author(author),
				m_url(url), m_references(0), m_state(0), m_data(0) { }
		/// Destructor.
		virtual ~Module() { }
		
		/// Returns configuration information.
		/** Returns configuration information, in the form of a
		  * conf::ConfQ parser filled with configuration directives
		  * that will be used to configure the module.
		  *
		  * \param[in] core The core module.
		  * \param[in] global The ConfGlobalQ object to which you
		  * can add global directives, variables and include paths.
		  *
		  * \return A pointer to the configuration parser, or 0
		  * if the module doesn't need to be configured.
		  * \sa conf::ConfQ */
		virtual conf::ConfQ *config(Core &core,
				conf::ConfGlobalQ &global) { return 0; }
		
		/// Retrieves an interface object exposed by the module.
		/** Retrieves an interface object exposed by the module,
		  * using an unique identifier.
		  *
		  * \param[in] iid The interface identifier.
		  *
		  * \return A pointer to an object of the requested
		  * interface class, or 0 if the module doesn't expose
		  * an object of this interface. The object returned
		  * by this function is owned by the caller, which must
		  * free it whenever it's no longer needed. To get an
		  * object owned by the provider, use the interface()
		  * function. Note that modules are not required to
		  * provide an object of a given interface through the
		  * interface() function even if they provide it
		  * through this function.
		  *
		  * \sa Interface, interface() */
		virtual Interface *factory(PSIID iid) { return 0; }
		
		/// Retrieves an interface object exposed by the module.
		/** Retrieves an interface object exposed by the module,
		  * using an unique identifier.
		  *
		  * \param[in] iid The interface identifier.
		  *
		  * \return A pointer to an object of the requested
		  * interface class, or 0 if the module doesn't expose
		  * an object of this interface. The object returned
		  * by this function is owned by its module and must
		  * not be freed by the caller. To get an object
		  * owned by the caller, use the factory() function.
		  * Note that modules are not required to provide an
		  * object of a given interface through the factory()
		  * function even if they provide it through this
		  * function.
		  *
		  * \sa Interface, factory() */
		virtual Interface *interface(PSIID iid) { return 0; }
		
		/// Loads the module.
		/** Asks the module to load itself. When this function
		  * returns, the module must be ready to get "hooked"
		  * (that is, used by other modules) and the msLoaded
		  * bit must have been added to the module state.
		  *
		  * This function is also ideal to ensure other modules
		  * you'll need are there and load them if not (but don't
		  * add references or hook other modules -- wait until
		  * the notify() function is called later). Modules
		  * loaded in this function are guaranteed to be 
		  * started BEFORE this module starts, which may be
		  * useful.
		  *
		  * The load() function inherited from Module just
		  * changes the module state and returns true.
		  *
		  * \param[in] core The core module.
		  *
		  * \return true if successful, false otherwise (in the
		  * latter case, module will be released by the modules
		  * manager.)
		  * \sa notify() */
		virtual bool load(Core &core);
		
		/// Returns information about module's memory usage.
		/** Returns information about module's memory usage.
		  *
		  * \return The module memory usage, expressed in
		  * bytes, or -1 if this information is unavailable
		  * (which is what does the default implementation). */
		ssize_t memory_usage() const { return -1; }
		
		/// Notification messages for the notify() function.
		enum
		{
			/// Don't use.
			mnNone = 0,
			/// The module is loaded.
			mnLoaded = 1,
			/// The module is started.
			mnStarted = 2,
			/// The module is stopped.
			mnStopped = 3,
			/// The module must sync.
			/** This message is generally handled only by database modules.
			  * They must sync their databases when they receive it
			  * (whatever that means, if anything, is up to them). */
			mnSync = 4,
			/// All modules have been reconfigured.
			/** This message is sent when the configuration files for
			  * all modules have been reloaded. The sender is the core
			  * module. */
			mnReconfig = 5,
			
			/// Core-defined messages will be (mnCoreDefined + x)
			mnCoreDefined = 512,
			/// User-defined messages must be (mnUserDefined + x).
			mnUserDefined = 4096
		};
		
		/// Notifies a message to a module.
		/** Notifies a message to a module, described by the message
		  * number and parameter. The function must ignore all
		  * messages it doesn't recognize.
		  *
		  * \param[in] sender The module that sent the notification.
		  * \param[in] msg The notification message. The standard
		  * messages are:
		  *   - mnLoaded: the module sending the message is loaded.
		  *     You may now hook it or add references to it if you
		  *     need to. After this module is loaded, this message
		  *     is sent automatically once for each loaded module,
		  *     including those that were loaded BEFORE this one.
		  *   - mnStarted: the module sending the message is started.
		  *     After this module is started, this message is sent
		  *     automatically once for each loaded module, including
		  *     those that were started BEFORE this one.
		  *   - mnStopped: the module sending the message is stopped.
		  *     You MUST remove clean up any hooks *and references* you
		  *     set up previously. After this module is started, this
		  *     message is sent automatically once for each stopped
		  *     module. Additionally, you'll get this message for each
		  *     module being started when this module is going to be
		  *     unloaded, after stop() has been called.
		  *
		  * \param[in] arg An optional parameter that may be used
		  * by the sender to pass additional data. */
		virtual void notify(Module &sender, int msg, void *arg) { }
		
		/// Queries a module.
		/** Queries a module. This function may be used to pass special
		  * module-defined commands to the modules for it to execute,
		  * with feedback.
		  *
		  * The default implementation always returns false.
		  *
		  * \param[in] query The query that must be executed by the
		  * module. What possible queries are and how they're
		  * interpreted is entirely left to the choice of the module
		  * writer. There are no standard queries.
		  * \param[out] feedback A feedback string should be set by
		  * the implementer to this parameter before returning, if
		  * the query is supported. It can be an empty string; in
		  * that case, the caller should consider that the module
		  * didn't provide any feedback.
		  * \return true if the query is supported (even if it was
		  * not successful), false otherwise. */
		virtual bool query(std::string const &query, std::string &feedback)
				{ return false; }
		
		/// Starts the module.
		/** Asks the module to start itself. The module state
		  * must be updated accordingly. What the module does
		  * in this function is completely free.
		  *
		  * This function must either success or the Core::fatal()
		  * function must be called.
		  *
		  * The start() function inherited from Module just
		  * changes the module state.
		  *
		  * \param[in] core The core module. */
		virtual void start(Core &core);
		
		/// Stops the module.
		/** Asks the module to stop itself. The module must stop
		  * its operations and do any important task that needs
		  * to be done, such as saving data, as this is the only
		  * function that will be called when the program stops.
		  * Removing hooks and references and freeing memory, on
		  * the other hand, is better left to the notify() and
		  * unload() function since they're meaningless if the
		  * program will stop immediately.
		  *
		  * The stop() function inherited from Module just
		  * changes the module state and returns true.
		  *
		  * \param[in] core The core module.
		  *
		  * \return true if successful, false otherwise (in that
		  * case, module unloading will fail there, so it's better
		  * to not return false unless it's a temporary error and
		  * module operation can continue; the return value is
		  * ignored when the program stops). */
		virtual bool stop(Core &core);
		  		
		/// Unloads the module.
		/** Asks the module to unload itself. The module must
		  * do all the remaining cleanup and update the state so
		  * that, if the load() was called afterwards, it
		  * would work perfectly.
		  *
		  * This function must either success or the Core::fatal()
		  * function must be called.
		  *
		  * The unload() function inherited from Module just
		  * changes the module state.
		  *
		  * \param[in] core The core module. */
		virtual void unload(Core &core);
		
		/// Returns the module name.
		virtual std::string const &name() const { return m_name; }
		/// Returns the module description.
		virtual std::string const &description() const { return m_description; }
		/// Returns the module author.
		virtual std::string const &author() const { return m_author; }
		/// Returns the module URL (where new versions can be found.)
		virtual std::string const &url() const { return m_url; }
		
		/// Returns the reference count.
		/** Returns the number of references that have been added
		  * to the module.
		  *
		  * \return The reference count.
		  * \sa references_add() */
		int references() const { return m_references; }
		/// Adds a reference to the module.
		/** When another module needs to prevent this module from
		  * being unloaded, it adds a reference through this
		  * function. A module can only be unloaded when its reference
		  * count is zero (it can be stopped regardless of it though --
		  * for example when the program stops).
		  *
		  * This function increments the reference count by one.
		  *
		  * You should only add a reference when a module absolutely CAN'T
		  * continue without the module being referenced. If it still can
		  * work without it (perhaps with altered functionality), then it
		  * should not add a reference and watch for its unloading through
		  * the notify() function.
		  *
		  * \return The new reference count.
		  * \sa notify() */
		int references_add() { return ++m_references; }
		/// Removes a reference to the module.
		/** Removes a reference to the module (decreases the
		  * reference count by one).
		  *
		  * \return The new reference count. */
		int references_remove() { return --m_references; }
		
		/// Module state bits.
		enum
		{
			/// The module is loaded, ready to be "hooked" by other modules.
			msLoaded = 0x1,
			/// The module is started.
			msStarted = 0x2
		};
		
		/// Returns the module state.
		/** Returns the module state, which is an OR'ed combination
		  * of any of the msLoaded or msStarted bits.
		  *
		  * This helps other modules to determine whether a module
		  * is loaded or started and must be updated accordingly
		  * by the load(), start(), stop() and unload() functions.
		  *
		  * \return the module state. */
		virtual int state() const { return m_state; }
		
		/// Returns the data set by the module manager.
		void *data() const { return m_data; }
		/// Sets arbitrary data (for use by the module manager only!)
		void data(void *data) { m_data = data; }
	protected:
		/// The module name
		std::string m_name;
		/// The module description
		std::string m_description;
		/// The module author.
		std::string m_author;
		/// The module URL.
		std::string m_url;
		
		/// The module reference count.
		int m_references;
		
		/// The module state.
		int m_state;
	private:
		void *m_data;
	
		Module(Module const&);
		Module &operator=(Module const &);
};

/*************************************************************************/

} /* namespace modules */

/*************************************************************************/

#endif /* MODULES_MODULE_H */
