/* Vaste programme.
 *
 * PegSoft scheduler library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file scheduler/scheduler.h
 * \brief Scheduler header file.
 */

#ifndef SCHEDULER_SCHEDULER_H
#define SCHEDULER_SCHEDULER_H

/*************************************************************************/

#include <list>
#include <string>

/*************************************************************************/

/// Namespace for all scheduler declarations.

namespace scheduler
{

/*************************************************************************/

class Task;

/*************************************************************************/

/// A task scheduler

/** The Scheduler class manages a list of Task objects and runs them
  * when they need.
  */
class Scheduler
{
	public:
		/// Type of the list of tasks.
		typedef std::list<Task *> list_type;
		
		/// Default constructor
		Scheduler() : m_tasks(), m_curtask(0), m_tasktype(0) { }
		/// Copy constructor
		Scheduler(Scheduler const &s) : m_tasks(s.m_tasks),
			m_curtask(0), m_tasktype(s.m_tasktype) { }
		/// Assignment operator
		Scheduler &operator=(Scheduler const &right);
		
		/// Registers a task type.
		/** Registers a task type. 
		  * \return A new task type ID.
		  * \sa Task */
		int register_type() { return m_tasktype++; }
		
		/// Adds a task.
		/** Adds a task to the list of scheduled tasks.
		  * \param[in] task The task to be added.
		  * \return true if successful, false otherwise. 
		  */
		bool add(Task &task);
		/// Removes a task.
		/** Removes a task from the list of scheduled tasks.
		  *
		  * \param[in] task The task to be removed. The function
		  * doesn't free the memory pointed by task; it is the
		  * caller's responsibility.
		  * \return true if successful, false otherwise. */
		bool remove(Task &task);
		/// Removes all tasks of a given type.
		/** Removes all tasks of a given type from the list of
		  * scheduled tasks. operator delete is called for each
		  * removed task.
		  *
		  * \param[in] type The type of the tasks to be removed.
		  * \return The number of tasks that have been removed. */
		list_type::size_type remove_bytype(int type);
		/// Removes all tasks with the given name and type.
		/** Removes all tasks with the given name and type from the
		  * list of scheduled tasks. operator delete is called for each
		  * removed task.
		  *
		  * \param[in] type The type of the tasks to be removed.
		  * \param[in] name The name of the tasks to be removed.
		  * \return The number of tasks that have been removed. */
		list_type::size_type remove_byname(int type, std::string const &name);
		
		/// Runs tasks.
		/** Iterates once in the list of tasks, calling Task::poll() on each
		  * task. For tasks that need to be run, Task::run() is called, and
		  * the task is either removed or left alone depending on the return
		  * value.
		  *
		  * \return true if the main loop must sleep, false if one of the tasks
		  * already slept (and there is therefore no second sleep needed). */
		bool run();
		
		/// Runs tasks of the given type.
		/** Iterates once in the list of tasks, calling Task::run() on each
		  * task of the given type. The task is either removed or left alone
		  * depending on the return value. Task::poll() is NOT called to
		  * determine whether the task should be run or not.
		  *
		  * \param[in] type The type of the tasks to be run.
		  * \return The number of tasks that have been run. */
		list_type::size_type run_bytype(int type);
		/// Runs tasks with the given name and type.
		/** Iterates once in the list of tasks, calling Task::run() on each
		  * task with the given name and type. The task is either removed or
		  * left alone depending on the return value. Task::poll() is NOT
		  * called to determine whether the task should be run or not.
		  *
		  * \param[in] type The type of the tasks to be run.
		  * \param[in] name The name of the tasks to be run.
		  * \return The number of tasks that have been run. */
		list_type::size_type run_byname(int type, std::string const &name);
		
		/// Returns the list of tasks.
		list_type const &tasks() const { return m_tasks; }
		/// Returns the current task being run by the scheduler (0 if none).
		Task const *current_task() const { return m_curtask; }
	private:
		list_type m_tasks;
		Task *m_curtask;
		
		int m_tasktype;
		
		bool run_task(list_type::iterator &task, list_type::iterator *next);
};

/*************************************************************************/

/// Base task class.

/** The Task class implements a task used by the Scheduler class.
  * Each Task object has its own task type, allowing you to remove 
  * all tasks of a specific type easily through the Scheduler
  * object.
  *
  * This class must not be used directly; instead, you must inherit it
  * and override the abstract functions. */

class Task
{
	public:
		/// Constructor setting the task type
		explicit Task(int type) : m_type(type) { }
		/// Copy constructor
		explicit Task(Task const &t) : m_type(t.m_type) { }
		/// Assignment operator
		virtual Task &operator=(Task const &right);
		/// Destructor
		virtual ~Task() { }
		
		/// Determines whether the task should be run.
		/** Determines whether the task should be run. This function
		  * is called periodically by the Scheduler object to which
		  * the class has been added.
		  *
		  * \return true if the task should be run (in that case,
		  * run() is then called by the scheduler), false otherwise.
		  * \sa run() */
		virtual bool poll() = 0;
		
		/// Possible return value bits for run()
		enum
		{
			/// Use this as an alias for 0.
			rOK = 0x0,
			/// The task has called sleep() -- no need to call it again.
			rSlept = 0x1,
			/// Remove the task from the list.
			rRemove = 0x2,
			/// Remove the task from the list and call operator delete on it.
			rFree = (rRemove | 0x4)
		};
		/// Runs the task.
		/** Runs the task. Override this function to do whatever you
		  * want there.
		  * \return An or'ed combination of any of the rSlept, rRemove
		  * or rFree bits as appropriate. There is no error return
		  * value.
		  * \sa rSlept, rRemove, rRemoveFree */
		virtual int run() = 0;
		
		/// Returns the description of the task.
		/** Returns the description of the task. It can be used by
		  * the program to show information about the running task.
		  * \return The description of the task. */
		virtual std::string description() const = 0;
		/// Returns the name of the task.
		/** Returns the name of the task. This name is used by the
		  * Scheduler::remove_bytype to remove tasks by type and name.
		  * \return The name of the task.
		  * \sa Scheduler::remove_bytype(int, std::string const &) */
		virtual std::string name() const = 0;
		/// Returns the task type.
		virtual int type() const { return m_type; }
		/// Sets the task type.
		/** Sets the task type.
		  * \param[in] type The new task type.
		  * \return The new task type. */
		virtual int type(int type) { return (m_type = type); }
	protected:
		/// The task type.
		int m_type;
};

/*************************************************************************/

/// Task class whose poll() function always yields true.

class TaskAlways : public Task
{
	public:
		/// Constructor setting the task type
		explicit TaskAlways(int type) : Task(type) { }
		
		/// Always returns true.
		virtual bool poll() { return true; }
};

/*************************************************************************/

/// Timeout task class

/** The TaskTimeout class is a task class whose poll() function yields true
  * when the given number of seconds has elapsed since the creation of the
  * task or its last run. */

class TaskTimeout : public Task
{
	public:
		/// Constructor
		/** This constructor sets the task type and the timeout.
		  * 
		  * \param[in] type The task type.
		  * \param[in] timeout The interval between two runs, in seconds
		  * (the first run being done when this amount of time
		  * expires). */
		TaskTimeout(int type, time_t timeout) : Task(type), m_timeout(timeout),
				m_lastrun(time(NULL)) { }
		/// Copy constructor
		explicit TaskTimeout(TaskTimeout const &t) : Task(t), m_timeout(t.m_timeout),
				m_lastrun(t.m_lastrun) { }
		/// Assignment operator
		virtual TaskTimeout &operator=(TaskTimeout const &right);
		
		/// Returns true and resets the last time the task was run if the interval has elapsed
		virtual bool poll();
		
		/// Returns the interval between two runs.
		time_t timeout() const { return m_timeout; }
		/// Sets the interval between two runs.
		/** Sets the interval between two runs.
		  * \param[in] timeout The new interval between two runs.
		  * \return The new interval between two runs. */
		time_t timeout(time_t timeout) { return (m_timeout = timeout); }
		
		/// Returns the last time the task was run.
		time_t last_run() const { return m_lastrun; }
	protected:
		/// Interval between two runs.
		time_t m_timeout;
		/// When was the task run for the last time?
		time_t m_lastrun;
};

/*************************************************************************/

/// Step task class

/** The TaskStep class is a task class whose poll() function yields true
  * once every predetermined amount of seconds.
  *
  * The main difference with other similar classes, such as TaskTimeout,
  * is that the base time is midnight of the day the task has been created.
  *
  * So, for example, if you create a task of this type at 1pm that must
  * run every four hours, it will be run at 4pm and 8pm of the same day,
  * and at 12am, 4am, 8am, 12pm, 4pm and 8pm of the next days until it
  * is removed.
  *
  * You can also set a shift that will modify the behaviour of the class.
  * For example, a 7200 seconds shift for the task described above will
  * make it run at 2am, 6am, 10am, 2pm, 6pm and 10pm. */

class TaskStep : public Task
{
	public:
		/// Constructor
		/** Sets the task type, step and shift (see the class description
		  * for more details).
		  *
		  * \param[in] type The task type.
		  * \param[in] step The interval between two runs, in seconds.
		  * \param[in] shift The shift, in seconds (must be less than
		  * step). */
		TaskStep(int type, time_t step, time_t shift);
		/// Copy constructor
		explicit TaskStep(TaskStep const &t) : Task(t), m_step(t.m_step),
				m_shift(t.m_shift), m_next(t.m_next) { }
		/// Assignment operator
		virtual TaskStep &operator=(TaskStep const &right);
		
		/// Returns true and when the interval has elapsed.
		virtual bool poll();
		
		/// Returns the interval between two runs, in seconds.
		time_t step() const { return m_step; }
		/// Sets the interval between two runs.
		/** Sets the interval between two runs.
		  * \param[in] step The new interval between two runs,
		  * in seconds.
		  * \return The new interval between two runs, in seconds. */
		time_t step(time_t step);
		
		/// Returns the shift, in seconds.
		time_t shift() const { return m_shift; }
		/// Sets the shift.
		/** Sets the shift.
		  * \param[in] shift The new shift.
		  * \return The new shift. */
		time_t shift(time_t shift);
		
		/// Returns when the task will run next time.
		time_t next() const { return m_next; } 
	protected:
		/// Interval between two runs.
		time_t m_step;
		/// Shift.
		time_t m_shift;
		/// When must the task run next?
		time_t m_next;
	private:
		void calc_next();
		time_t timezone_offset(time_t now);
};

/*************************************************************************/

} /* namespace scheduler */

/*************************************************************************/

#endif /* SCHEDULER_SCHEDULER_H */
