/* Dynamic buffer class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <algorithm>
#include <iostream>

#include "sockets/dbuffer.h"

using namespace std;

/*************************************************************************/

static size_t chunks_inuse[64];
static bool is_init = false;

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

DBuffer::chunks_list DBuffer::chunks_pool[64];

/*************************************************************************/

DBuffer::DBuffer(uint8_t chunkmul, size_t maxsize) : m_chunks(),
		m_chunksize(chunk_unit * chunkmul), m_poolidx(chunkmul - 1),
		m_size(0), m_maxsize(maxsize)
{
	if (is_init == false)
	{
		memset(&chunks_inuse, 0, sizeof(chunks_inuse));
		is_init = true;
	}
}

/*************************************************************************/

DBuffer::~DBuffer()
{
	clear();
}

/*************************************************************************/

DBuffer::Chunk *DBuffer::alloc()
{
	Chunk *chunk;
	
	if (!chunks_pool[m_poolidx].empty())
	{
		chunk = chunks_pool[m_poolidx].front();
		chunks_pool[m_poolidx].pop_front();
	} else
	{
		chunk = new Chunk(m_chunksize);
	}
	
	m_chunks.push_back(chunk);
	return chunk;
}

/*************************************************************************/

bool DBuffer::clear()
{
	while (m_chunks.size())
		recycle();
		
	m_size = 0;
	return true;
}

/*************************************************************************/

bool DBuffer::erase(size_t len)
{
	if (!len)
		return true;
	
	if (len > m_size)
		return false;
	
	if (len == m_size)
		return clear();
	
	for (Chunk *chunk = m_chunks.front();
			len; chunk = (len ? m_chunks.front() : 0))
	{
		size_t toerase = min(chunk->size(), len);
		
		chunk->erase(toerase);
		m_size -= toerase;
		
		if (chunk->empty() && chunk->remaining() == 0)
			recycle();
		
		len -= toerase;
	}
	
	return true;
}

/*************************************************************************/

bool DBuffer::read(size_t max, char const **data, size_t *len) const
{
	if (empty())
		return false;
	
	Chunk *chunk = m_chunks.front();
	*data = chunk->first();
	
	if (!max)
		*len = chunk->size();
	else
		*len = min(max, chunk->size());
	
	return true;
}

/*************************************************************************/

bool DBuffer::read_byte(size_type pos, char *ch) const
{
	/* If the offset is greater or equal than the size, then it
	 * is incorrect. */
	if (pos >= m_size)
		return false;
	
	Chunk *chunk = 0;
	size_t rel = 0;
	chunks_list::const_iterator it, end;
		
	for (it = m_chunks.begin(), end = m_chunks.end(); it != end; ++it)
	{
		chunk = (*it);
		
		if (rel + chunk->size() > pos)
			break;
		rel += chunk->size();
	}
	
	*ch = chunk->first()[pos - rel];
	return true; 
}

/*************************************************************************/

void DBuffer::recycle()
{
	if (m_chunks.empty())
		return;
	
	Chunk *chunk = m_chunks.front();
	m_chunks.pop_front();
	
	if (chunks_pool[m_poolidx].size() > 8 &&
			chunks_inuse[m_poolidx] < chunks_pool[m_poolidx].size())
	{
		delete chunk;
		return;
	}
	
	chunk->clear();
	chunks_pool[m_poolidx].push_back(chunk);
}

/*************************************************************************/

bool DBuffer::write(char const *data, size_t len)
{
	if (!len)
		return true;
	
	if (len > remaining())
		return false;
	
	for (Chunk *chunk = (m_chunks.empty() ? alloc() : m_chunks.back());
			len; chunk = (len ? alloc() : 0))
	{
		if (!chunk->remaining())
			continue;
		
		size_t towrite = min(chunk->remaining(), len);
		
		chunk->write(data, towrite);
		m_size += towrite;

		data += towrite;
		len -= towrite;
	}
	
	return true;
}

/*************************************************************************/

void DBuffer::Chunk::clear()
{
	m_first = 0;
	m_next = m_data;
	
	m_size = 0;
	m_remaining = m_maxsize;
}

/*************************************************************************/

bool DBuffer::Chunk::erase(size_t len)
{
	if (!len)
		return true;
	
	if (m_size < len)
		return false;
	
	m_first += len;

	m_size -= len;	
	return true;
}

/*************************************************************************/

bool DBuffer::Chunk::write(char const *data, size_t len)
{
	if (!len)
		return true;
	
	if (len > m_remaining)
		return false;
	
	if (!m_first)
		m_first = m_data;
	
	memcpy(m_next, data, len);
	m_next += len;
	
	m_size += len;
	m_remaining -= len;
	
	return true;
}

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/
