/* Static buffer class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include "sockets/sbuffer.h"

using namespace std;

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

SBuffer::SBuffer(size_t maxsize) : m_buf(new char [maxsize]), m_first(0),
		m_next(m_buf), m_end(m_buf + maxsize), m_size(0), m_maxsize(maxsize)
{
	
}

/*************************************************************************/

SBuffer::SBuffer(SBuffer const &sb) : m_buf(0), m_first(0), m_next(0),
		m_end(0), m_size(sb.m_size), m_maxsize(sb.m_maxsize)
{
	m_buf = new char [m_maxsize];
	memcpy(m_buf, sb.m_buf, m_maxsize);
	
	if (sb.m_first)
		m_first = m_buf + (sb.m_first - sb.m_buf);
	m_next = m_buf + (sb.m_next - sb.m_buf);
	m_end = m_buf + m_maxsize;
}

/*************************************************************************/

SBuffer::~SBuffer()
{
	delete [] m_buf;
}

/*************************************************************************/

SBuffer &SBuffer::operator=(SBuffer const &right)
{
	if (this != &right)
	{
		m_size = right.m_size;
		m_maxsize = right.m_maxsize;
		
		delete [] m_buf;
		m_buf = new char [m_maxsize];
		memcpy(m_buf, right.m_buf, m_maxsize);
	
		if (right.m_first)
			m_first = m_buf + (right.m_first - right.m_buf);
		else
			m_first = 0;
		m_next = m_buf + (right.m_next - right.m_buf);
		m_end = m_buf + m_maxsize;
	}
	
	return *this;
}

/*************************************************************************/

bool SBuffer::clear()
{
	m_first = 0;
	m_next = m_buf;
	m_size = 0;
	
	return true;
}

/*************************************************************************/

bool SBuffer::erase(size_t len)
{
	if (!len)
		return true;
	
	if (len > m_size)
		return false;
	
	if (len == m_size)
		return clear();
	
	if ((size_t)(m_end - m_first) > len)
		m_first += len;
	else
		m_first = m_buf + (len - (m_end - m_first));
	
	m_size -= len;
	return true;
}

/*************************************************************************/

bool SBuffer::read(size_t max, char const **data, size_t *len) const
{
	if (empty())
		return false;
	
	*data = m_first;
	
	if (m_first < m_next)
	{
		if (!max)
			*len = m_size;
		else
			*len = ((max < m_size) ? max : m_size);
		return true;
	}
	
	if (!max)
		*len = m_end - m_first;
	else
		*len = ((max < (size_t)(m_end - m_first)) ? max : (m_end - m_first));
	return true;
}

/*************************************************************************/

bool SBuffer::read_byte(size_type pos, char *ch) const
{
	/* If the offset is greater or equal than the size, then it
	 * is incorrect. */
	if (pos >= m_size)
		return false;
	
	if ((size_t)(m_end - m_first) > pos)
	{
		*ch = m_first[pos];
		return true;
	}
	
	*ch = m_buf[pos - (m_end - m_first)];
	return true;
}

/*************************************************************************/

bool SBuffer::write(char const *data, size_t len)
{
	if (!len)
		return true;
	
	if (len > remaining())
		return false;
	
	if (!m_first || m_next < m_first || ((size_t)(m_end - m_next) >= len))
	{
		if (!m_first)
			m_first = m_next;
		
		memcpy(m_next, data, len);
		m_next += len;
		
		if (m_next == m_end)
			m_next = m_buf;
		
		m_size += len;
		return true;
	}
	
	memcpy(m_next, data, (m_end - m_next));
	memcpy(m_buf, data + (m_end - m_next), len - (m_end - m_next));
	
	m_next = m_buf + (len - (m_end - m_next));

	m_size += len;
	return true;
}

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/
