/* ChanServ header file.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#ifndef CHANSERV_H
#define CHANSERV_H

/*************************************************************************/

#include <string>

#include <databases/collection.h>
#include <databases/record.h>
#include <log/levels.h>
#include <misc/string.h>

#include "tweak.h"

#include "botserv.h"
#include "services.h"

/*************************************************************************/

namespace modules { class Module; }
namespace password { class Encrypter; }

class Access;
class Akick;

/*************************************************************************/

class ChannelInfo : public databases::NameRecord
{
	public:
		/* Classes */
		
		/// Mode lock data class
		/** This class is used to store and handle the (hidden)
		  * mlock data. */
		class MLockData
		{
			public:
				/// Constructor
				MLockData() : m_on(0), m_off(0), m_limit(0),
						m_key()
#ifdef HAS_FMODE
						, m_flood()
#endif
#ifdef HAS_JMODE
						, m_joinrate()
#endif
#ifdef HAS_LMODE
						, m_redirect()
#endif
				{ }
				
				/// Copy constructor
				MLockData(MLockData const &mld) : m_on(mld.m_on),
						m_off(mld.m_off), m_limit(mld.m_limit),
						m_key(mld.m_key)
#ifdef HAS_FMODE
						, m_flood(mld.m_flood)
#endif
#ifdef HAS_JMODE
						, m_joinrate(mld.m_joinrate)
#endif
#ifdef HAS_LMODE
						, m_redirect(mld.m_redirect)
#endif
				{ }
				
				/// Assignment operator
				MLockData &operator=(MLockData const &right);

				/// == operator
				bool operator==(MLockData const &right) const;
				/// != operator
				bool operator!=(MLockData const &right) const
						{ return (!(*this == right)); }

				/// Swaps two MLockData objects.
				void swap(MLockData &mld);
				
				/// Compute modes
				char const *compute() const;
	
				/// Returns the always-on modes
				uint32_t on() const { return m_on; }
				/// Adds always-on modes
				uint32_t on_add(uint32_t on);
				/// Clears always-on modes
				uint32_t on_clear();
				/// Removes always-on modes
				uint32_t on_remove(uint32_t on);
				
				/// Returns the always-off modes
				uint32_t off() const { return m_off; }
				/// Adds always-off modes
				uint32_t off_add(uint32_t off);
				/// Clears always-off modes
				uint32_t off_clear();
				/// Removes always-off modes
				uint32_t off_remove(uint32_t off);
				
				/// Returns the limit
				int32_t limit() const { return m_limit; }
				/// Sets the limit
				int32_t limit(int32_t limit);
				
				/// Returns the key
				char const *key() const { return m_key; }
				/// Sets the key
				char const *key(char const *key);
		
#ifdef HAS_FMODE
				/// Returns the mode +f parameter
				char const *flood() const { return m_flood; }
				/// Sets the mode +f parameter
				char const *flood(char const *flood);
#endif    
		
#ifdef HAS_JMODE
				/// Returns the mode +j parameter
				char const *joinrate() const { return m_joinrate; }
				/// Sets the mode +j parameter
				char const *joinrate(char const *joinrate);
#endif    
		
#ifdef HAS_LMODE
				/// Returns the mode +L parameter
				char const *redirect() const { return m_redirect; }
				/// Sets the mode +L parameter
				char const *redirect(char const *redirect);
#endif
			private:
				uint32_t m_on, m_off;		/* Channel modes */
				int32_t m_limit;			/* 0 if no limit */
				misc::cstring m_key;		/* NULL if no key */
#ifdef HAS_FMODE
				misc::cstring m_flood;		/* NULL if no +f */
#endif
#ifdef HAS_JMODE
				misc::cstring m_joinrate;	/* NULL if no +j */
#endif
#ifdef HAS_LMODE
				misc::cstring m_redirect;	/* NULL if no +L */
#endif
		};
		
		/* Typedefs */
		
		typedef containers::VectorSP<databases::Record *, uint16_t>
				levels_type;
		typedef containers::VectorSP<databases::Record *, uint16_t>
				accesses_type;
		typedef containers::VectorSP<databases::Record *, uint16_t>
				akicks_type;
		
		typedef containers::VectorSP<databases::Record *, uint16_t>
				kbb_type;
		typedef containers::VectorSP<databases::Record *, uint16_t>
				badwords_type;
		
		/* Data */
		
		static databases::Field *f_name;
		static databases::Field *f_owner;
		static databases::Field *f_successor;
		static databases::Field *f_created;
		static databases::Field *f_last_used;
		static databases::Field *f_enc;
		static databases::Field *f_pass;
		static databases::Field *f_flags;
		static databases::Field *f_desc;
		static databases::Field *f_email;
		static databases::Field *f_url;
		static databases::Field *f_entry_message;
		static databases::Field *f_bantype;
		static databases::Field *f_mlock_ircd;
		static databases::Field *f_mlock_modes;
		static databases::Field *f_topic;
		static databases::Field *f_topic_setter;
		static databases::Field *f_topic_time;
		static databases::Field *f_bot;
		static databases::Field *f_botflags;
		static databases::Field *f_caps_min;
		static databases::Field *f_caps_percent;
		static databases::Field *f_flood_lines;
		static databases::Field *f_flood_secs;
		static databases::Field *f_repeats;
	
		Channel *c;							/* Pointer to channel record (if
						 					 * channel is currently in use) */
		
		/* Constants */
		
		// Maximum lengths
		
		/// Maximum length of a channel name.
		static size_t const name_len = 63;
		/// Maximum length of a password.
		static size_t const pass_len = 192;
		/// Real maximum length of a password.
		static size_t const pass_rlen = 31;
		/// Maximum length of the description.
		static size_t const desc_len = 255;
		/// Maximum length of the e-mail address.
		static size_t const email_len = 127;
		/// Maximum length of the URL.
		static size_t const url_len = 255;
		/// Maximum length of the entry message.
		static size_t const entry_message_len = 410;
		/// Maximum length of the mlock modes.
		static size_t const mlock_modes_len = 450;
		/// Maximum length of the topic.
		static size_t const topic_len = 450;
		/// Maximum length of the topic setter.
		static size_t const topic_setter_len = 63;
		
		/// Custom database status
		enum
		{
			/// The IRCd has changed
			dbsIRCdChange = 0x8000,
			/// The topic has changed
			dbsTopicChange = 0x4000
		};
		
		// Channel flags
		
		/// Retain topic even after last person leaves channel
		static uint32_t const cfKeepTopic      = 0x00000001;
		/// Don't allow non-authorized users to be opped
		static uint32_t const cfSecureOps      = 0x00000002;
		/// Hide channel from ChanServ LIST command
		static uint32_t const cfPrivate        = 0x00000004;
		/// Topic can only be changed by ChanServ TOPIC command
		static uint32_t const cfTopicLock      = 0x00000008;
		/// Those not allowed ops are kickbanned
		static uint32_t const cfRestricted     = 0x00000010;
		/// Prevent ChanServ and BotServ commands from harming higher levels
		static uint32_t const cfPeace          = 0x00000020;
		/// Don't allow any privileges unless a user is IDENTIFY'd with NickServ
		static uint32_t const cfSecure         = 0x00000040;
		/// Channel does not expire
		static uint32_t const cfNoExpire       = 0x00000200;
		/// Send notice to channel on use of OP/DEOP
		static uint32_t const cfOpNotice       = 0x00000800;
		/// Stricter control of channel founder status
		static uint32_t const cfSecureFounder  = 0x00001000;
		/// Always sign kicks
		static uint32_t const cfSignKick       = 0x00002000;
		/// Sign kicks if level is < than the one defined by the SIGNKICK level
		static uint32_t const cfSignKickLevel  = 0x00004000;
		/// Pending expiration
		static uint32_t const cfPendingExpire  = 0x00010000;
		
		/// Set the flags to the conf default flags
		static uint32_t const cfDefaults       = 0x40000000;
		
		/// ChanServ is on the channel (temporary)
		static uint32_t const cfInhabit        = 0x80000000;
		/// Temporary flags
		static uint32_t const cfTemporary      = (cfInhabit);
		
		/// Unused flags
		static uint32_t const cfUnused         = 0x3FFE8580;
		/// Default flags (for config)
		static uint32_t const cfDefaultConf    =
				(cfKeepTopic | cfPeace | cfSecure | cfSecureFounder |
				 cfSignKick);
		
		/// Bot flags
		
		/// Never kick operators
		static uint32_t const bfDontKickOps    = 0x00000001;
		/// Never kick voices
		static uint32_t const bfDontKickVoices = 0x00000002;
		/// Enable public channel commands
		static uint32_t const bfFantasy        = 0x00000004;
		/// ChanServ operations made by the bot
		static uint32_t const bfSymbiosis      = 0x00000008;
		/// The bot greets newcomers
		static uint32_t const bfGreet          = 0x00000010;
		/// No bot allowed on the channel
		static uint32_t const bfNoBot          = 0x00000020;
		
		/// Set the flags to the conf default flags
		static uint32_t const bfDefaults       = 0x00000040;
		
		/// Kick bold characters
		static uint32_t const bfKickBolds      = 0x80000000;
		/// Kick colors
		static uint32_t const bfKickColors     = 0x40000000;
		/// Kick reverses
		static uint32_t const bfKickReverses   = 0x20000000;
		/// Kick underlines
		static uint32_t const bfKickUnderlines = 0x10000000;
		/// Kick bad words
		static uint32_t const bfKickBadWords   = 0x08000000;
		/// Kick caps
		static uint32_t const bfKickCaps       = 0x04000000;
		/// Kick flood
		static uint32_t const bfKickFlood      = 0x02000000;
		/// Kick repeats
		static uint32_t const bfKickRepeats    = 0x01000000;
		
		/// Unused flags
		static uint32_t const bfUnused         = 0x00FFFF80;
		/// Default flags (for config)
		static uint32_t const bfDefaultConf    = (bfDontKickOps |
				bfFantasy | bfSymbiosis);
		
		/* Functions */
    
    	/// Constructor.
    	ChannelInfo();
    	/// Constructor setting identifier.
		ChannelInfo(identifier_type id);
		/// Destructor.
		~ChannelInfo();
		
		/// Overrides NameRecord::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides NameRecord::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides NameRecord::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides NameRecord::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides NameRecord::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides NameRecord::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		/// Overrides NameRecord::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides NameRecord::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Overrides Record::end_update().
		virtual bool end_update();
		
		/// Transfers channel ownership to the successor.
		/** Transfers channel ownership to the successor,
		  * provided there is one set and it doesn't have too many
		  * registered channels.
		  *
		  * \return true if successful, false otherwise. */
		bool transfer();
		
		/// Changes the password.
		bool change_password(char const *pass);
		/// Checks the password.
		bool check_password(char const *pass);
		/// Generates a new password.
		bool generate_password(char *passbuf, size_t size);
		
		/// Changes the topic.
		bool change_topic(char const *topic, char const *topic_setter,
				time_t topic_time);
		/// Restores saved topic.
		void restore_topic() const;
		/// Checks for topic lock.
		/** Checks for topic lock. Called on topic change.
		  *
		  * \return true if the previous topic has been restored,
		  * false otherwise. */
		bool check_topiclock() const;
		
		/// Resets levels to the default values.
		bool reset_levels();
		
		/// Changes the caps settings.
		bool change_caps(uint8_t caps_min, uint8_t caps_percent);
		/// Changes the flood settings.
		bool change_flood(uint8_t flood_lines, uint8_t flood_secs);
		
		/// Called after the channel is added to collection.
		void post_registered();
		/// Called after the channel is removed from collection.
		void post_unregistered();
		
		/// Clears the successor field (called by channels.remove_owned_list()).
		void clear_successor();
		/// Clears the bot field (called by channels.remove_owned_list()).
		void clear_bot();
		
		/// Overrides Record::owned().
		virtual bool owned(databases::Collection const *coll,
				databases::Record const *record);
		
		/// Overrides NameRecord::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Implements NameRecord::name().
		virtual char const *name() const { return m_name; }
		/// Sets the channel name.
		char const *name(char const *name);
		
		/// Returns the owner of the channel.
		NickGroup *owner() const { return m_owner; }
		/// Sets the owner of the channel.
		NickGroup *owner(NickGroup *owner);
		/// Sets the owner ID of the channel (database only).
		identifier_type owner_id(identifier_type owner_id);
		
		/// Returns the successor of the channel.
		NickGroup *successor() const { return m_successor; }
		/// Sets the successor of the channel.
		NickGroup *successor(NickGroup *successor);
		/// Sets the successor ID of the channel (database only).
		identifier_type successor_id(identifier_type successor_id);
		
		/// Returns the time of the channel creation.
		time_t created() const { return m_created; }
		/// Sets the time of the channel creation.
		time_t created(time_t created);
		
		/// Returns the last time the channel was used.
		time_t last_used() const { return m_last_used; }
		/// Sets the last time the channel was used.
		time_t last_used(time_t last_used);
		
		/// Returns the password encoding.
		char const *enc() const;
		/// Sets the password encoding.
		char const *enc(char const *enc);
		
		/// Returns the password.
		char const *pass() const { return m_pass; }
		/// Sets the password.
		char const *pass(char const *pass);
		
		/// Returns the channel flags.
		uint32_t flags() const { return m_flags; }
		/// Sets the channel flags (database modules only!)
		uint32_t flags(uint32_t flags);
		/// Adds flags.
		uint32_t flags_add(uint32_t flags);
		/// Removes flags.
		uint32_t flags_remove(uint32_t flags);
		
		/// Returns the channel description
		char const *desc() const { return m_desc; }
		/// Sets the channel description
		char const *desc(char const *desc);
		
		/// Returns the e-mail address associated with the channel
		char const *email() const { return m_email; }
		/// Sets the e-mail address associated with the channel
		char const *email(char const *email);
		
		/// Returns the URL associated with the channel
		char const *url() const { return m_url; }
		/// Sets the URL associated with the channel
		char const *url(char const *url);
		
		/// Returns the notice message sent on JOIN
		char const *entry_message() const { return m_entry_message; }
		/// Sets the notice message sent on JOIN
		char const *entry_message(char const *entry_message);
		
		/// Returns the channel ban type
		uint8_t bantype() const { return m_bantype; }
		/// Sets the channel ban type
		uint8_t bantype(uint8_t bantype);
		
		/// Returns the mlock IRCd
		char const *mlock_ircd() const { return IRCD_NAME; }
		/// Sets the mlock IRCd
		char const *mlock_ircd(char const *mlock_ircd);
		
		/// Returns the mlocked modes
		char const *mlock_modes() const;
		/// Sets the mlocked modes
		char const *mlock_modes(char const *mlock_modes, User *u = 0);
		
		/// Returns the mlock data
		MLockData &mlock_data() { return m_mlock_data; }
		/// Returns the mlock data (const version)
		MLockData const &mlock_data() const { return m_mlock_data; }
		
		uint32_t mlock_on() const { return m_mlock_data.on(); }
		uint32_t mlock_on_add(uint32_t mlock_on)
				{ return m_mlock_data.on_add(mlock_on); }
		uint32_t mlock_on_clear()
				{ return m_mlock_data.on_clear(); }
		uint32_t mlock_on_remove(uint32_t mlock_on)
				{ return m_mlock_data.on_remove(mlock_on); }
		
		uint32_t mlock_off() const { return m_mlock_data.off(); }
		uint32_t mlock_off_add(uint32_t mlock_off)
				{ return m_mlock_data.off_add(mlock_off); }
		uint32_t mlock_off_clear()
				{ return m_mlock_data.off_clear(); }
		uint32_t mlock_off_remove(uint32_t mlock_off)
				{ return m_mlock_data.off_remove(mlock_off); }
		
		int32_t mlock_limit() const { return m_mlock_data.limit(); }
		int32_t mlock_limit(int32_t mlock_limit)
				{ return m_mlock_data.limit(mlock_limit); }
		
		char const *mlock_key() const { return m_mlock_data.key(); }
		char const *mlock_key(char const *mlock_key)
				{ return m_mlock_data.key(mlock_key); }

#ifdef HAS_FMODE
		char const *mlock_flood() const { return m_mlock_data.flood(); }
		char const *mlock_flood(char const *mlock_flood)
				{ return m_mlock_data.flood(mlock_flood); }
#endif    

#ifdef HAS_JMODE
		char const *mlock_joinrate() const { return m_mlock_data.joinrate(); }
		char const *mlock_joinrate(char const *mlock_joinrate)
				{ return m_mlock_data.joinrate(mlock_joinrate); }
#endif    

#ifdef HAS_LMODE
		char const *mlock_redirect() const { return m_mlock_data.redirect(); }
		char const *mlock_redirect(char const *mlock_redirect)
				{ return m_mlock_data.redirect(mlock_redirect); }
#endif
		/// Returns the last topic.
		char const *topic() const { return m_topic; }
		/// Sets the last topic.
		char const *topic(char const *topic);
		
		/// Returns the last topic setter.
		char const *topic_setter() const { return m_topic_setter; }
		/// Sets the last topic setter.
		char const *topic_setter(char const *topic_setter);
		
		/// Returns the last topic time.
		time_t topic_time() const { return m_topic_time; }
		/// Sets the last topic time.
		time_t topic_time(time_t topic_time);
		
		/// Returns the levels.
		levels_type &levels() { return m_levels; }
		/// Returns the level for the given index.
		int16_t levels(uint16_t index) const;
		/// Sets the level for the given index.
		int16_t levels(uint16_t index, int16_t level);
		
		/// Returns the access list.
		accesses_type &accesses() { return m_accesses; }
		
		/// Returns the akick list.
		akicks_type &akicks() { return m_akicks; }

		/* For BotServ */
		
		/// Returns the bot assigned to the channel.
		BotInfo *bi() const { return m_bi; }
		/// Sets the bot assigned to the channel.
		BotInfo *bi(BotInfo *bi);
		/// Sets the identifier of the bot assigned to the channel (db only).
		identifier_type bot_id(identifier_type bot_id);
		
		/// Returns the bot flags.
		uint32_t botflags() const { return m_botflags; }
		/// Sets the bot flags (database only).
		uint32_t botflags(uint32_t botflags);
		/// Adds flags to the bot flags.
		uint32_t botflags_add(uint32_t botflags);
		/// Removes flags from the bot flags.
		uint32_t botflags_remove(uint32_t botflags);
		
		/// Returns the minimum number of caps required to trigger the kicker.
		uint8_t caps_min() const { return m_caps_min; }
		/// Sets the minimum number of caps required to trigger the kicker.
		uint8_t caps_min(uint8_t caps_min);
		
		/// Returns the min. percentage of caps required to trigger the kicker.
		uint8_t caps_percent() const { return m_caps_percent; }
		/// Sets the min. percentage of caps required to trigger the kicker.
		uint8_t caps_percent(uint8_t caps_percent);
		
		/// Returns the minimum number of lines required to trigger the kicker.
		uint8_t flood_lines() const { return m_flood_lines; }
		/// Sets the minimum number of lines required to trigger the kicker.
		uint8_t flood_lines(uint8_t flood_lines);
		
		/// Returns the number of seconds in which the minimum number of lines is required.
		uint8_t flood_secs() const { return m_flood_secs; }
		/// Sets the number of seconds in which the minimum number of lines is required.
		uint8_t flood_secs(uint8_t flood_secs);
		
		/// Returns the minimum number of repeats required to trigger the kicker.
		uint8_t repeats() const { return m_repeats; }
		/// Sets the minimum number of repeats required to trigger the kicker.
		uint8_t repeats(uint8_t repeats);
		
		/// Returns the bad words list.
		badwords_type &badwords() { return m_badwords; }
		
		/// Returns the list of KBB
		kbb_type &kbb() { return m_kbb; }
		/// Returns the KBB for the given index.
		uint8_t kbb(uint8_t index) const;
		/// Sets the number of KBB for the given index.
		uint8_t kbb(uint8_t index, uint8_t count);
	private:
		misc::cstring m_name;
		
		NickGroup *m_owner;
		NickGroup *m_successor;
		
		time_t m_created;
		time_t m_last_used;
		
		password::Encrypter *m_enc;
		misc::cstring m_pass;
		
		uint32_t m_flags;
		
		misc::cstring m_desc;
		misc::cstring m_email;
		misc::cstring m_url;
		
		misc::cstring m_entry_message;
		
		uint8_t m_bantype;
		
		mutable misc::cstring m_mlock_modes;
		MLockData m_mlock_data;
		
		misc::cstring m_topic;
		misc::cstring m_topic_setter;
		time_t m_topic_time;
		
		/* Unsaved data */
		
		levels_type m_levels;
		
		accesses_type m_accesses;
		akicks_type m_akicks;
	
		/* For BotServ */

		BotInfo *m_bi;
		uint32_t m_botflags;
		
		uint8_t m_caps_min, m_caps_percent;
		uint8_t m_flood_lines, m_flood_secs;
		uint8_t m_repeats;
		
		/* Unsaved data */
		
		kbb_type m_kbb;
		
		badwords_type m_badwords;
};

/* Don't allow the channel to be registered or used */
#define OLD_CI_VERBOTEN	0x00000080
/* Channel password is encrypted */
#define CI_ENCRYPTEDPW	0x00000100
/* Channel memo limit may not be changed */
#define OLD_CI_MEMO_HARDMAX	0x00000400
/* Use the xOP lists */
#define OLD_CI_XOP 0x00008000

/* Modes that have been removed. */
#ifdef IRC_UNREAL
#define REMOVED_CMODE_H 0x00001000
#endif

/* Indices for KBB (Kicks Before Ban) */
#define KBB_BOLDS 		0
#define KBB_COLORS 		1
#define KBB_REVERSES	2
#define KBB_UNDERLINES  3
#define KBB_BADWORDS	4
#define KBB_CAPS		5
#define KBB_FLOOD		6
#define KBB_REPEAT		7

/*************************************************************************/

class Channels : public databases::NameCollection
{
	public:
		/// Constructor.
		Channels();
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides NameCollection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		/// Overrides NameCollection::remove().
		virtual bool remove(databases::Record *record, int *err = 0);
		/// Overrides NameCollection::remove_owned().
		virtual bool remove_owned(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		/// Overrides NameCollection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Removes expired channels.
		void expire();
		
		/// Finds a record by ID.
		virtual ChannelInfo *ifind(databases::Record::identifier_type id)
				{ return (ChannelInfo *)NameCollection::find(id); }
		/// Finds a record by name.
		virtual ChannelInfo *find(char const *name) const
				{ return (ChannelInfo *)NameCollection::nfind(name); }
		
		/// Configures owner collection.
		void setup_owner();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Loads a password encrypter.
		password::Encrypter *load_encrypter(char const *name);
		/// Decrements the use of the password encrypter by one.
		void release_encrypter(password::Encrypter *enc);
		
		/// Overrides NameCollection::owner_field().
		virtual databases::Field const
				*owner_field(Collection const *coll) const;
		
		/// Returns the default password encrypter.
		password::Encrypter *encrypter() { return m_encrypter; }
		/// Sets the default password encrypter.
		password::Encrypter *encrypter(password::Encrypter *enc)
				{ return (m_encrypter = enc); }
	protected:
		/// Overrides NameCollection::remove_owned_db().
		virtual bool remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		
		/// Overrides NameCollection::remove_owned_list().
		virtual bool remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
	private:
		password::Encrypter *m_encrypter;
		
		Channels(Channels const &);
		Channels &operator=(Channels const &);
};

/*************************************************************************/

extern Channels channels;

/*************************************************************************/

/* Access levels for users. */

class Access : public databases::NameRecord
{	
	public:
		/* Data */
		static databases::Field *f_owner;
		static databases::Field *f_group;
		static databases::Field *f_level;
		static databases::Field *f_last_seen;
		
		/* Functions */
		
		/// Constructor.
		Access();
		/// Constructor setting the record ID.
		Access(identifier_type id);
		
		/// Overrides Record::get_i().
		virtual int32_t get_i(databases::Field const *field) const;
		/// Overrides Record::set_i().
		virtual int32_t set_i(databases::Field const *field, int32_t value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		/// Overrides Record::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides Record::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Implements NameRecord::name().
		virtual char const *name() const;
		
		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the owner channel of the access.
		ChannelInfo *owner() const { return m_owner; }
		/// Sets the owner channel of the access.
		ChannelInfo *owner(ChannelInfo *owner);
		/// Sets the owner channel ID of the access (for databases only!)
		identifier_type owner_id(identifier_type owner_id);
		
		/// Returns the group of the person getting the access
		NickGroup *ng() const { return m_ng; }
		/// Sets the group of the person getting the access
		NickGroup *ng(NickGroup *ng);
		/// Sets the group ID of the person getting the access (for databases only!)
		identifier_type ng_id(identifier_type owner_id);
		
		/// Returns the access level of the group on the channel.
		int16_t level() const { return m_level; }
		/// Sets the access level of the group on the channel.
		int16_t level(int16_t level);
		
		/// Returns the last time the group was present on the channel.
		time_t last_seen() const { return m_last_seen; }
		/// Sets the last time the group was present on the channel.
		time_t last_seen(time_t last_seen);
	private:
		ChannelInfo *m_owner;
		
		NickGroup *m_ng;
		int16_t m_level;
		time_t m_last_seen;
};

/* Note that these two levels also serve as exclusive boundaries for valid
 * access levels.  ACCESS_FOUNDER may be assumed to be strictly greater
 * than any valid access level, and ACCESS_INVALID may be assumed to be
 * strictly less than any valid access level.
 */
#define ACCESS_FOUNDER	10000	/* Numeric level indicating founder access */
#define ACCESS_INVALID	-10000	/* Used in levels[] for disabled settings */

/*************************************************************************/

class Accesses : public databases::Collection
{
	public:
		/* Consts */
		
		/// Custom error codes
		enum
		{
			/// Access entry with the same name already exists.
			ceName = ceCustom + 0
		};
		
		/* Functions */
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides Collection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		
		/// Overrides Collection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Configures owner collection.
		void setup_owner();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Overrides Collection::owner_field().
		virtual databases::Field const
				*owner_field(Collection const *coll) const;
	protected:
		/// Overrides Collection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides Collection::remove_list().
		virtual void remove_list(databases::Record *record);
		
		/// Overrides Collection::remove_owned_db().
		virtual bool remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		/// Overrides Collection::remove_owned_list().
		virtual bool remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
};

/*************************************************************************/

extern Accesses accesses;

/*************************************************************************/

/* Akick data. */

class Akick : public databases::NameRecord
{
	public:
		/* Data */
		static databases::Field *f_owner;
		static databases::Field *f_flags;
		static databases::Field *f_mask;
		static databases::Field *f_group;
		static databases::Field *f_creator;
		static databases::Field *f_reason;
		static databases::Field *f_since;
		static databases::Field *f_expires;
		
		/* Constants */
		
		/// Maximum length of the mask field.
		static size_t const mask_len = 191;
		/// Maximum length of the creator field.
		static size_t const creator_len = 63;
		/// Maximum length of the reason field.
		static size_t const reason_len = 255;
		
		// Akick flags
		
		/// The akick is for a group (the group field must be non-zero).
		static uint8_t const afIsGroup = 0x02;
		/// The akick is stuck.
		static uint8_t const afStuck = 0x04;
		
		/// Unused flags.
		static uint8_t const afUnused = 0xF9;
		
		/* Functions */
		
		/// Constructor.
		Akick();
		/// Constructor setting the record ID.
		Akick(identifier_type id);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides Record::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		/// Overrides Record::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides Record::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Overrides Record::end_update().
		virtual bool end_update();
		
		/// Places the ban on the channel, if channel exists.
		void stick();
		
		/// Implements NameRecord::name().
		virtual char const *name() const;
		
		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the owner channel of the access.
		ChannelInfo *owner() const { return m_owner; }
		/// Sets the owner channel of the access.
		ChannelInfo *owner(ChannelInfo *owner);
		/// Sets the owner channel ID of the access (for databases only!)
		identifier_type owner_id(identifier_type owner_id);
		
		/// Returns akick flags.
		uint8_t flags() const { return m_flags; }
		/// Sets akick flags (for databases only!)
		uint8_t flags(uint8_t flags);
		/// Adds akick flags.
		uint8_t flags_add(uint8_t flags);
		/// Removes akick flags.
		uint8_t flags_remove(uint8_t flags);
		
		/// Returns banned mask (if afIsGroup flag is NOT set).
		char const *mask() const { return m_mask; }
		/// Sets banned mask.
		char const *mask(char const *mask);
		
		/// Returns banned group (if afIsGroup flag is set).
		NickGroup *ng() const { return m_ng; }
		/// Sets banned group.
		NickGroup *ng(NickGroup *ng);
		/// Sets banned group (for databases only!)
		identifier_type ng_id(identifier_type ng_id);
		
		/// Returns the creator of the akick (may be NULL for histo. reasons)
		char const *creator() const { return m_creator; }
		/// Sets the creator of the akick.
		char const *creator(char const *creator);
		
		/// Returns the reason of the akick (may be NULL).
		char const *reason() const { return m_reason; }
		/// Sets the reason of the akick.
		char const *reason(char const *reason);
		
		/// Returns the creation date of the akick (0 means "unknown" for histo. reasons).
		time_t since() const { return m_since; }
		/// Sets the creation date of the akick.
		time_t since(time_t since);
		
		/// Returns the expiration time of the akick (0 means never).
		time_t expires() const { return m_expires; }
		/// Sets the expiration time of the akick.
		time_t expires(time_t expires);
	private:
		ChannelInfo *m_owner;
		
		uint8_t m_flags;
		
		misc::cstring m_mask;		/* Guaranteed to be non-NULL if in use, NULL if not */
		NickGroup *m_ng;			/* Same */
		
		misc::cstring m_creator;
		misc::cstring m_reason;
		
		time_t m_since;
		time_t m_expires;
};

#define AK_USED 0x0001

/*************************************************************************/

class Akicks : public databases::Collection
{
	public:
		/* Consts */
		
		/// Custom error codes
		enum
		{
			/// Access entry with the same name already exists.
			ceName = ceCustom + 0
		};
		
		/* Functions */
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides Collection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		
		/// Overrides Collection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Configures owner collection.
		void setup_owner();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Overrides Collection::owner_field().
		virtual databases::Field const
				*owner_field(Collection const *coll) const;
	protected:
		/// Overrides Collection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides Collection::remove_list().
		virtual void remove_list(databases::Record *record);
		
		/// Overrides Collection::remove_owned_db().
		virtual bool remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		/// Overrides Collection::remove_owned_list().
		virtual bool remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
};

/*************************************************************************/

extern Akicks akicks;

/*************************************************************************/

class Level : public databases::Record
{
	public:
		/* Data */
		static databases::Field *f_owner;
		static databases::Field *f_name;
		static databases::Field *f_level;
		
		/* Constants */
		
		/// Maximum length of the name.
		static size_t const name_len = 31;
		
		/* Functions */
		
		/// Constructor.
		Level();
		/// Constructor setting the record ID.
		Level(identifier_type id);
		
		/// Overrides Record::get_i().
		virtual int32_t get_i(databases::Field const *field) const;
		/// Overrides Record::set_i().
		virtual int32_t set_i(databases::Field const *field, int32_t value);
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides Record::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Overrides Record::end_update().
		virtual bool end_update();
		
		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the owner channel of the level.
		ChannelInfo *owner() const { return m_owner; }
		/// Sets the owner channel of the level.
		ChannelInfo *owner(ChannelInfo *owner);
		/// Sets the owner channel ID of the level (for databases only!)
		identifier_type owner_id(identifier_type owner_id);
		
		/// Returns the level index.
		uint16_t index() const { return m_index; }
		/// Sets the level index.
		uint16_t index(uint16_t index);
		
		/// Returns the number of kicks before ban.
		int16_t level() const { return m_level; }
		/// Sets the the number of kicks before ban.
		int16_t level(int16_t level);
	private:
		ChannelInfo *m_owner;
		
		uint16_t m_index;
		int16_t m_level;
		
		char const *name(char const *name);
};

/* Level indexes */

#define CA_INVITE			0
#define CA_AKICK			1
#define CA_SET				2	/* but not FOUNDER or PASSWORD */
#define CA_UNBAN			3
#define CA_AUTOOP			4
#define CA_AUTODEOP			5	/* Maximum, not minimum */
#define CA_AUTOVOICE		6
#define CA_OPDEOP			7	/* ChanServ commands OP and DEOP */
#define CA_ACCESS_LIST		8
#define CA_CLEAR			9
#define CA_NOJOIN			10	/* Maximum */
#define CA_ACCESS_CHANGE 	11
#define CA_MEMO				12
#define CA_ASSIGN       	13  /* BotServ ASSIGN command */
#define CA_BADWORDS     	14  /* BotServ BADWORDS command */
#define CA_NOKICK       	15  /* Not kicked by the bot */
#define CA_FANTASIA   		16
#define CA_SAY				17
#define CA_GREET            18
#define CA_VOICEME			19
#define CA_VOICE			20
#define CA_GETKEY           21
#define CA_AUTOHALFOP		22
#define CA_AUTOPROTECT		23
#define CA_OPDEOPME         24
#define CA_HALFOPME         25
#define CA_HALFOP           26
#define CA_PROTECTME        27
#define CA_PROTECT          28
#define CA_KICKME           29
#define CA_KICK             30
#define CA_SIGNKICK			31
#define CA_BANME            32
#define CA_BAN              33
#define CA_TOPIC            34
#define CA_INFO             35

#define CA_SIZE             36

/*************************************************************************/

class Levels : public databases::Collection
{
	public:
		/* Functions */
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides Collection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		
		/// Overrides Collection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Configures the map.
		void setup_map();
		
		/// Configures owner collection.
		void setup_owner();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Overrides Collection::owner_field().
		virtual databases::Field const
				*owner_field(Collection const *coll) const
				{ return Level::f_owner; }
	protected:
		/// Overrides Collection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides Collection::remove_list().
		virtual void remove_list(databases::Record *record);
		
		/// Overrides Collection::remove_owned_db().
		virtual bool remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		/// Overrides Collection::remove_owned_list().
		virtual bool remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
};

/*************************************************************************/

extern Levels levels;

/*************************************************************************/

/* ChanServ mode utilities commands */

struct CSModeUtil {
	char *name;				/* Name of the ChanServ command */
	char *bsname;			/* Name of the BotServ fantasy command */
	
	char *mode;				/* Mode (ie. +o) */
	int32_t notice;			/* Notice flag (for the damn OPNOTICE) */
	
	int level;				/* Level required to use the command */
	int levelself;			/* Level required to use the command for himself */
};

#define MUT_DEOP   		0
#define MUT_OP			1
#define MUT_DEVOICE		2
#define MUT_VOICE		3
#ifdef HAS_HALFOP
#define MUT_DEHALFOP	4
#define MUT_HALFOP		5
#endif
#ifdef IRC_UNREAL
#define MUT_DEPROTECT   6
#define MUT_PROTECT     7
#endif

/*************************************************************************/

class DefEntry
{
	public:
		/// Constructor.
		DefEntry(int index, char const *name, int defval) :
				m_index(index), m_name(name), m_defval(defval) { }
		/// Copy constructor.
		DefEntry(DefEntry const &de) : m_index(de.m_index), m_name(de.m_name),
				m_defval(de.m_defval) { }
		/// Assignment operator.
		DefEntry &operator=(DefEntry const &right);
		
		/// Returns the index.
		int index() const { return m_index; }
		/// Returns the name.
		char const *name() const { return m_name.c_str(); }
		/// Returns the default value.
		int defval() const { return m_defval; }
	private:
		int m_index;
		std::string m_name;
		int m_defval;
};

/*************************************************************************/

/* For Channel structure */

/* This structure stocks ban data since it must not be removed when
 * user is kicked.
 */

struct BanData {
    BanData *next, *prev;	

    char *mask;				/* Since a nick is unsure and a User structure
    					   	   is unsafe */
    time_t last_use;		/* Since time is the only way to check
    					   	   whether it's still useful */
    int16_t kbb[KBB_SIZE];
};

/* This structure stocks information on every user that will be used by
 * BotServ. */

struct UserData {
    /* Data validity */
    time_t last_use;
	
    /* for flood kicker */
    int16_t lines;
    time_t last_start;
    
    /* for repeat kicker */
    char *lastline;
    int16_t times;
};

/*************************************************************************/

/*extern ChannelInfo *chanlists[256];*/
extern CSModeUtil csmodeutils[];

/*************************************************************************/

LOGLEVELH(InfoChanServ, xlog::LogInfo, "Info::ChanServ", "chanserv: ");
LOGLEVELH(InfoChanServList, InfoChanServ, "Info::ChanServ::List", "chanserv list: ");
LOGLEVELH(InfoChanServRights, InfoChanServ, "Info::ChanServ::Rights", "chanserv rights: ");

LOGLEVELH(ErrorChanServ, xlog::LogError, "Error::ChanServ", "Error: chanserv: ");
LOGLEVELH(ErrorChanServList, ErrorChanServ, "Error::ChanServ::List", "Error: chanserv list: ");

/*************************************************************************/

extern DefEntry def_kbb[KBB_SIZE];

extern void listchans(int count_only, const char *chan);

extern bool chanserv_load();
extern void chanserv_notify(modules::Module &sender, int msg, void *arg);
extern void chanserv_start();

extern void cs_init(void);
extern void chanserv(User *u, char *buf);

extern void check_modes(Channel *c);
#ifdef HAS_HALFOP
extern int check_valid_halfop(User *user, Channel *chan, int servermode);
#endif
extern int check_valid_op(User *user, Channel *chan, int servermode);
extern int check_should_op(User *user, const char *chan);
extern int check_should_voice(User *user, const char *chan);
#ifdef HAS_HALFOP
extern int check_should_halfop(User *user, const char *chan);
#endif 
#ifdef IRC_UNREAL
extern int check_should_owner(User *user, const char *chan);
extern int check_should_protect(User *user, const char *chan); 
#endif
extern int check_kick(User *user, char *chan);

extern int check_access(User *user, ChannelInfo *ci, int what);
extern int is_owner(User *user, ChannelInfo *ci);
extern int get_access(User *user, ChannelInfo *ci);
extern Access *get_access_entry(NickGroup *nc, ChannelInfo *ci);
extern void update_cs_lastseen(User *user, ChannelInfo *ci);
extern int get_idealban(ChannelInfo *ci, User *u, char *ret, int retlen);
extern Akick *is_stuck(ChannelInfo *ci, char const *mask);
extern void stick_all(ChannelInfo *ci);

#ifdef HAS_FMODE
extern char const *cs_get_flood(ChannelInfo::MLockData const *mld);
extern void cs_set_flood(ChannelInfo::MLockData *mld, char *value);
#endif
#ifdef HAS_JMODE
extern char const *cs_get_joinrate(ChannelInfo::MLockData const *mld);
extern void cs_set_joinrate(ChannelInfo::MLockData *mld, char *value);
#endif
extern char const *cs_get_key(ChannelInfo::MLockData const *mld);
extern void cs_set_key(ChannelInfo::MLockData *mld, char *value);
extern char const *cs_get_limit(ChannelInfo::MLockData const *mld);
extern void cs_set_limit(ChannelInfo::MLockData *mld, char *value);
#ifdef HAS_LMODE
extern char const *cs_get_redirect(ChannelInfo::MLockData const *mld);
extern void cs_set_redirect(ChannelInfo::MLockData *mld, char *value);
#endif

/*************************************************************************/

#endif /* CHANSERV_H */
