/* Routines for looking up commands in a *Serv command list.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#include <utility>

#include <core/epona.h>
#include <log/levels.h>
#include <misc/misc.h>

#include "epona.h"
#include "language.h"
#include "operserv.h"

#include "services.h"
#include "commands.h"

using namespace std;

/*************************************************************************/

int ListCmd::clear(User *u, container_type &ctn, char const *sub)
{
	if (!changeable(u, ctn, sub))
	{
		notice_lang(m_service, u, m_msgs[cmDisabled]);
		return crStop;
	}
	
	if (!allow_clear(u, ctn, sub)) 
	{
		notice_lang(m_service, u, PERMISSION_DENIED);
		return crStop;
	}
	
	if (ctn.empty())
	{
		notice_lang(m_service, u, m_msgs[cmEmpty], context(u));
		return crStop;
	}
	
	if (clear_records(u, ctn, sub))
		notice_lang(m_service, u, m_msgs[cmClearPartially], context(u));
	else
		notice_lang(m_service, u, m_msgs[cmClear], context(u));
	
	return crStop;
}

/*************************************************************************/

bool ListCmd::clear_records(User *u, container_type &ctn, char const *sub)
{
	bool dont_clear = false, partial = false;
	container_type::iterator it, end;
	
	for (it = ctn.begin(), end = ctn.end(); it != end; ++it)
	{
		if (!(*it))
			continue;
		
		if (!is_member(u, ctn, sub, *it))
		{
			dont_clear = true;
			continue;
		}
		
		if (!del_record(u, ctn, sub, *it))
			dont_clear = partial = true;
		else
			*it = 0;
	}
	
	if (!dont_clear)
		ctn.clear();
	
	return partial;
}

/*************************************************************************/

int ListCmd::del(User *u, container_type &ctn, char const *sub)
{	
	if (!changeable(u, ctn, sub))
	{
		notice_lang(m_service, u, m_msgs[cmDisabled]);
		return crStop;
	}
	
	if (ctn.empty())
	{
		notice_lang(m_service, u, m_msgs[cmEmpty], context(u));
		return crStop;
	}
	
	char *what = strtok(NULL, "");
	
	for (int i = (what ? ((long)strlen(what) - 1) : -1); i >= 0; --i)
	{
		if (isspace(what[i]))
			what[i] = 0;
		if (!isspace(what[i]))
			break;
	}
	
	if (what && !*what)
		what = 0;
	
	if (!what) {
		syntax_error(m_service, u, m_name, m_msgs[cmSyntaxDel], sub);
	} 
	else if (isdigit(*what) && strspn(what, "0123456789-,") == strlen(what))
	{
		/* what is probably a range. */
		char const *expr = what;
		unsigned long first, last;
		
		int entry_num = 0;
		bool perm = true, rw = true;
		uint16_t count = 0;
		
		while (misc::complex_range(&expr, &first, &last, true))
		{
			if (first == last)
			{
				if (!entry_num)
					entry_num = first;
				else if ((unsigned int)entry_num != first)
					entry_num = -1;
			}
			
			/* We support descending ranges, so we invert them
			 * now! */
			if (first != last && max(first, last) == first)
				swap(first, last);

			for (unsigned int i = (first - 1); i != last; ++i)
			{
				/* If the index is greater than or equal to the size
				 * of the list itself... */
				if (i >= ctn.size())
					break; /* We can't delete anything anymore. */
				
				/* If the record is NULL, step to the next */
				if (ctn[i] == 0)
					continue;
				
				if (!is_member(u, ctn, sub, ctn[i]))
					continue;
				
				if ((perm = allow_del(u, ctn, sub, ctn[i])) && 
						(rw = del_record(u, ctn, sub, ctn[i])))
				{
					ctn[i] = 0; /* NULLify the list */
					++count;
				}
			}
		}
		
		if (!count)
		{
			if (!perm)
				notice_lang(m_service, u, PERMISSION_DENIED);
			else if (!rw)
				notice_lang(m_service, u, m_msgs[cmDisabled]);
			else if (entry_num && entry_num != -1)
				notice_lang(m_service, u, m_msgs[cmNoSuchEntry], entry_num, context(u));
			else
				notice_lang(m_service, u, m_msgs[cmNoMatch], context(u));
		} else if (count == 1) {
			notice_lang(m_service, u, m_msgs[cmDeletedOne], context(u));
		} else {
			notice_lang(m_service, u, m_msgs[cmDeletedSeveral], count, context(u));
		}
	} else if (!m_named) {
		syntax_error(m_service, u, m_name, m_msgs[cmSyntaxDel], sub);
	} else 
	{
		/* We're now going to delete by name! */
		container_type::iterator it, end;
		databases::NameRecord *nr;
		
		for (it = ctn.begin(), end = ctn.end(); it != end; ++it)
		{
			/* Ensure the record is valid */
			if (!(*it))
				continue;
			
			if (!is_member(u, ctn, sub, (*it)))
				continue;
			
			nr = (databases::NameRecord *)*it;
			
			if (*nr == what)
			{
				if (!allow_del(u, ctn, sub, *it)) {
					notice_lang(m_service, u, PERMISSION_DENIED);
				} else if (!del_record(u, ctn, sub, *it)) {
					notice_lang(m_service, u, m_msgs[cmDisabled]);
				} else {
					*it = 0;
					notice_lang(m_service, u, m_msgs[cmDeleted], what, context(u));
				}
				
				break;
			}
		}
		
		if (it == end)
			notice_lang(m_service, u, m_msgs[cmNotFound], what, context(u));
	}
	
	return crStop;
}

/*************************************************************************/

int ListCmd::dispatch(User *u, container_type &ctn, char const *sub)
{
	if (!stricmp(sub, "ADD"))
		add(u, ctn, sub);
	else if (!stricmp(sub, "DEL"))
		del(u, ctn, sub);
	else if (!stricmp(sub, "LIST"))
		list(u, ctn, sub);
	else if (!stricmp(sub, "CLEAR"))
		clear(u, ctn, sub);
	else
		syntax_error(m_service, u, m_name, m_msgs[cmSyntax]);
	
	return crStop;
}

/*************************************************************************/

int ListCmd::execute(User *u)
{
	container_type *ctn = container(u);
	if (!ctn)
		return crStop;
	
	char *sub = strtok(NULL, " ");
	
	if (!sub)
	{
		syntax_error(m_service, u, m_name, m_msgs[cmSyntax]);
		return crStop;
	}
	
	/* Transform the sub-command name in uppercase */
	char *end = sub;
	
	while (*end)
	{
		*end = toupper(*end);
		++end;
	}
	
	if (!allow_dispatch(u, *ctn, sub))
		notice_lang(m_service, u, ACCESS_DENIED);
	else
		return dispatch(u, *ctn, sub);
	
	return crStop;
}

/*************************************************************************/

int ListCmd::list(User *u, container_type &ctn, char const *sub)
{
	if (ctn.empty())
	{
		notice_lang(m_service, u, m_msgs[cmEmpty], context(u));
		return crStop;
	}
	
	char *what = strtok(NULL, "");
	
	for (int i = (what ? ((long)strlen(what) - 1) : -1); i >= 0; --i)
	{
		if (isspace(what[i]))
			what[i] = 0;
		if (!isspace(what[i]))
			break;
	}
	
	if (what && !*what)
		what = 0;
	
	if (!what || (isdigit(*what) && strspn(what, "0123456789-,") == strlen(what)))
	{
		/* what is probably a range. */
		char const *expr = what;
		unsigned long first, last;
		int entry_num = 0, step;
		
		bool perm = true, header = false;
		uint16_t count = 0;
		
		while (misc::complex_range(&expr, &first, &last, true))
		{
			if (first == last)
			{
				if (!entry_num)
					entry_num = first;
				else if ((unsigned int)entry_num != first)
					entry_num = -1;
			}
			
			/* Reset the step */
			step = 1;
			/* Descending ranges requires a different step */
			if (first != last && max(first, last) == first)
				step = -1;
			
			if (first > ctn.size())
				first = ctn.size() + 1;
			if (last > ctn.size())
				last = ctn.size() + 1;
			
			for (unsigned int i = (first - 1); (i + 1) != (last + step); i += step)
			{
				/* If the index is greater than or equal to the size
				 * of the list itself... */
				if (i >= ctn.size())
					continue; /* We can't list this record. */
				
				/* If the record is NULL, step to the next */
				if (ctn[i] == 0)
					continue;
					
				if (!is_member(u, ctn, sub, ctn[i]))
					continue;
				
				if ((perm = allow_list(u, ctn, sub, ctn[i])))
				{
					if (!header)
					{
						notice_lang(m_service, u, m_msgs[cmListHeader], context(u));
						header = true;
					}
					if (list_record(u, ctn, sub, (i + 1), ctn[i], count))
						++count;
				}
			}
			
			/* Stop there to prevent infinite loop ;) */
			if (!what)
				break;
		}
		
		if (!count)
		{
			if (!perm)
				notice_lang(m_service, u, PERMISSION_DENIED);
			else if (!what)
				notice_lang(m_service, u, m_msgs[cmEmpty], context(u));
			else if (entry_num && entry_num != -1)
				notice_lang(m_service, u, m_msgs[cmNoSuchEntry], entry_num, context(u));
			else
				notice_lang(m_service, u, m_msgs[cmNoMatch], context(u));
		} else if (count == 1) {
			notice_lang(m_service, u, m_msgs[cmListedOne]);
		} else {
			notice_lang(m_service, u, m_msgs[cmListedSeveral], count);
		}
	} else if (!m_named) {
		syntax_error(m_service, u, m_name, m_msgs[cmSyntaxList], sub);
	} else 
	{
		/* We're now going to list entries whose name match the given pattern */
		container_type::iterator it, end;
		databases::NameRecord *nr;
		
		bool perm = true, header = false;
		uint16_t num = 0, count = 0;
		
		for (it = ctn.begin(), end = ctn.end(); it != end; ++it)
		{
			++num;
			
			/* Ensure the record is valid */
			if (!(*it))
				continue;
			
			if (!is_member(u, ctn, sub, (*it)))
				continue;
			
			nr = (databases::NameRecord *)*it;
			
			if (!nr->matches(what))
				continue;
				
			if (!allow_list(u, ctn, sub, *it)) 
			{
				perm = true;
				continue;
			}
			
			if (!header)
			{
				notice_lang(m_service, u, m_msgs[cmListHeader], context(u));
				header = true;
			}
			
			if (list_record(u, ctn, sub, num, *it, count))
				++count;
		}
		
		if (!count)
		{
			if (!perm)
				notice_lang(m_service, u, PERMISSION_DENIED);
			else
				notice_lang(m_service, u, m_msgs[cmNoMatch], context(u));
		} else if (count == 1) {
			notice_lang(m_service, u, m_msgs[cmListedOne]);
		} else {
			notice_lang(m_service, u, m_msgs[cmListedSeveral], count);
		}
	}
	
	return crStop;
}

/*************************************************************************/

/* Return the Command corresponding to the given name, or NULL if no such
 * command exists.
 */

Command *lookup_cmd(Command *list, const char *cmd)
{
    Command *c;

    for (c = list; c->name; c++) {
	if (stricmp(c->name, cmd) == 0)
	    return c;
    }
    return NULL;
}

/*************************************************************************/

/* Run the routine for the given command, if it exists and the user has
 * privilege to do so; if not, print an appropriate error message.
 */

void run_cmd(const char *service, User *u, Command *list, const char *cmd)
{
    Command *c = lookup_cmd(list, cmd);
    if (c && c->routine) {
	if ((c->has_priv == NULL) || c->has_priv(u))
	    c->routine(u);
	else
	    notice_lang(service, u, ACCESS_DENIED);
    } else {
    	notice_lang(service, u, UNKNOWN_COMMAND_HELP, cmd, service);
    }
}

/*************************************************************************/

/* Print a help message for the given command. */

void help_cmd(const char *service, User *u, Command *list, const char *cmd)
{
    Command *c = lookup_cmd(list, cmd);

    if (c) {
	const char *p1 = c->help_param1,
	           *p2 = c->help_param2,
	           *p3 = c->help_param3,
	           *p4 = c->help_param4;

	if (c->helpmsg_all >= 0) {
	    notice_help(service, u, c->helpmsg_all, p1, p2, p3, p4);
	}

	if (is_services_root(u)) {
	    if (c->helpmsg_root >= 0)
		notice_help(service, u, c->helpmsg_root, p1, p2, p3, p4);
	    else if (c->helpmsg_all < 0)
		notice_lang(service, u, NO_HELP_AVAILABLE, cmd);
	} else if (is_services_admin(u)) {
	    if (c->helpmsg_admin >= 0)
		notice_help(service, u, c->helpmsg_admin, p1, p2, p3, p4);
	    else if (c->helpmsg_all < 0)
		notice_lang(service, u, NO_HELP_AVAILABLE, cmd);
	} else if (is_services_oper(u)) {
	    if (c->helpmsg_oper >= 0)
		notice_help(service, u, c->helpmsg_oper, p1, p2, p3, p4);
	    else if (c->helpmsg_all < 0)
		notice_lang(service, u, NO_HELP_AVAILABLE, cmd);
	} else {
	    if (c->helpmsg_reg >= 0)
		notice_help(service, u, c->helpmsg_reg, p1, p2, p3, p4);
	    else if (c->helpmsg_all < 0)
		notice_lang(service, u, NO_HELP_AVAILABLE, cmd);
	}

    } else {

	notice_lang(service, u, NO_HELP_AVAILABLE, cmd);

    }
}

/*************************************************************************/
