/* Definitions of IRC message functions and list of messages.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#include <core/directives.h>
#include <core/epona.h>
#include <core/version.h>
#include <log/levels.h>

#include "epona.h"

#include "botserv.h"
#include "chanserv.h"
#include "nickserv.h"
#include "operserv.h"
#include "users.h"

#include "services.h"
#include "messages.h"
#include "language.h"

using namespace std;
using namespace core;

/* List of messages is at the bottom of the file. */

/*************************************************************************/

LOGLEVEL (InfoUserIgnore, InfoUser, "Info::User::Ignore");

/*************************************************************************/
/*************************************************************************/

static void m_nickcoll(char *source, int ac, char **av)
{
    if (ac < 1)
	return;
    
    introduce_user(av[0]);
}

/*************************************************************************/

static void m_ping(char *source, int ac, char **av)
{
    if (ac < 1)
	return;
    send_cmd(epona->cd().services.sname, "PONG %s %s", ac>1 ? av[1] :
    		(char const *)epona->cd().services.sname, av[0]);
}

/*************************************************************************/

static void m_away(char *source, int ac, char **av)
{
    User *u = finduser(source);

    if (u && (ac == 0 || *av[0] == 0))	/* un-away */
	check_memos(u);
}

/*************************************************************************/

#if (defined(IRC_SEQUANA) || defined(IRC_SEQUANA_LIGHT))

static void m_bs(char *source, int ac, char **av)
{   
    User *u;
	time_t starttime, stoptime;	/* When processing started and finished */

    if (ac < 1)
		return;

	u = finduser(source);

	if (!u) {
		epona->log(ErrorUser(), "Record for %s not found", source);
		notice(s_BotServ, source, getstring(NULL, USER_RECORD_NOT_FOUND));
		return;
    }

    /* Check if we should ignore.  Operators always get through. */
    if (allow_ignore && !is_oper(u)) {
		IgnoreData *ign = get_ignore(source);
		if (ign && ign->time > time(NULL)) {
	    	epona->log(InfoUserIgnore(), "Ignored message from %s: \"%s\"",
	    			source, inbuf);
	    	return;
		}
    }

    starttime = time(NULL);
    
    botserv(u, av[0]);

    /* Add to ignore list if the command took a significant amount of time. */
    if (allow_ignore) {
		stoptime = time(NULL);
		if (stoptime > starttime && *source && !strchr(source, '.'))
	    	add_ignore(source, stoptime-starttime);
    }
}

#endif

/*************************************************************************/

#ifdef HAS_CHGHOST

#if (defined(IRC_SEQUANA) || defined(IRC_SEQUANA_LIGHT))
static void m_svshost(char *source, int ac, char **av)
#else
static void m_chghost(char *source, int ac, char **av)
#endif
{
    User *u;
    
    if (ac != 2)
		return;
    
    u = finduser(av[0]);
	if (!u) {
	    epona->log(ErrorUser(), "CHGHOST for nonexistent user %s", av[0]);
	    return;
	}
	
	change_user_host(u, av[1]);
}

#endif

/*************************************************************************/

#if defined(IRC_ULTIMATE) || defined(IRC_UNREAL)

static void m_chgident(char *source, int ac, char **av)
{
	User *u;
    
    if (ac != 2)
		return;
    
    u = finduser(av[0]);
	if (!u) {
	    epona->log(ErrorUser(), "CHGIDENT for nonexistent user %s", av[0]);
	    return;
	}
	
	change_user_username(u, av[1]);
}

/*************************************************************************/

static void m_chgname(char *source, int ac, char **av)
{
	User *u;
    
    if (ac != 2)
		return;
    
    u = finduser(av[0]);
	if (!u) {
	    epona->log(ErrorUser(), "CHGNAME for nonexistent user %s", av[0]);
	    return;
	}
	
	change_user_realname(u, av[1]);
}

#endif

/*************************************************************************/

#ifdef IRC_BAHAMUT

static void m_cs(char *source, int ac, char **av)
{   
    User *u;
	time_t starttime, stoptime;	/* When processing started and finished */

    if (ac < 1)
		return;

	u = finduser(source);

	if (!u) {
		epona->log(ErrorUser(), "Record for %s not found", source);
		notice(s_ChanServ, source, getstring3(USER_RECORD_NOT_FOUND));
		return;
    }

    /* Check if we should ignore.  Operators always get through. */
    if (allow_ignore && !is_oper(u)) {
		IgnoreData *ign = get_ignore(source);
		if (ign && ign->time > time(NULL)) {
	    	epona->log(InfoUserIgnore(), "Ignored message from %s: \"%s\"", source, inbuf);
	    	return;
		}
    }

    starttime = time(NULL);
    
    chanserv(u, av[0]);

    /* Add to ignore list if the command took a significant amount of time. */
    if (allow_ignore) {
		stoptime = time(NULL);
		if (stoptime > starttime && *source && !strchr(source, '.'))
	    	add_ignore(source, stoptime-starttime);
    }
}

#endif

/*************************************************************************/

#ifdef IRC_BAHAMUT

static void m_hs(char *source, int ac, char **av)
{   
    User *u;
	time_t starttime, stoptime;	/* When processing started and finished */

    if (ac < 1)
		return;

	u = finduser(source);

	if (!u) {
		epona->log(ErrorUser(), "Record for %s not found", source);
		notice(s_HelpServ, source, getstring3(USER_RECORD_NOT_FOUND));
		return;
    }

    /* Check if we should ignore.  Operators always get through. */
    if (allow_ignore && !is_oper(u)) {
		IgnoreData *ign = get_ignore(source);
		if (ign && ign->time > time(NULL)) {
	    	epona->log(InfoUserIgnore(), "Ignored message from %s: \"%s\"", source, inbuf);
	    	return;
		}
    }

    starttime = time(NULL);
    
    notice_help(s_HelpServ, u, HELP_HELP, s_NickServ, s_ChanServ, s_MemoServ);
    notice_help(s_HelpServ, u, HELP_HELP_BOT, s_BotServ);

    /* Add to ignore list if the command took a significant amount of time. */
    if (allow_ignore) {
		stoptime = time(NULL);
		if (stoptime > starttime && *source && !strchr(source, '.'))
	    	add_ignore(source, stoptime-starttime);
    }
}

#endif

/*************************************************************************/

static void m_join(char *source, int ac, char **av)
{
    if (ac != 1)
	return;
    do_join(source, ac, av);
}

/*************************************************************************/

static void m_kick(char *source, int ac, char **av)
{
    if (ac != 3)
	return;
    do_kick(source, ac, av);
}

/*************************************************************************/

static void m_kill(char *source, int ac, char **av)
{
    BotInfo *bi;
	
    if (ac != 2)
	return;
    /* Recover if someone kills us. */
    if (stricmp(av[0], s_OperServ) == 0 ||
        stricmp(av[0], s_NickServ) == 0 ||
        stricmp(av[0], s_ChanServ) == 0 ||
        stricmp(av[0], s_MemoServ) == 0 ||
        stricmp(av[0], s_BotServ) == 0 ||
        stricmp(av[0], s_HelpServ) == 0 ||
        stricmp(av[0], s_GlobalNoticer) == 0
    ) {
		introduce_user(av[0]);
    } else if ((bi = bots.find(av[0]))) { 
		introduce_user(av[0]);
		bot_rejoin_all(bi);
    } else {
	do_kill(source, ac, av);
    }
}

/*************************************************************************/

static void m_mode(char *source, int ac, char **av)
{
	if (*av[0] == '#' || *av[0] == '&') {
#ifdef IRC_BAHAMUT
		int i;
		
		if (ac < 3)
			return;
		
		for (i = 2; i < ac; i++)
			av[i-1] = av[i];
		ac--;
#else
		if (ac < 2)
			return;
#endif
		do_cmode(source, ac, av);
    } else {
	if (ac != 2)
	    return;
	do_umode(source, ac, av);
    }
}

/*************************************************************************/

static void m_motd(char *source, int ac, char **av)
{
    FILE *f;
    char buf[BUFSIZE];

    f = fopen(epona->cd().services.motd, "r");
    send_cmd(epona->cd().services.sname, "375 %s :- %s Message of the Day",
		source, (char const *)epona->cd().services.sname);
   if (f) {
	while (fgets(buf, sizeof(buf), f)) {
	    buf[strlen(buf)-1] = 0;
	    send_cmd(epona->cd().services.sname, "372 %s :- %s", source, buf);
	}
	fclose(f);
    } else {
	send_cmd(epona->cd().services.sname, "372 %s :- MOTD file not found!  Please "
			"contact your IRC administrator.", source);
    }
    send_cmd(epona->cd().services.sname, "376 %s :End of /MOTD command.", source);
}

/*************************************************************************/

#ifdef IRC_BAHAMUT

static void m_ms(char *source, int ac, char **av)
{   
    User *u;
	time_t starttime, stoptime;	/* When processing started and finished */

    if (ac < 1)
		return;

	u = finduser(source);

	if (!u) {
		epona->log(ErrorUser(), "Record for %s not found", source);
		notice(s_MemoServ, source, getstring3(USER_RECORD_NOT_FOUND));
		return;
    }

    /* Check if we should ignore.  Operators always get through. */
    if (allow_ignore && !is_oper(u)) {
		IgnoreData *ign = get_ignore(source);
		if (ign && ign->time > time(NULL)) {
	    	epona->log(InfoUserIgnore(), "Ignored message from %s: \"%s\"", source, inbuf);
	    	return;
		}
    }

    starttime = time(NULL);
    
    memoserv(u, av[0]);

    /* Add to ignore list if the command took a significant amount of time. */
    if (allow_ignore) {
		stoptime = time(NULL);
		if (stoptime > starttime && *source && !strchr(source, '.'))
	    	add_ignore(source, stoptime-starttime);
    }
}

#endif

/*************************************************************************/

static void m_nick(char *source, int ac, char **av)
{
	if (ac != 2) {
#if defined(IRC_BAHAMUT)
# if defined(IRC_SEQUANA)
		do_nick(source, av[0], av[4], av[5], av[6], av[10], strtoul(av[2], NULL, 10), strtoul(av[7], NULL, 0), av[3], strtoul(av[8], NULL, 0), av[9]);
# else
		do_nick(source, av[0], av[4], av[5], av[6], av[9], strtoul(av[2], NULL, 10), strtoul(av[7], NULL, 0), av[3], strtoul(av[8], NULL, 0));
# endif
#elif defined(IRC_UNREAL)
		if (ac == 7) {
			/* For some reasons, Unreal sends this sometimes */
			do_nick(source, av[0], av[3], av[4], av[5], av[6], strtoul(av[2], NULL, 10), 0, 0, "*");
		} else {
			do_nick(source, av[0], av[3], av[4], av[5], av[10], strtoul(av[2], NULL, 10), strtoul(av[6], NULL, 0), av[7], av[9], av[8]);
		}
#else
# if defined(IRC_ULTIMATE)
		if (ac == 7) {
			do_nick(source, av[0], av[3], av[4], av[5], av[6], strtoul(av[2], NULL, 10), 0, 0);
		} else {
			do_nick(source, av[0], av[3], av[4], av[5], av[7], strtoul(av[2], NULL, 10), strtoul(av[6], NULL, 0), 0);
		}
# else
		do_nick(source, av[0], av[3], av[4], av[5], av[7], strtoul(av[2], NULL, 10), strtoul(av[6], NULL, 0), 0);
# endif
#endif
	} else {
		do_nick(source, av[0], NULL, NULL, NULL, NULL, strtoul(av[1], NULL, 10), 0, 0);
	}	
}

/*************************************************************************/

#ifdef IRC_BAHAMUT

static void m_ns(char *source, int ac, char **av)
{   
    User *u;
	time_t starttime, stoptime;	/* When processing started and finished */

    if (ac < 1)
		return;

	u = finduser(source);

	if (!u) {
		epona->log(ErrorUser(), "Record for %s not found", source);
		notice(s_NickServ, source, getstring3(USER_RECORD_NOT_FOUND));
		return;
    }

    /* Check if we should ignore.  Operators always get through. */
    if (allow_ignore && !is_oper(u)) {
		IgnoreData *ign = get_ignore(source);
		if (ign && ign->time > time(NULL)) {
	    	epona->log(InfoUserIgnore(), "Ignored message from %s: \"%s\"", source, inbuf);
	    	return;
		}
    }

    starttime = time(NULL);
    
    nickserv(u, av[0]);

    /* Add to ignore list if the command took a significant amount of time. */
    if (allow_ignore) {
		stoptime = time(NULL);
		if (stoptime > starttime && *source && !strchr(source, '.'))
	    	add_ignore(source, stoptime-starttime);
    }
}

#endif

/*************************************************************************/

#ifdef IRC_BAHAMUT

static void m_os(char *source, int ac, char **av)
{   
    User *u;
	time_t starttime, stoptime;	/* When processing started and finished */

    if (ac < 1)
		return;

	u = finduser(source);

	if (!u) {
		epona->log(ErrorUser(), "Record for %s not found", source);
		notice(s_OperServ, source, getstring3(USER_RECORD_NOT_FOUND));
		return;
    }

    /* Check if we should ignore.  Operators always get through. */
    if (allow_ignore && !is_oper(u)) {
		IgnoreData *ign = get_ignore(source);
		if (ign && ign->time > time(NULL)) {
	    	epona->log(InfoUserIgnore(), "Ignored message from %s: \"%s\"", source, inbuf);
	    	return;
		}
    }

    starttime = time(NULL);
    
    if (is_oper(u)) {
	    operserv(u, av[0]);
	} else {
		notice_lang(s_OperServ, u, ACCESS_DENIED);
	    
	    if (epona->cd().operserv.wall & Epona::Directives::OperServ::Wall::waNonOper)
			wallops(s_OperServ, "Denied access to %s from %s (non-oper)", s_OperServ, u->nick);
	}

    /* Add to ignore list if the command took a significant amount of time. */
    if (allow_ignore) {
		stoptime = time(NULL);
		if (stoptime > starttime && *source && !strchr(source, '.'))
	    	add_ignore(source, stoptime-starttime);
    }
}

#endif

/*************************************************************************/

static void m_part(char *source, int ac, char **av)
{
    if (ac < 1 || ac > 2)
	return;
    do_part(source, ac, av);
}

/*************************************************************************/

static void m_privmsg(char *source, int ac, char **av)
{
    char *s;
    time_t starttime, stoptime;	/* When processing started and finished */

    BotInfo *bi;
    ChannelInfo *ci;
    User *u;

    if (ac != 2)
		return;

	u = finduser(source);

	if (!u) {
		epona->log(ErrorUser(), "Record for %s not found", source);
		notice(av[1], source, getstring3(USER_RECORD_NOT_FOUND));
		return;
    }

	if (*av[0] == '#') {
		if ((ci = channels.find(av[0])))
    		if (ci->c && ci->bi()) /* Some paranoia checks */
    			botchanmsgs(u, ci, av[1]);
	} else {
    	/* Check if we should ignore.  Operators always get through. */
    	if (allow_ignore && !is_oper(u)) {
			IgnoreData *ign = get_ignore(source);
			if (ign && ign->time > time(NULL)) {
	    		epona->log(InfoUserIgnore(), "Ignored message from %s: \"%s\"", source, inbuf);
	    		return;
			}
    	}

    	/* If a server is specified (nick@server format), make sure it matches
    	 * us, and strip it off. */
    	s = strchr(av[0], '@');
    	if (s) {
			*s++ = 0;
			if (stricmp(s, epona->cd().services.sname) != 0)
	    		return;
    	}
    
		starttime = time(NULL);

    	if (stricmp(av[0], s_OperServ) == 0) {
			if (is_oper(u)) {
	    		operserv(u, av[1]);
			} else {
				notice_lang(s_OperServ, u, ACCESS_DENIED);
	    
	    		if (epona->cd().operserv.wall & Epona::Directives::OperServ::Wall::waNonOper)
					wallops(s_OperServ, "Denied access to %s from %s (non-oper)", s_OperServ, u->nick);
			}
    	} else if (stricmp(av[0], s_NickServ) == 0) {
			nickserv(u, av[1]);
    	} else if (stricmp(av[0], s_ChanServ) == 0) {
			chanserv(u, av[1]);
    	} else if (stricmp(av[0], s_MemoServ) == 0) {
			memoserv(u, av[1]);
    	} else if (stricmp(av[0], s_BotServ) == 0) {
			botserv(u, av[1]);
    	} else if (stricmp(av[0], s_HelpServ) == 0) {
			notice_help(s_HelpServ, u, HELP_HELP, s_NickServ, s_ChanServ, s_MemoServ);
			notice_help(s_HelpServ, u, HELP_HELP_BOT, s_BotServ);
    	} 
    	else if ((bi = bots.find(av[0]))) {
    		botmsgs(u, bi, av[1]);
    	}

    	/* Add to ignore list if the command took a significant amount of time. */
    	if (allow_ignore) {
			stoptime = time(NULL);
			if (stoptime > starttime && *source && !strchr(source, '.'))
	    		add_ignore(source, stoptime-starttime);
    	}
	}
}

/*************************************************************************/

static void m_quit(char *source, int ac, char **av)
{
    if (ac != 1)
	return;
    do_quit(source, ac, av);
}

/*************************************************************************/

static void m_server(char *source, int ac, char **av)
{	
	if (ac != 3)
		return;
	
	do_server(source, ac, av);
}

/*************************************************************************/

#if defined(IRC_ULTIMATE) || defined(IRC_UNREAL)

static void m_sethost(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
	return;
    
    u = finduser(source);
	if (!u) {
	    epona->log(ErrorUser(), "SETHOST for nonexistent user %s", source);
	    return;
	}
	
	change_user_host(u, av[0]);
}

/*************************************************************************/

static void m_setident(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
		return;
    
    u = finduser(source);
	if (!u) {
	    epona->log(ErrorUser(), "SETIDENT for nonexistent user %s", source);
	    return;
	}
	
	change_user_username(u, av[0]);
}

/*************************************************************************/

static void m_setname(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
	return;
    
    u = finduser(source);
	if (!u) {
	    epona->log(ErrorUser(), "SETNAME for nonexistent user %s", source);
	    return;
	}
	
	change_user_realname(u, av[0]);
}

#endif

/*************************************************************************/

#ifdef IRC_BAHAMUT

static void m_sjoin(char *source, int ac, char **av)
{
	do_sjoin(source, ac, av);
}

#endif

/*************************************************************************/

static void m_squit(char *source, int ac, char **av)
{	
	if (ac != 2)
		return;
	
	do_squit(source, ac, av);
}

/*************************************************************************/

static void m_stats(char *source, int ac, char **av)
{
	User *u;
	
	if (ac < 1)
		return;

	switch (*av[0]) {
		case 'l':
			u = finduser(source);
				
			if (u && is_oper(u)) {

				send_cmd(NULL, "211 %s Server SendBuf SentBytes SentMsgs RecvBuf "
						"RecvBytes RecvMsgs ConnTime", source);
		
				send_cmd(NULL, "211 %s %s %d %d %d %d %d %d %ld", source,
						(server_uplink())->name,
						write_buffer_len(), total_written, -1,
						read_buffer_len(), total_read, -1,
						(long) time(NULL) - epona->start_time());
			}
		
			send_cmd(NULL, "219 %s l :End of /STATS report.", source);
			break;
		case 'o':
		case 'O':
		{
			conf::ConfStrCharA::container_type::const_iterator it, end =
					epona->cd().operserv.root->end();
			
			for (it = epona->cd().operserv.root->begin(); it != end; ++it) 
			{
				send_cmd(NULL, "243 %s O * * %s Root 0", source,
						(char const *) *it);
			}
			
			VectorNG::const_iterator vit, vend;
			
			for (vit = servadmins.begin(), vend = servadmins.end();
					vit != vend; ++vit)
			{
				NickGroup *ng = (NickGroup *)(*vit);
				
				if (!ng)
					continue;
				
				send_cmd(NULL, "243 %s O * * %s Admin 0", source, ng->name());
			}
			
			for (vit = servopers.begin(), vend = servopers.end();
					vit != vend; ++vit)
			{
				NickGroup *ng = (NickGroup *)(*vit);
				
				if (!ng)
					continue;
				
				send_cmd(NULL, "243 %s O * * %s Oper 0", source, ng->name());
			}
					
			send_cmd(NULL, "219 %s %c :End of /STATS report.", source, *av[0]);
			break;
		}
		case 'u': {
			int uptime = time(NULL) - epona->start_time();
			send_cmd(NULL, "242 %s :Server Up %d day%s, %02d:%02d:%02d",
					source, uptime/86400, (uptime/86400 == 1) ? "" : "s",
					(uptime/3600) % 24, (uptime/60) % 60, uptime % 60);
			send_cmd(NULL, "250 %s :Current users: %d (%d ops); maximum %d",
					source, usercnt, opcnt,
					(stats_rec ? stats_rec->maxusers() : 0));
					send_cmd(NULL, "219 %s u :End of /STATS report.", source);
			break;
		} /* case 'u' */

		default:
			send_cmd(NULL, "219 %s %c :End of /STATS report.", source, *av[0]);
			break;
	}
}

/*************************************************************************/

static void m_time(char *source, int ac, char **av)
{
    time_t t;
    struct tm *tm;
    char buf[64];

    time(&t);
    tm = localtime(&t);
    strftime(buf, sizeof(buf), "%a %b %d %H:%M:%S %Y %Z", tm);
    send_cmd(NULL, "391 :%s", buf);
}

/*************************************************************************/

static void m_topic(char *source, int ac, char **av)
{
    if (ac != 4)
	return;
    do_topic(source, ac, av);
}

/*************************************************************************/

void m_version(char *source, int ac, char **av)
{
	if (source)
		send_cmd(epona->cd().services.sname, 
			"351 %s %s-%s %s :%s -- compiled %s %s",
			source, program_name, epona->version_string().c_str(),
			(char const *) epona->cd().services.sname, program_protocol_l,
			__DATE__, __TIME__);
}

/*************************************************************************/

void m_whois(char *source, int ac, char **av)
{
	if (!source || ac < 1)
		return;
	
	BotInfo *bi;
	char const *user, *host, *real;
	
	if (stricmp(av[0], s_NickServ) == 0)
	{
		user = epona->cd().nickserv.service.username;
		host = epona->cd().nickserv.service.hostname;
		real = epona->cd().nickserv.service.realname;
	} else if (stricmp(av[0], s_ChanServ) == 0) {
		user = epona->cd().chanserv.service.username;
		host = epona->cd().chanserv.service.hostname;
		real = epona->cd().chanserv.service.realname;
	} else if (stricmp(av[0], s_MemoServ) == 0) {
		user = epona->cd().memoserv.service.username;
		host = epona->cd().memoserv.service.hostname;
		real = epona->cd().memoserv.service.realname;
	} else if (stricmp(av[0], s_BotServ) == 0) {
		user = epona->cd().botserv.service.username;
		host = epona->cd().botserv.service.hostname;
		real = epona->cd().botserv.service.realname;
	} else if (stricmp(av[0], s_HelpServ) == 0) {
		user = epona->cd().helpserv.username;
		host = epona->cd().helpserv.hostname;
		real = epona->cd().helpserv.realname;
	} else if (stricmp(av[0], s_OperServ) == 0) {
		user = epona->cd().operserv.service.username;
		host = epona->cd().operserv.service.hostname;
		real = epona->cd().operserv.service.realname;
	} else if (stricmp(av[0], s_GlobalNoticer) == 0) {
		user = epona->cd().global.service.username;
		host = epona->cd().global.service.hostname;
		real = epona->cd().global.service.realname;
	} else if ((bi = bots.find(av[0]))) {
	    /* Bots are handled separately */
	    send_cmd(epona->cd().services.sname, "311 %s %s %s %s * :%s", source,
	    		bi->nick(), bi->username(), bi->hostname(), bi->realname());
	    send_cmd(epona->cd().services.sname, "312 %s %s %s :%s", source,
	    		bi->nick(), (const char *)epona->cd().services.sname,
	    		(const char *) epona->cd().services.description);
	    send_cmd(epona->cd().services.sname,
	    		"317 %s %s %ld %ld :seconds idle, signon time",
	    		source, bi->nick(), (long) time(NULL) - bi->lastmsg(),
	    		(long) epona->start_time());
	    send_cmd(epona->cd().services.sname, "318 %s %s :End of /WHOIS list.",
	    		source, bi->nick());
	    return;
	}
	else {
		send_cmd(epona->cd().services.sname, "401 %s %s :No such service.",
				source, av[0]);
		return;
	}
	send_cmd(epona->cd().services.sname, "311 %s %s %s %s * :%s", source,
			av[0], user, host, real);
	send_cmd(epona->cd().services.sname, "312 %s %s %s :%s", source, av[0],
			(const char *)epona->cd().services.sname,
			(const char *) epona->cd().services.description);
	send_cmd(epona->cd().services.sname,
			"317 %s %s %ld %ld :seconds idle, signon time",
			source, av[0], (long) time(NULL) - epona->start_time(),
			(long) epona->start_time());
	send_cmd(epona->cd().services.sname, "318 %s %s :End of /WHOIS list.",
			source, av[0]);
}

/*************************************************************************/
/*************************************************************************/

static Message irc_messages[] = 
{
    { "401",       NULL },
    { "436",       m_nickcoll },
    { "AWAY",      m_away },
    { "INVITE",    NULL },
    { "JOIN",      m_join },
    { "KICK",      m_kick },
    { "KILL",      m_kill },
    { "MODE",      m_mode },
    { "MOTD",      m_motd },
    { "NICK",      m_nick },
    { "NOTICE",    NULL },
    { "PART",      m_part },
    { "PASS",      NULL },
    { "PING",      m_ping },
    { "PRIVMSG",   m_privmsg },
    { "QUIT",      m_quit },
    { "SERVER",    m_server },
    { "SQUIT",     m_squit },
    { "STATS",     m_stats },
    { "TIME",      m_time },
    { "TOPIC",     m_topic },
    { "USER",      NULL },
    { "VERSION",   m_version },
    { "WALLOPS",   NULL },
    { "WHOIS",     m_whois },

    /* DALnet specific messages */
    { "AKILL",     NULL },
    { "GLOBOPS",   NULL },
    { "GNOTICE",   NULL },
    { "GOPER",     NULL },
    { "RAKILL",    NULL },
    { "SILENCE",   NULL },
    { "SVSKILL",   NULL },
    { "SVSMODE",   NULL },
    { "SVSNICK",   NULL },
    { "SVSNOOP",   NULL },
    { "SQLINE",    NULL },
    { "UNSQLINE",  NULL },
    
    /* DreamForge specific messages */
#ifdef IRC_DREAMFORGE
    { "PROTOCTL",  NULL },
#endif
    
    /* Bahamut specific messages */
#ifdef IRC_BAHAMUT
    { "CAPAB", 	   NULL },
    { "CS",        m_cs },
    { "HS",        m_hs },
    { "LUSERSLOCK", NULL },
    { "MS",        m_ms },
    { "NS",        m_ns },
    { "OS",        m_os },
    { "RESYNCH",   NULL },
    { "RS",        NULL },
    { "SGLINE",    NULL },
    { "SJOIN",     m_sjoin },
    { "SS",        NULL },
    { "SVINFO",    NULL },
    { "SZLINE",    NULL },
    { "UNSGLINE",  NULL },
    { "UNSZLINE",  NULL },
#endif

#if (defined(IRC_SEQUANA) || defined(IRC_SEQUANA_LIGHT))
    { "BS",        m_bs },
#endif

#if (defined(IRC_SEQUANA) || defined(IRC_SEQUANA_LIGHT))
    { "SVSHOST",   m_svshost },
#endif

#ifdef IRC_ULTIMATE
    { "CHGHOST",   m_chghost },
    { "CHGIDENT",  m_chgident },
    { "CHGNAME",   m_chgname },
    { "NETINFO",   NULL },
	{ "SETHOST",   m_sethost },
	{ "SETIDENT",  m_setident },
	{ "SETNAME",   m_setname },
	{ "VCTRL",     NULL },
#endif

#ifdef IRC_UNREAL
    { "CHGHOST",   m_chghost },
    { "CHGIDENT",  m_chgident },
    { "CHGNAME",   m_chgname },
	{ "EOS",       NULL },
	{ "NETINFO",   NULL },
	{ "SENDSNO",   NULL },
	{ "SETHOST",   m_sethost },
	{ "SETIDENT",  m_setident },
	{ "SETNAME",   m_setname },
	{ "SMO",       NULL },
	{ "TKL",       NULL },
#endif

    { NULL }

};

/*************************************************************************/

Message *find_message(const char *name)
{
    Message *m;

    for (m = irc_messages; m->name; m++) {
	if (stricmp(name, m->name) == 0)
	    return m;
    }
    return NULL;
}

/*************************************************************************/
