/* Configuration data module.
 *
 * Configuration data module (c) 2006 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <conf/directives.h>
#include <conf/quick.h>
#include <conf/parser.h>
#include <core/version.h>
#include <databases/database.h>
#include <log/levels.h>
#include <modules/core.h>
#include <modules/manager.h>

#include "conf.h"
#include "decl.h"
#include "module.h"

using namespace std;
using namespace modules;
using namespace confdata;

/*************************************************************************/

extern "C" Module *confdata_get(Core &core)
{
	/* Ensure the core version is the same as the version we were
	 * compiled for so that there are no compatibility problems later. */
	if (core.version() != core::core_version)
	{
		core.log(xlog::LogCritical(), "Core module version mismatch");
		return 0;
	}
	
	return (me = new ConfData(core));
}

/*************************************************************************/

extern "C" void confdata_release(Core &core, Module *module)
{
	delete module;
	me = 0;
}

/*************************************************************************/

namespace confdata
{

/*************************************************************************/

ConfData *me;

/*************************************************************************/

ConfData::ConfData(Core &core) : 
		Module("confdata", "Configuration data exporter"),
		m_core(&core), m_cd(new Directives(core.conf_logger()))
{
	
}

/*************************************************************************/

ConfData::~ConfData()
{
	delete m_cd;
}

/*************************************************************************/

bool ConfData::add_confdirs()
{
	conf::ConfGlobalQ global(m_core->conf_logger());
	
	if (!global.setup())
		return false;
	
	conf::ConfQ *parser = m_core->config(*m_core, global);
	if (parser && !add_confdirs(m_core->name().c_str(), parser->conf()))
		return false;
	
	modules::ModManager::list_type::const_iterator it, end;
	
	for (it = m_core->modules().modules().begin(),
			end = m_core->modules().modules().end(); it != end; ++it)
	{
		modules::Module *m = (*it);
		
		parser = m->config(*m_core, global);
		if (parser && !add_confdirs(m->name().c_str(), parser->conf()))
			return false;
	}
	
	return true;
}

/*************************************************************************/

bool ConfData::add_confdirs(char const *name, conf::Conf const &parser)
{
	conf::Conf::DirList::const_iterator it, end;
	
	for (it = parser.directives().begin(), end = parser.directives().end();
			it != end; ++it)
	{
		conf::ConfDir const *cd = it->second;
		
		char dirname[ConfDir::name_len];
		snprintf(dirname, sizeof(dirname), "%s::%s",
				name, cd->name().c_str());
		
		size_t count = cd->label_count();
		
		for (size_t i = 0; i < count; ++i)
		{
			char const *label = cd->label(i);
			
			if (!label)
				continue;
			
			ConfDir *cdi = (ConfDir *)m_confdirs.create();
			if (!cdi)
				return false;
			
			cdi->name(dirname);
			cdi->value(label);
			
			if (!m_confdirs.add(cdi, 0))
				return false;
		}
		
		conf::Conf const *sub = cd->sub();
		
		if (sub)
		{
			if (!add_confdirs(dirname, *sub))
				return false;
		}
	}
	
	return true;
}

/*************************************************************************/

conf::ConfQ *ConfData::config(modules::Core &core, conf::ConfGlobalQ &global)
{
	return &m_cd->parser();
}

/*************************************************************************/

bool ConfData::load(modules::Core &core)
{
	if (!core.modules().load(m_cd->database_directives))
	{
		core.log(ErrorConfData(), "Can't load database module %s: %s",
				m_cd->database_directives.value().c_str(),
				core.modules().error().c_str());
		return false;
	}
	
	return Module::load(core);
}

/*************************************************************************/

void ConfData::notify(modules::Module &sender, int msg, void *arg)
{
	switch (msg)
	{
		case mnLoaded:
			if (sender.name() == m_cd->database_directives.value())
			{
				databases::Database *db = (databases::Database *)
						sender.factory(databases::Database::IID);
				if (!db)
				{
					m_core->fatalf("Can't create database confdata_directives: "
							"module %s does not expose the Database interface",
							sender.name().c_str());	
				}
				
				/* Add a reference to the module */
				sender.references_add();
				/* Setup the database */
				m_confdirs.setup_database(db);
			}
			
			if (state() & msStarted)
			{
				conf::ConfGlobalQ global(m_core->conf_logger());
					
				if (global.setup())
				{
				
					conf::ConfQ *parser = sender.config(*m_core, global);
			
					if (parser)
						add_confdirs(sender.name().c_str(), parser->conf());
				}
			}
			
			break;
		
		case mnStopped:
			if (sender.name() == m_cd->database_directives.value())
				sender.references_remove();
			
			if (state() & msStarted)
				remove_confdirs(sender.name().c_str());
			
			break;
		
		case mnReconfig:
			if (m_confdirs.writable() && m_confdirs.clear())
				add_confdirs();
			break;
			
		default:
			break;
	}
	
	Module::notify(sender, msg, arg);
}

/*************************************************************************/

bool ConfData::query(std::string const &query, std::string &feedback)
{
	if (query.empty())
		return false;
	
	feedback = "";
	
	char buf[(ConfDir::name_len + ConfDir::value_len / 16) + 100], *ebuf = buf;
	ConfDirs::const_iterator it, end;
	
	for (it = m_confdirs.begin(), end = m_confdirs.end(); it != end; ++it)
	{
		ConfDir *cd = (ConfDir *)*it;
		
		if (misc::match(query.c_str(), cd->name()))
		{
			if (!feedback.empty())
				feedback += "\n";
			
			ebuf += snprintf(ebuf, sizeof(buf)-(ebuf-buf), "%s = %s",
					cd->name(), cd->value());
			feedback += buf;
			
			ebuf = buf;
		}
	}
	
	if (feedback.empty())
		feedback = "No matching configuration directives.";
	
	return true;
}

/*************************************************************************/

bool ConfData::remove_confdirs(char const *name)
{
	if (!m_confdirs.writable())
		return false;
	
	size_t const len = strlen(name) + 2;
	char prefix[PATH_MAX + 1];
	
	snprintf(prefix, sizeof(prefix), "%s::", name);
	
	ConfDirs::const_iterator it, cur, end;
	
	for (it = m_confdirs.begin(), end = m_confdirs.end(); it != end; )
	{
		cur = it++;
		ConfDir *cd = (ConfDir *) *cur;
		
		if (!strnicmp(cd->name(), prefix, len))
		{
			if (!m_confdirs.remove(*cur))
				return false;
		}
	}
	
	return true;
}

/*************************************************************************/

void ConfData::start(modules::Core &core)
{
	if (!m_confdirs.db() || !m_confdirs.db()->start())
		me->core().fatal("Can't start database confdata_directives");
	
	add_confdirs();
	
	Module::start(core);
}

/*************************************************************************/

void ConfData::unload(modules::Core &core)
{
	if (m_confdirs.db())
		delete m_confdirs.db();
	
	Module::unload(core);
}

/*************************************************************************/

} /* namespace confdata */

/*************************************************************************/
