/* Database import module.
 *
 * Database import module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <string>

#include <cmdline/cmdline.h>
#include <core/version.h>
#include <databases/xml.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <modules/core.h>
#include <modules/manager.h>

#include "decl.h"
#include "module.h"

using namespace std;
using namespace modules;
using namespace dbimport;

/*************************************************************************/

static string strip_ext(std::string const &filename, std::string const &ext)
{
	if (string(filename.end() - 4, filename.end()) == ext)
		return string(filename.begin(), filename.end() - 4);
	else
		return filename;
}

/*************************************************************************/

extern "C" Module *dbimport_get(Core &core)
{
	/* Ensure the core version is the same as the version we were
	 * compiled for so that there are no compatibility problems later. */
	if (core.version() != core::core_version)
	{
		core.log(xlog::LogCritical(), "Core module version mismatch");
		return 0;
	}
	
	return (me = new DBImport(core));
}

/*************************************************************************/

extern "C" void dbimport_release(Core &core, Module *module)
{
	delete module;
	me = 0;
}

/*************************************************************************/

namespace dbimport
{

/*************************************************************************/

LOGLEVEL (InfoDBImport, xlog::LogInfo, "Info::DBImport");

/*************************************************************************/

DBImport *me;

/*************************************************************************/

DBImport::DBImport(Core &core) : Module("dbimport", "Database import utility"),
		m_core(&core), m_tgt(0), m_target(), m_dir(), m_index("index.xml"),
		m_db()
{
	
}

/*************************************************************************/

bool DBImport::load(modules::Core &core)
{
	/* Make the program run in foreground. */
	core.flags_add(Core::pfForeground);
	
	/* Parse command-line parameters. */
	typedef cmdline::CmdLineParser::ParamsList::const_iterator PIType;
	
	PIType it;
	
	it = core.clp().params().find("target");
	if (it == core.clp().params().end())
	{
		core.log(xlog::LogCritical(),
				"Missing --target command-line parameter.");
		return false;
	}
	m_target = it->second;
	
	it = core.clp().params().find("dir");
	if (it != core.clp().params().end())
		m_dir = it->second;

	it = core.clp().params().find("index");
	if (it != core.clp().params().end())
		m_index = it->second;
	
	pair<PIType, PIType> range;
	
	range = core.clp().params().equal_range("db");
	if (range.first != range.second)
	{
		PIType end;
		
		for (it = range.first, end = range.second; it != end; ++it)
			m_db.push_back(it->second);
	}
	
	if (!core.modules().load(m_target))
	{
		core.log(xlog::LogCritical(), "Can't load module %s",
				m_target.c_str());
		return false;
	}
	
	return Module::load(core);
}

/*************************************************************************/

void DBImport::notify(modules::Module &sender, int msg, void *arg)
{	
	Module::notify(sender, msg, arg);
	
	switch (msg)
	{
		case Module::mnLoaded:
			if (sender.name() == m_target)
			{
				m_tgt = &sender;
				m_tgt->references_add();
			}
			break;
		case Module::mnStopped:
			if (&sender == m_tgt)
				m_tgt->references_remove();
			m_tgt = 0;
		default:
			break;
	}
}

/*************************************************************************/

void DBImport::start(modules::Core &core)
{
	databases::XMLImport *xim;
	
	if (!m_tgt || !(m_tgt->state() & Module::msStarted))
		m_core->fatalf("Module %s has not been started", m_target.c_str());
	
	if (!(xim = (databases::XMLImport *) 
			m_tgt->factory(databases::XMLImport::IID)))
		m_core->fatalf("Module %s does not expose the XMLImport interface",
				m_target.c_str());
	
	databases::XMLIndex xi;
	
	if (!xi.load(misc::absolute_path(m_dir, m_index)))
		m_core->fatalf("Can't load index file %s: %s",
				misc::absolute_path(m_dir, m_index).c_str(),
				xi.error().c_str());
	
	m_core->log(InfoDBImport(), "Loaded index file");
	
	if (m_db.empty())
	{
		databases::XMLIndex::const_iterator it, end;
		
		for (it = xi.begin(), end = xi.end(); it != end; ++it)
		{
			if (!xim->import_data(m_dir, strip_ext(*it, ".xml")))
				m_core->fatalf("Can't import databases through module %s: %s",
						m_target.c_str(), xim->error());
		}
	} else 
	{
		dblist_type::const_iterator it, end;
		
		for (it = m_db.begin(), end = m_db.end(); it != end; ++it)
		{
			if (!xi.exists(strip_ext(*it, ".xml") + ".xml"))
				m_core->fatalf("Database %s is not listed in the index file",
						(*it).c_str());
			
			if (!xim->import_data(m_dir, strip_ext(*it, ".xml")))
				m_core->fatalf("Can't export database %s through module %s: %s",
						it->c_str(), m_target.c_str(), xim->error());
		}
	}
	
	delete xim;
	
	m_core->log(InfoDBImport(), "Database import complete.");
	
	Module::start(core);
	exit(EXIT_SUCCESS);
}

/*************************************************************************/

} /* namespace dbimport */

/*************************************************************************/
