/* Database removal module.
 *
 * Database removal module (c) 2006 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <cmdline/cmdline.h>
#include <core/version.h>
#include <databases/database.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <modules/core.h>
#include <modules/manager.h>

#include "decl.h"
#include "module.h"

using namespace std;
using namespace modules;
using namespace dbremove;

/*************************************************************************/

extern "C" Module *dbremove_get(Core &core)
{
	/* Ensure the core version is the same as the version we were
	 * compiled for so that there are no compatibility problems later. */
	if (core.version() != core::core_version)
	{
		core.log(xlog::LogCritical(), "Core module version mismatch");
		return 0;
	}
	
	return (me = new DBRemove(core));
}

/*************************************************************************/

extern "C" void dbremove_release(Core &core, Module *module)
{
	delete module;
	me = 0;
}

/*************************************************************************/

namespace dbremove
{

/*************************************************************************/

LOGLEVEL (InfoDBRemove, xlog::LogInfo, "Info::DBRemove");
LOGLEVELH(CriticalDBRemove, xlog::LogCritical, "Critical::DBRemove",
		"Critical error: dbremove: ");

/*************************************************************************/

DBRemove *me;

/*************************************************************************/

DBRemove::DBRemove(Core &core) : Module("dbremove", "Database removal utility"),
		m_core(&core), m_tgt(0), m_target(), m_db(), m_all(false)
{
	
}

/*************************************************************************/

bool DBRemove::load(modules::Core &core)
{
	/* Make the program run in foreground. */
	core.flags_add(Core::pfForeground);
	
	/* Parse command-line parameters. */
	typedef cmdline::CmdLineParser::ParamsList::const_iterator PIType;
	
	PIType it;
	
	it = core.clp().params().find("target");
	if (it == core.clp().params().end())
	{
		core.log(CriticalDBRemove(),
				"Missing --target command-line parameter.");
		return false;
	}
	m_target = it->second;
	
	it = core.clp().params().find("all");
	if (it != core.clp().params().end())
		m_all = true;
	
	if (!m_all)
	{
		pair<PIType, PIType> range;
		
		range = core.clp().params().equal_range("db");
		if (range.first != range.second)
		{
			PIType end;
			
			for (it = range.first, end = range.second; it != end; ++it)
				m_db.push_back(it->second);
		} else 
		{
			core.log(CriticalDBRemove(),
					"Missing --db command-line parameter.");
			return false;
		}
	}
	
	if (!core.modules().load(m_target))
	{
		core.log(CriticalDBRemove(), "Can't load module %s",
				m_target.c_str());
		return false;
	}
	
	return Module::load(core);
}

/*************************************************************************/

void DBRemove::notify(modules::Module &sender, int msg, void *arg)
{	
	Module::notify(sender, msg, arg);
	
	switch (msg)
	{
		case Module::mnLoaded:
			if (sender.name() == m_target)
			{
				m_tgt = &sender;
				m_tgt->references_add();
			}
			break;
		case Module::mnStopped:
			if (&sender == m_tgt)
				m_tgt->references_remove();
			m_tgt = 0;
		default:
			break;
	}
}

/*************************************************************************/

void DBRemove::start(modules::Core &core)
{
	databases::DBRemove *dbr;
	
	if (!m_tgt || !(m_tgt->state() & Module::msStarted))
		m_core->fatalf("Module %s has not been started", m_target.c_str());
	
	if (!(dbr = (databases::DBRemove *) 
			m_tgt->factory(databases::DBRemove::IID)))
		m_core->fatalf("Module %s does not expose the DBRemove interface",
				m_target.c_str());
	
	if (m_all)
	{
		if (!dbr->clear())
			m_core->fatalf("Can't remove all databases through module %s: %s",
					m_target.c_str(), dbr->error());
	} else 
	{
		dblist_type::const_iterator it, end;
		
		for (it = m_db.begin(), end = m_db.end(); it != end; ++it)
		{
			if (!dbr->remove(*it))
				m_core->fatalf("Can't remove database %s through module %s: %s",
						it->c_str(), m_target.c_str(), dbr->error());
		}
	}
	
	delete dbr;
	
	m_core->log(InfoDBRemove(), "Database removal complete.");
	
	Module::start(core);
	exit(EXIT_SUCCESS);
}

/*************************************************************************/

} /* namespace dbremove */

/*************************************************************************/
