/* Database information.
 *
 * MySQL database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <sstream>

#include <databases/collection.h>
#include <databases/field.h>
#include <databases/record.h>
#include <log/levels.h>
#include <modules/core.h>
#include <mysql-cli/conn.h>
#include <mysql-cli/result.h>

#include "conf.h"
#include "dbinfo.h"
#include "decl.h"
#include "module.h"

using namespace std;

using mysql_cli::Result;

/*************************************************************************/

namespace mysql
{

/*************************************************************************/

DBIndex::~DBIndex()
{
	container_type::iterator it, end;
	
	for (it = m_dbs.begin(), end = m_dbs.end(); it != end; ++it)
		delete (*it);
}

/*************************************************************************/

void DBIndex::add(DBInfo *dbi)
{
	m_dbs.push_back(dbi);
}

/*************************************************************************/

bool DBIndex::create_tables()
{
	char qbuf[2048];
	
	/* Check if the mysql_databases table exists */
	snprintf(qbuf, sizeof(qbuf), "SHOW TABLES LIKE '%smysql_databases'",
			me->cd().table_prefix.value().c_str());
	
	Result res;
	
	if (!m_conn->query(qbuf, &res))
		return false;
	
	/* If no rows were returned, then the table doesn't exist */
	if (!res.num_rows())
	{
		snprintf(qbuf, sizeof(qbuf), "CREATE TABLE `%smysql_databases` ("
				"id SMALLINT(5) UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, "
				"name VARCHAR(255) NOT NULL DEFAULT '' UNIQUE KEY, "
				"version SMALLINT(5) UNSIGNED NOT NULL DEFAULT 1, "
				"mem INT(10) UNSIGNED NOT NULL DEFAULT 0) "
				"DEFAULT CHARACTER SET %s %s%s%s ENGINE=MyISAM",
				me->cd().table_prefix.value().c_str(),
				me->cd().table_charset.value().c_str(),
				(me->cd().table_collation ? "COLLATE" : ""),
				(me->cd().table_collation ? " " : ""),
				(me->cd().table_collation ? 
					me->cd().table_collation.value().c_str() : "")
				);
		
		if (!m_conn->query(qbuf, 0))
			return false;
	}
	
	snprintf(qbuf, sizeof(qbuf), "SHOW COLUMNS FROM `%smysql_databases` "
			"LIKE 'mem'", me->cd().table_prefix.value().c_str());
	
	if (!m_conn->query(qbuf, &res))
		return false;
	
	if (!res.num_rows())
	{
		snprintf(qbuf, sizeof(qbuf), "ALTER TABLE `%smysql_databases` "
				"ADD COLUMN (mem INT(10) UNSIGNED NOT NULL DEFAULT 0)",
				me->cd().table_prefix.value().c_str());
		
		if (!m_conn->query(qbuf, 0))
			return false;
	}
	
	/* Check if the mysql_fields table exists */
	snprintf(qbuf, sizeof(qbuf), "SHOW TABLES LIKE '%smysql_fields'",
			me->cd().table_prefix.value().c_str());
	
	if (!m_conn->query(qbuf, &res))
		return false;
	
	/* If no rows were returned, then the table doesn't exist */
	if (!res.num_rows())
	{
		snprintf(qbuf, sizeof(qbuf), "CREATE TABLE `%smysql_fields` ("
				"id MEDIUMINT(8) UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, "
				"owner SMALLINT(5) UNSIGNED NOT NULL DEFAULT 0, "
				"name VARCHAR(255) NOT NULL DEFAULT '', "
				"type SMALLINT(5) UNSIGNED NOT NULL DEFAULT 0, "
				"size INT(10) UNSIGNED NOT NULL DEFAULT 0, "
				"attrs SMALLINT(5) UNSIGNED NOT NULL DEFAULT 0, "
				"defval LONGTEXT NOT NULL DEFAULT '', "
				"INDEX owner_id (owner, id))"
				"DEFAULT CHARACTER SET %s %s%s%s ENGINE=MyISAM",
				me->cd().table_prefix.value().c_str(),
				me->cd().table_charset.value().c_str(),
				(me->cd().table_collation ? "COLLATE" : ""),
				(me->cd().table_collation ? " " : ""),
				(me->cd().table_collation ? 
					me->cd().table_collation.value().c_str() : "")
				);
		
		if (!m_conn->query(qbuf, 0))
			return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBIndex::clear()
{
	container_type::iterator it, end;
	
	for (it = m_dbs.begin(), end = m_dbs.end(); it != end; ++it)
	{
		if (!(*it)->drop_table())
			return false;
		
		delete (*it);
	}
	
	m_dbs.clear();
	return true;
}

/*************************************************************************/

bool DBIndex::fetch()
{
	/* First ensure that the required tables exist. */
	if (!create_tables())
	{
		me->core().log(ErrorMySQL(),
				"Creation or validation of requisite tables failed");
		return false;
	}
	
	char qbuf[1024];
	
	snprintf(qbuf, sizeof(qbuf),
			"SELECT `id`, `name`, `version`, `mem` FROM `%smysql_databases`",
			me->cd().table_prefix.value().c_str());
	
	Result res;
	
	if (!m_conn->query(qbuf, &res))
		return false;
	
	MYSQL_ROW row;
	
	while ((row = res.fetch_row()))
	{
		DBInfo *dbi = new DBInfo(m_conn, strtoul(row[0], 0, 10), row[1],
				atoi(row[2]), strtoul(row[3], 0, 10));
		
		if (!dbi->fetch())
		{
			me->core().log(ErrorMySQL(),
					"Can't retrieve field information for database %s",
					dbi->name());
			delete dbi;
			return false;
		}
		
		m_dbs.push_back(dbi);
		dbi->inserted(true);
		
		if (me->core().debuglevel())
			me->core().log(DebugMySQL(),
					"Added database %s to database index", dbi->name());
	}
	
	return true;
}

/*************************************************************************/

DBInfo *DBIndex::find(char const *name) const
{
	container_type::const_iterator it, end;
	
	for (it = m_dbs.begin(), end = m_dbs.end(); it != end; ++it)
		if (!strcmp((*it)->name(), name))
			return (*it);
	
	return 0;
}

/*************************************************************************/

bool DBIndex::remove(DBInfo *dbi)
{
	if (!dbi->drop_table())
		return false;
	
	container_type::const_iterator it, end;
	
	m_dbs.remove(dbi);
	return true;
}

/*************************************************************************/

DBInfo::~DBInfo()
{
	container_type::iterator it, end;
	
	for (it = m_fields.begin(), end = m_fields.end(); it != end; ++it)
		delete (*it);
}

/*************************************************************************/

void DBInfo::add(FieldInfo *fi)
{
	m_fields.push_back(fi);
}

/*************************************************************************/

bool DBInfo::alter_table()
{
	ostringstream q, def;
	bool need_alter = false, need_default = false;
	
	q << "ALTER TABLE `" << me->cd().table_prefix.value() << m_name << "` ADD (";
	
	def << "UPDATE `" << me->cd().table_prefix.value() << m_name << "` SET ";
	
	container_type::iterator it, end;
	
	for (it = m_fields.begin(), end = m_fields.end(); it != end; ++it)
	{
		FieldInfo *fi = (*it);
		
		if (fi->inserted())
			continue;
		
		if (need_alter)
			q << ", ";
		q << fi->decl();
		
		need_alter = true;
		
		if (fi->type() == databases::Field::ftString &&
				fi->size() >= 256 && fi->defval())
		{
			if (need_default)
				def << ", ";
			def << "`" << fi->name() << "`" << " = '" <<
					m_conn->escape_string(fi->pdefval()) << "'";
			
			need_default = true;
		}
	}
	
	if (!need_alter)
		return true;
	
	q << ")";
	
	if (!m_conn->query(q.str().c_str(), 0))
		return false;
	
	if (need_default && !m_conn->query(def.str().c_str(), 0))
		return false;
	
	q.str("");
	
	q << "INSERT INTO `" << me->cd().table_prefix.value() <<
				"mysql_fields` (id, owner, name, type, size, attrs, defval) VALUES";
	
	bool need_comma = false;
	size_t count = 0;
	
	for (it = m_fields.begin(), end = m_fields.end(); it != end; ++it)
	{
		FieldInfo *fi = (*it);
		
		if (fi->inserted())
			continue;
		
		if (need_comma)
			q << ", ";
		else
			need_comma = true;
		
		q << "(NULL, " << m_id << ", '" <<
				m_conn->escape_string(fi->name()) << "', " << fi->type() <<
				", " << fi->size() << ", " << fi->attrs() << ", '";
		q << m_conn->escape_string(fi->pdefval()) << "')";
		
		++count;
	}
	
	if (!m_conn->query(q.str().c_str(), 0))
		return false;
	
	uint32_t iid = m_conn->insert_id();
	iid -= (count - 1);
	
	for (it = m_fields.begin(), end = m_fields.end(); it != end; ++it)
	{
		FieldInfo *fi = (*it);
		
		if (fi->inserted())
			continue;
		
		fi->id(iid++);
		fi->inserted(true);
	}
	
	return true;
}

/*************************************************************************/

bool DBInfo::create_table()
{
	ostringstream cre, ins;
	
	cre << "CREATE TABLE `" << me->cd().table_prefix.value() << m_name << "` (" <<
			"id BIGINT(10) UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY, "
			"res_act SET('CREATE', 'UPDATE', 'DELETE') NOT NULL DEFAULT '', ";
	
	container_type::iterator it, end;
	
	for (it = m_fields.begin(), end = m_fields.end(); it != end; ++it)
		cre << (*it)->decl() << ", ";
			
	cre << "INDEX res_act (res_act)) ENGINE = MyISAM "
			"AUTO_INCREMENT = " << databases::Collection::external_id <<
			" DEFAULT CHARACTER SET " << me->cd().table_charset.value();
	
	if (me->cd().table_collation)
		cre << " COLLATE " << me->cd().table_collation.value();

	if (!m_conn->query(cre.str().c_str(), 0))
		return false;

	ins << "INSERT INTO `" << me->cd().table_prefix.value() <<
			"mysql_databases` (id, name, version) VALUES("
			"NULL, '" << m_conn->escape_string(m_name) << "', " <<
			m_version << ")";
	
	if (!m_conn->query(ins.str().c_str(), 0))
		return false;
	m_id = m_conn->insert_id();
	
	ins.str("");
		ins << "INSERT INTO `" << me->cd().table_prefix.value() <<
				"mysql_fields` (id, owner, name, type, size, attrs, defval) VALUES";
	
	bool need_comma = false;
	
	for (it = m_fields.begin(), end = m_fields.end(); it != end; ++it)
	{
		FieldInfo *fi = (*it);
		
		if (need_comma)
			ins << ", ";
		else
			need_comma = true;
		
		ins << "(NULL, " << m_id << ", '" <<
				m_conn->escape_string(fi->name()) << "', " << fi->type() <<
				", " << fi->size() << ", " << fi->attrs() << ", '";
		ins << m_conn->escape_string(fi->pdefval()) << "')";
	}
	
	if (!m_conn->query(ins.str().c_str(), 0))
		return false;
	
	uint32_t iid = m_conn->insert_id();
	iid -= (m_fields.size() - 1);
	
	for (it = m_fields.begin(), end = m_fields.end(); it != end; ++it)
	{
		(*it)->id(iid++);
		(*it)->inserted(true);
	}

	return (m_inserted = true);
}

/*************************************************************************/

bool DBInfo::drop_table()
{
	char qbuf[1024];
	
	snprintf(qbuf, sizeof(qbuf), "DELETE FROM `%smysql_databases` "
			"WHERE `id` = %u", me->cd().table_prefix.value().c_str(), m_id);
	
	if (!m_conn->query(qbuf, 0))
		return false;
	
	snprintf(qbuf, sizeof(qbuf), "DELETE FROM `%smysql_fields` "
			"WHERE `owner` = %u", me->cd().table_prefix.value().c_str(), m_id);
	
	if (!m_conn->query(qbuf, 0))
		return false;
	
	snprintf(qbuf, sizeof(qbuf), "DROP TABLE `%s%s`",
			me->cd().table_prefix.value().c_str(), m_name.str());
	
	if (!m_conn->query(qbuf, 0))
		return false;
	
	return true;
}

/*************************************************************************/

bool DBInfo::fetch()
{
	char qbuf[1024];
	
	snprintf(qbuf, sizeof(qbuf), "SELECT `id`, `name`, `type`, `size`, `attrs`, `defval` "
			"FROM `%smysql_fields` WHERE `owner` = %u ORDER BY `id`",
			me->cd().table_prefix.value().c_str(), m_id);
	
	Result res;
	
	if (!m_conn->query(qbuf, &res))
		return false;
	
	MYSQL_ROW row;
	
	while ((row = res.fetch_row()))
	{
		FieldInfo *fi = new FieldInfo(m_conn, strtoul(row[0], 0, 10), row[1],
				atoi(row[2]), strtoul(row[3], 0, 10), atoi(row[4]), row[5]);
		
		m_fields.push_back(fi);
		fi->inserted(true);
		
		if (me->core().debuglevel() >= 2)
			me->core().log(DebugMySQL(),
					"Added field %s to database %s", fi->name(), name());
	}
	
	return true;
}

/*************************************************************************/

FieldInfo *DBInfo::find(char const *name) const
{
	container_type::const_iterator it, end;
	
	for (it = m_fields.begin(), end = m_fields.end(); it != end; ++it)
		if (!strcmp((*it)->name(), name))
			return (*it);
	
	return 0;
}

/*************************************************************************/

bool DBInfo::mem(size_t mem)
{
	if (m_mem == mem)
		return true;
	
	char qbuf[1024];
	
	snprintf(qbuf, sizeof(qbuf), "UPDATE `%smysql_databases` SET `mem` = %u "
			"WHERE `id` = %u", me->cd().table_prefix.value().c_str(),
			mem, m_id);
	
	if (!m_conn->query(qbuf, 0))
		return false;
	
	m_mem = mem;
	return true;
}

/*************************************************************************/

bool DBInfo::update_table()
{
	if (m_inserted)
	{
		if (!alter_table())
		{
			me->core().log(ErrorMySQL(),
					"Can't alter the MySQL table for database %s",
					name());
			return false;
		}
		
		return true;
	}
	else
	{
		if (!create_table())
		{
			me->core().log(ErrorMySQL(),
					"Can't create the MySQL table for database %s",
					name());
			return false;
		}
		
		return true;
	}
}

/*************************************************************************/

bool DBInfo::version(uint16_t version)
{
	if (m_version == version)
		return true;
	
	char qbuf[1024];
	
	snprintf(qbuf, sizeof(qbuf), "UPDATE `%smysql_databases` SET `version` = %u "
			"WHERE `id` = %u", me->cd().table_prefix.value().c_str(),
			version, m_id);
	
	if (!m_conn->query(qbuf, 0))
		return false;
	
	m_version = version;
	return true;
}

/*************************************************************************/

char const *FieldInfo::decl() const
{
	static char decl[1024];
	char *end = decl;
	
	end += snprintf(end, sizeof(decl) - (end-decl), "`%s` ", name());
	
	switch (m_type)
	{
		case databases::Field::ftString:
			if (m_size <= 255)
				end += snprintf(end, sizeof(decl) - (end-decl), "VARCHAR(%u) ",
						m_size);
			else if (m_size <= 65535)
				end += snprintf(end, sizeof(decl) - (end-decl), "TEXT ");
			else if (m_size <= 16777215)
				end += snprintf(end, sizeof(decl) - (end-decl), "MEDIUMTEXT ");
			else
				end += snprintf(end, sizeof(decl) - (end-decl), "LONGTEXT ");
			
			if (m_attrs & databases::Field::faCase)
				end += snprintf(end, sizeof(decl) - (end-decl), "BINARY ");
			
			end += snprintf(end, sizeof(decl) - (end-decl), "NOT NULL");
			
			/* Default values are not supported by *TEXT fields */
			if (m_size <= 255)
				end += snprintf(end, sizeof(decl)-(end-decl), " DEFAULT '%s'",
						m_conn->escape_string(pdefval()));
			
			return decl;
		case databases::Field::ftInteger:
			if (m_size == 1)
				end += snprintf(end, sizeof(decl) - (end-decl),
						"TINYINT(4) NOT NULL ");
			else if (m_size == 2)
				end += snprintf(end, sizeof(decl) - (end-decl),
						"SMALLINT(6) NOT NULL ");
			else if (m_size == 4)
				end += snprintf(end, sizeof(decl) - (end-decl),
						"INT(11) NOT NULL ");
			else
			{
				me->core().log(ErrorMySQL(), "Bad size %u for field %s",
						m_size, name());
				return 0;
			}
			
			if (m_defval)
				end += snprintf(end, sizeof(decl) - (end-decl), "DEFAULT %ld",
						strtol(m_defval, 0, 10));
			else
			{
				me->core().log(ErrorMySQL(),
						"Missing default value for field %s", name());
				return 0;
			}
			
			return decl;
		case databases::Field::ftUnsigned:
			if (m_size == 1)
				end += snprintf(end, sizeof(decl) - (end-decl),
						"TINYINT(3) UNSIGNED NOT NULL ");
			else if (m_size == 2)
				end += snprintf(end, sizeof(decl) - (end-decl),
						"SMALLINT(5) UNSIGNED NOT NULL ");
			else if (m_size == 4)
				end += snprintf(end, sizeof(decl) - (end-decl),
						"INT(10) UNSIGNED NOT NULL ");
			else
			{
				me->core().log(ErrorMySQL(), "Bad size %u for field %s",
						m_size, name());
				return 0;
			}
			
			if (m_defval)
				end += snprintf(end, sizeof(decl) - (end-decl), "DEFAULT %lu",
						strtoul(m_defval, 0, 10));
			else
			{
				me->core().log(ErrorMySQL(),
						"Missing default value for field %s", name());
				return 0;
			}
		
			return decl;
		case databases::Field::ftUnixTime:
			end += snprintf(end, sizeof(decl) - (end-decl),
					"DATETIME NOT NULL ");
			
			if (m_defval)
			{
				time_t t = strtol(m_defval, 0, 10);
				struct tm tm = *localtime(&t);
				
				char timebuf[64];
				strftime(timebuf, sizeof(timebuf), "%Y-%m-%d %H:%M:%S", &tm);
				
				end += snprintf(end, sizeof(decl) - (end-decl),
						"DEFAULT '%s'", timebuf);
			}
			else
			{
				me->core().log(ErrorMySQL(),
						"Missing default value for field %s", name());
				return 0;
			}
			
			return decl;
		case databases::Field::ftIdentifier:
			end += snprintf(end, sizeof(decl) - (end-decl),
					"BIGINT(10) UNSIGNED NOT NULL ");
			
			if (m_defval)
				end += snprintf(end, sizeof(decl) - (end-decl), "DEFAULT "
						RECORD_FORMAT,
						databases::Record::strtori(m_defval, 0, 10));
			else
			{
				me->core().log(ErrorMySQL(),
						"Missing default value for field %s", name());
				return 0;
			}
			
			if (m_attrs & databases::Field::faOwnerUnique)
				end += snprintf(end, sizeof(decl) - (end-decl),
						" UNIQUE KEY");
			else if (m_attrs & databases::Field::faOwner)
				end += snprintf(end, sizeof(decl) - (end-decl),
						", INDEX `%s` (`%s`)", name(), name());
			
			return decl;
		default:
			me->core().log(ErrorMySQL(),
					"Unknown type %u for field %s", m_type, name());
			return 0;
	};
}

/*************************************************************************/

} /* namespace mysql */

/*************************************************************************/
