/* Database information.
 *
 * MySQL database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef DBINFO_H
#define DBINFO_H

/*************************************************************************/

#include <list>

#include <misc/string.h>

/*************************************************************************/

namespace mysql_cli { class Conn; }

/*************************************************************************/
	
namespace mysql 
{

/*************************************************************************/

class DBInfo;
class FieldInfo;

/*************************************************************************/

class DBIndex
{
	public:
		/// Container type
		typedef std::list<DBInfo *> container_type;
		
		/// Constructor.
		DBIndex(mysql_cli::Conn *conn) : m_dbs(), m_conn(conn) { }
		/// Destructor.
		~DBIndex();
		
		/// Adds a database to the index.
		void add(DBInfo *dbi);
		/// Retrieves the database info.
		bool fetch();
		/// Removes a database from the index.
		bool remove(DBInfo *dbi);
		/// Removes all databases.
		bool clear();
		
		/// Does the given database exist?
		bool exists(char const *name) const { return (find(name) != 0); }
		/// Finds the given database
		DBInfo *find(char const *name) const;
		
		/// Return the databases in the index. */
		container_type const &dbs() const { return m_dbs; }
	private:
		container_type m_dbs;
	
		mysql_cli::Conn *m_conn;
		
		DBIndex(DBIndex const &);
		DBIndex &operator=(DBIndex const &);
		
		bool create_tables();
};

/*************************************************************************/

class DBInfo
{
	public:
		/// Container type
		typedef std::list<FieldInfo *> container_type;
		
		/// Constructor.
		DBInfo(mysql_cli::Conn *conn, uint32_t id, char const *name,
				uint16_t version, size_t mem) : m_fields(), m_conn(conn),
				m_id(id), m_name(name), m_version(version), m_mem(mem),
				m_inserted(false) { }
		/// Destructor.
		~DBInfo();
		
		/// Adds a field to the database.
		void add(FieldInfo *fi);
		/// Finds the given field.
		FieldInfo *find(char const *name) const;
		
		/// Retrieves the fields info.
		bool fetch();
		/// Updates the database table as needed.
		bool update_table();
		/// Drops the database table.
		bool drop_table();
		
		/// Returns the database ID.
		uint32_t id() const { return m_id; }
		/// Returns the database name.
		char const *name() const { return m_name; }
		/// Returns the database version.
		uint16_t version() const { return m_version; }
		/// Sets the database version
		bool version(uint16_t version);
		
		/// Returns the database memory usage.
		size_t mem(size_t mem) const { return m_mem; }
		/// Sets the database memory usage.
		bool mem(size_t mem);
		
		/// Return the fields in the database. */
		container_type const &fields() const { return m_fields; }
		
		/// Has this database already been inserted?
		bool inserted() const { return m_inserted; }
		/// Sets whether this database has already been inserted.
		bool inserted(bool inserted) { return (m_inserted = inserted); }
	private:
		container_type m_fields;
	
		mysql_cli::Conn *m_conn;
		
		uint32_t m_id;
		misc::cstring m_name;
		uint16_t m_version;
		
		size_t m_mem;
		
		bool m_inserted;
		
		DBInfo(DBInfo const &);
		DBInfo &operator=(DBInfo const &);
		
		bool alter_table();
		bool create_table();
};

/*************************************************************************/

class FieldInfo
{
	public:
		/// Constructor.
		FieldInfo(mysql_cli::Conn *conn, uint32_t id, char const *name,
				uint16_t type, uint32_t size, uint16_t attrs,
				char const *defval) : m_conn(conn), m_id(id), m_name(name),
				m_type(type), m_size(size), m_attrs(attrs),
				m_defval(defval && *defval ? defval : 0), m_inserted(false) { }
		
		/// Returns field's id
		uint32_t id() const { return m_id; }
		/// Sets field's id
		uint32_t id(uint32_t id) { return (m_id = id); }
		/// Returns field's name
		char const *name() const { return m_name; }
		/// Returns field's type
		uint16_t type() const { return m_type; }
		/// Returns field's size
		uint32_t size() const { return m_size; }
		/// Returns field's attributes
		uint16_t attrs() const { return m_attrs; }
		/// Returns field's default value
		char const *defval() const { return m_defval; }

		/// Returns field's printable default value
		char const *pdefval() const
				{ return (m_defval ? (char const *)m_defval : ""); }
		
		/// Returns MySQL declaration.
		char const *decl() const;
		
		/// Has this database already been inserted?
		bool inserted() const { return m_inserted; }
		/// Sets whether this database has already been inserted.
		bool inserted(bool inserted) { return (m_inserted = inserted); }
	private:
		mysql_cli::Conn *m_conn;
	
		uint32_t m_id;
		misc::cstring m_name;
		uint16_t m_type;
		uint32_t m_size;
		uint16_t m_attrs;
		misc::cstring m_defval;
		
		bool m_inserted;
		
		FieldInfo(FieldInfo const &);
		FieldInfo &operator=(FieldInfo const &);
};

/*************************************************************************/

} /* namespace mysql */

/*************************************************************************/

#endif /* DBINFO_H */
