/* MySQL module.
 *
 * MySQL database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <core/version.h>
#include <log/levels.h>
#include <modules/core.h>
#include <modules/manager.h>

#include "conf.h"
#include "conn.h"
#include "database.h"
#include "dbinfo.h"
#include "decl.h"
#include "module.h"
#include "xml.h"

using namespace std;
using namespace modules;
using namespace mysql;

/*************************************************************************/

extern "C" Module *mysql_get(Core &core)
{
	/* Ensure the core version is the same as the version we were
	 * compiled for so that there are no compatibility problems later. */
	if (core.version() != core::core_version)
	{
		core.log(xlog::LogCritical(), "Core module version mismatch");
		return 0;
	}
	
	return (me = new MySQL(core));
}

/*************************************************************************/

extern "C" void mysql_release(Core &core, Module *module)
{
	delete module;
	me = 0;
}

/*************************************************************************/

namespace mysql
{

/*************************************************************************/

MySQL *me;

/*************************************************************************/

MySQL::MySQL(Core &core) : 
		Module("mysql", "MySQL database module"),
		m_core(&core), m_cd(new Directives(core.conf_logger())),
		m_conn(0), m_index(0), m_stype(-1), m_recotask(), m_synctask()
{
	
}

/*************************************************************************/

MySQL::~MySQL()
{
	if (m_index)
		delete m_index;
	/* m_conn is free'd in MySQL::notify() */
	
	delete m_cd;
}

/*************************************************************************/

conf::ConfQ *MySQL::config(modules::Core &core, conf::ConfGlobalQ &global)
{
	return &m_cd->parser();
}

/*************************************************************************/

modules::Interface *MySQL::factory(modules::PSIID iid)
{
	if (databases::Database::IID == iid)
		return new Database(*m_core);
	else if (databases::DBRemove::IID == iid)
		return new DBRemove;
	else if (databases::XMLExport::IID == iid)
		return new XMLExport;
	else if (databases::XMLImport::IID == iid)
		return new XMLImport;
	else
		return Module::factory(iid);
}

/*************************************************************************/

bool MySQL::load(modules::Core &core)
{
	m_stype = core.tasks().register_type();
	
	if (!core.modules().load("mysql-cli"))
	{
		core.log(CriticalMySQL(), "Can't load required module mysql-cli: %s",
				core.modules().error().c_str());
		return false;
	}
	
	return Module::load(core);
}

/*************************************************************************/

void MySQL::notify(modules::Module &sender, int msg, void *arg)
{
	switch (msg)
	{
		case mnLoaded:
			if (sender.name() == "mysql-cli")
			{
				sender.references_add();
				
				m_conn = new mysql_cli::Conn(me->cd().connection.hostname,
				me->cd().connection.username, me->cd().connection.password,
				me->cd().connection.database, me->cd().connection.port,
				me->cd().connection.unix_socket);
	
				m_conn->connect_timeout(me->cd().connection.timeout);
				m_conn->reconnect(true);
				m_conn->charset(me->cd().connection.charset);
	
				m_conn->events(new Events);
				m_conn->log_queries(me->cd().connection.log_queries);
				m_conn->log_warnings(me->cd().connection.log_warnings);
			}
			break;
		case mnStopped:
			if (sender.name() == "mysql-cli")
			{
				m_conn->close();
				
				delete m_conn;
				m_conn = 0;
				
				sender.references_remove();
			}
			break;
		case mnSync:
			Database::sync_all();
			break;
		default:
			break;
	}
		
	Module::notify(sender, msg, arg);
}

/*************************************************************************/

bool MySQL::query(std::string const &query, std::string &feedback)
{
	char fb[256];
	
	if (!stricmp(query.c_str(), "INFO"))
	{
		double avg = 0;
		
		if (m_conn->queries())
			avg = static_cast<double>(m_conn->queries()) /
					static_cast<double>(time(0) - m_conn->since());
		
		snprintf(fb, sizeof(fb), "connected: %d; gone: %d; thread id %lu; "
				"client version: %s; host: %s; protocol: %u; "
				"server version: %s; queries: %u; "
				"average queries per second: %.3f - per minute: %.3f",
				m_conn->connected(), m_conn->gone(), m_conn->thread_id(),
				m_conn->client_info(), 
				(m_conn->host_info() ? m_conn->host_info() : "none"),
				m_conn->proto_info(),
				(m_conn->server_info() ? m_conn->server_info() : "unknown"),
				m_conn->queries(), avg, avg * 60);
		goto setfb;
	}
	else if (!strnicmp(query.c_str(), "QUERY", 5))
	{
		if (query.size() < 7)
		{
			snprintf(fb, sizeof(fb), "ERROR No query");
			goto setfb;
		}
		
		if (!m_conn->query(&(query.c_str()[6]), 0))
		{
			snprintf(fb, sizeof(fb),
					"ERROR (code %u; connected: %d; gone: %d)",
					m_conn->errnum(), m_conn->connected(), m_conn->gone());
			goto setfb;
		}
		
		if (m_conn->info())
			snprintf(fb, sizeof(fb), "OK (%s)", m_conn->info());
		else if (m_conn->field_count())
			snprintf(fb, sizeof(fb), "OK (Rows: %lu; Warnings: %u)",
					(unsigned long)m_conn->affected_rows(),
					m_conn->warning_count());
		else
			snprintf(fb, sizeof(fb), "OK (Affected: %lu; Warnings: %u)",
					(unsigned long)m_conn->affected_rows(),
					m_conn->warning_count());
		
		goto setfb;
	}
	else if (!stricmp(query.c_str(), "STATUS"))
	{
		char const *status = m_conn->stat();
			
		feedback = (status ? status : m_conn->error());
		return true;
	}
	
	return false;

setfb:
	feedback = fb;
	return true;
}

/*************************************************************************/

void MySQL::start(modules::Core &core)
{
	if (!m_conn->connect())
		core.fatal("mysql: Connection to the MySQL server failed");
	
	m_index = new DBIndex(m_conn);
	if (!m_index->fetch())
		core.fatal("mysql: Can't load database index");
	
	m_recotask.type(m_stype);
	m_recotask.timeout(me->cd().reconnect_frequency);
	
	if (!core.tasks().add(m_recotask))
		core.fatal("mysql: can't add reconnection attempt task");
	
	m_synctask.type(m_stype);
	m_synctask.timeout(me->cd().sync.frequency);
	
	if (!core.tasks().add(m_synctask))
		core.fatal("mysql: can't add sync task");
	
	Module::start(core);
}

/*************************************************************************/

bool MySQL::stop(modules::Core &core)
{	
	/* Close the connection */
	m_conn->close();
	
	return Module::stop(core);
}

/*************************************************************************/

void MySQL::unload(modules::Core &core)
{
	if (!core.tasks().remove(m_recotask))
		core.fatal("mysql: Can't remove reconnection attempt task");
	
	if (!core.tasks().remove(m_synctask))
		core.fatal("mysql: Can't remove sync task");
	
	Database::free_all();
	
	Module::unload(core);
}

/*************************************************************************/

int MySQL::RecoTask::run()
{
	if (!me->conn().connected())
		me->conn().connect();
	else if (me->conn().gone())
		me->conn().ping();
	
	return rOK;
}

/*************************************************************************/

int MySQL::SyncTask::run()
{
	Database::sync_all();
	return rOK;
}

/*************************************************************************/

} /* namespace mysql */

/*************************************************************************/
