/* Password interface.
 *
 * Password interface (c) 2006 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file password/password.h
  * \brief Password interface header file. */

#ifndef PASSWORD_PASSWORD_H
#define PASSWORD_PASSWORD_H

/*************************************************************************/

#include <modules/interface.h>
#include <modules/module.h>

/*************************************************************************/

/// Namespace of the password interface

namespace password 
{

/*************************************************************************/

/// Password interface.

/** This interface class can be used as a base class by modules providing
  * password encryption.
  *
  * You must override the check(), encrypt() (both versions), length()
  * and module() functions in derived classes.
  *
  * Modules must provide objects of this interface class through the
  * Module::interface() function. */

class Encrypter : public modules::Interface
{
	public:
		/// Interface ID.
		static modules::PSIID const IID = 0x43f29461;
		
		/// Destructor.
		virtual ~Encrypter() { }
		
		/// Returns the name of the module the implementation belongs to.
		/** Returns the name of the module the implementation belongs to.
		  * This is used to easily retrieve the module that generated
		  * an encrypted password.
		  *
		  * \return The module name. */
		virtual modules::Module *module() = 0;
		
		/// Returns the length of encrypted passwords.
		/** Returns the length of passwords encrypted with this
		  * encryption scheme.
		  *
		  * \return The length of encrypted passwords. */
		virtual size_t length() const = 0;
		
		/// Encrypts a password to a buffer.
		/** Encrypts a password to a submitted buffer. The encryption
		  * scheme is up to the implementer.
		  *
		  * \param[in] pass A NULL-terminated password.
		  * \param[out] buf The buffer the encrypted password will be
		  * written to. It must be encoded in a human-readable way and
		  * be NULL terminated.
		  * \param[in] bufsize The size of the output buffer, which
		  * should be at least length() + 1.
		  * \return true if successful, false otherwise. */
		virtual bool encrypt(char const *pass, char *buf,
				size_t bufsize) = 0;
		
		/// Encrypts a password in-place.
		/** Encrypts a password in-place. The encryption scheme is up
		  * to the implementer.
		  *
		  * \param[in] buf On input, this contains the NULL-terminated
		  * password that needs to be encrypted. On output, it contains
		  * the encrypted password, NULL-terminated and human-readable. 
		  * \param[in] size The maximum length of the buffer, which
		  * should be at least length() + 1.
		  * \return true if successful, false otherwise. */
		virtual bool encrypt(char *buf, size_t size) = 0;
		
		/// Checks a password.
		/** Checks a password against a plaintext representation.
		  *
		  * \param[in] plain The plaintext password to check.
		  * \param[in] pass The encrypted password.
		  * \return true if it matches, false otherwise. */
		virtual bool check(char const *plain, char const *pass) = 0;
};

/*************************************************************************/

} /* namespace password */

/*************************************************************************/

#endif /* PASSWORD_PASSWORD_H */
