/* Twilight database module.
 *
 * Twilight database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <core/version.h>
#include <databases/xml.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <modules/core.h>

#include "conf.h"
#include "decl.h"
#include "database.h"
#include "module.h"
#include "xml.h"

using namespace std;
using namespace modules;
using namespace twilight;

/*************************************************************************/

extern "C" Module *twilight_get(Core &core)
{
	/* Ensure the core version is the same as the version we were
	 * compiled for so that there are no compatibility problems later. */
	if (core.version() != core::core_version)
	{
		core.log(xlog::LogCritical(), "Core module version mismatch");
		return 0;
	}
	
	return (me = new Twilight(core));
}

/*************************************************************************/

extern "C" void twilight_release(Core &core, Module *module)
{
	delete module;
	me = 0;
}

/*************************************************************************/

namespace twilight
{

/*************************************************************************/

Twilight *me;

/*************************************************************************/

Twilight::Twilight(Core &core) : 
		Module("twilight", "Twilight database module"), m_core(&core),
		m_cd(new Directives(core.conf_logger())), m_index(), sched_type(-1),
		m_task()
{
	
}

/*************************************************************************/

Twilight::~Twilight()
{
	delete m_cd;
}

/*************************************************************************/

conf::ConfQ *Twilight::config(modules::Core &core, conf::ConfGlobalQ &global)
{
	return &m_cd->parser();
}

/*************************************************************************/

modules::Interface *Twilight::factory(modules::PSIID iid)
{
	if (databases::Database::IID == iid)
		return new Database(*m_core);
	else if (databases::DBRemove::IID == iid)
		return new DBRemove;
	else if (databases::XMLExport::IID == iid)
		return new XMLExport;
	else if (databases::XMLImport::IID == iid)
		return new XMLImport;
	else
		return Module::factory(iid);
}

/*************************************************************************/

bool Twilight::load(modules::Core &core)
{
	sched_type = core.tasks().register_type();
	
	return Module::load(core);
}

/*************************************************************************/

void Twilight::notify(modules::Module &sender, int msg, void *arg)
{
	if (msg == mnSync)
	{
		Database::save_all();
		return;
	}
	
	Module::notify(sender, msg, arg);
}

/*************************************************************************/

bool Twilight::save_index()
{
	return m_index.save(misc::absolute_path(m_cd->directory, "twilight.xml"));
}

/*************************************************************************/

void Twilight::start(modules::Core &core)
{
	std::string path = misc::absolute_path(m_cd->directory, "twilight.xml");
	
	if (misc::file_exists(path.c_str()))
	{
		if (!m_index.load(path))
			core.fatal("twilight: can't load databases index twilight.xml");
	} else
	{
		if (!m_index.save(path))
			core.fatal("twilight: can't save databases index twilight.xml");
	}
	
	m_task.type(sched_type);
	m_task.step(m_cd->save_frequency);
	
	if (!core.tasks().add(m_task))
		core.fatal("twilight: can't add automatic save task");
	
	Module::start(core);
}

/*************************************************************************/

bool Twilight::stop(modules::Core &core)
{
	if (!core.tasks().remove(m_task))
	{
		core.log(ErrorTwilight(), "Can't remove automatic save task");
		return false;
	}
	
	Database::save_all();
	
	return Module::stop(core);
}

/*************************************************************************/

void Twilight::unload(modules::Core &core)
{
	/* Remove all (remaining) databases (if there are still any ;) */
	Database::free_all();
	
	Module::unload(core);
}

/*************************************************************************/

Twilight::Task::Task() : scheduler::TaskStep(-1, 1, 0)
{
	
}

/*************************************************************************/

int Twilight::Task::run()
{
	Database::save_all();
	return rOK;
}

/*************************************************************************/

} /* namespace twilight */

/*************************************************************************/
