/* News functions.
 *
 * Based on code by Andrew Kempe (TheShadow)
 *     E-mail: <theshadow@shadowfire.org>
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#include <core/directives.h>
#include <core/epona.h>
#include <databases/database.h>
#include <log/levels.h>
#include <modules/manager.h>

#include "epona.h"
#include "operserv.h"

#include "services.h"
#include "pseudo.h"

using namespace std;
using namespace core;

/*************************************************************************/

NewsList news;

/*************************************************************************/

LOGLEVELH(ErrorNews, xlog::LogError, "Error::News", "Error: news: ");

/*************************************************************************/

static int news_msgs[ListCmd::cmListCmdTotal];

class NewsCmd : public ListCmd
{
	public:
		NewsCmd() : ListCmd("NEWS", s_OperServ, news_msgs, false),
				m_context(0) { }
	protected:
		virtual char const *context(User *u) const;
		virtual container_type *container(User *u);
		
		virtual bool allow_dispatch(User *u, container_type &ctn,
				char const *sub);
		
		virtual bool changeable(User *u, container_type &ctn,
				char const *sub) const;
		virtual bool is_member(User *u, container_type &ctn,
				char const *sub, databases::Record *rec) const;
		
		virtual int add(User *u, container_type &ctn, char const *sub);
		
		virtual bool del_record(User *u, container_type &ctn,
				char const *sub, databases::Record *rec);
		
		virtual bool list_record(User *u, container_type &ctn,
				char const *sub, uint16_t num, databases::Record const *rec,
				uint16_t count);
	private:
		uint16_t m_context;
};

/*************************************************************************/

bool news_load()
{
	if (!epona->modules().load(epona->cd().operserv.database_news))
	{
		epona->log(ErrorNews(), "Can't load database module %s: %s",
				epona->cd().operserv.database_news.value().c_str(),
				epona->modules().error().c_str());
		return false;
	}
	
	news_msgs[ListCmd::cmSyntax] = OPER_NEWS_SYNTAX;
	news_msgs[ListCmd::cmDisabled] = OPER_NEWS_DISABLED;
	news_msgs[ListCmd::cmEmpty] = OPER_NEWS_EMPTY;
	news_msgs[ListCmd::cmSyntaxDel] = OPER_NEWS_DEL_SYNTAX;
	news_msgs[ListCmd::cmNoSuchEntry] = OPER_NEWS_NO_SUCH_ENTRY;
	news_msgs[ListCmd::cmNoMatch] = OPER_NEWS_NO_MATCH;
	news_msgs[ListCmd::cmDeletedOne] = OPER_NEWS_DELETED_ONE;
	news_msgs[ListCmd::cmDeletedSeveral] = OPER_NEWS_DELETED_SEVERAL;
	news_msgs[ListCmd::cmNotFound] = -1; /* Not used for NEWS */
	news_msgs[ListCmd::cmDeleted] = -1; /* Not used for NEWS */
	news_msgs[ListCmd::cmSyntaxList] = OPER_NEWS_LIST_SYNTAX;
	news_msgs[ListCmd::cmListHeader] = OPER_NEWS_LIST_HEADER;
	news_msgs[ListCmd::cmListedOne] = OPER_NEWS_LISTED_ONE;
	news_msgs[ListCmd::cmListedSeveral] = OPER_NEWS_LISTED_SEVERAL;
	news_msgs[ListCmd::cmClearPartially] = OPER_NEWS_CLEAR_PARTIALLY;
	news_msgs[ListCmd::cmClear] = OPER_NEWS_CLEAR;
	
	return true;
}

/*************************************************************************/

void news_notify(modules::Module &sender, int msg, void *arg)
{
	switch (msg)
	{
		case modules::Module::mnLoaded:
			if (sender.name() == epona->cd().operserv.database_news.value())
			{
				databases::Database *db = (databases::Database *)
						sender.factory(databases::Database::IID);
				if (!db)
				{
					epona->fatalf("Can't create database operserv_news: "
							"module %s does not expose the Database interface",
							sender.name().c_str());	
				}
				
				/* Add a reference to the module */
				sender.references_add();
				/* Setup the database */
				news.setup_database(db);
			}
		default:
			break;
	}
}

/*************************************************************************/

void news_start()
{
	if (!news.db() || !news.db()->start())
		epona->fatal("Can't start database operserv_news");
}

/*************************************************************************/
/***************************** News display ******************************/
/*************************************************************************/

void display_news(User *u, uint16_t type)
{
	int msg;

	if (type == News::ntLogon) {
		msg = OPER_NEWS_LOGON;
	} else if (type == News::ntOper) {
		msg = OPER_NEWS_OPER;
	} else if (type == News::ntRandom) {
		msg = OPER_NEWS_RANDOM;
	} else {
		return;
	}

	if (type == News::ntRandom) 
	{
		static int current_news = -1;
		int count = 0;
		
		while (count++ < news.array().size()) 
		{
			if (++current_news >= news.array().size())
				current_news = 0;
			
			News *n = (News *)news.array()[current_news];
			
			if (!n)
				continue;
			
			if (n->type() == type) 
			{
				struct tm *tm;
				char timebuf[64];
				time_t t = n->since();

				tm = localtime(&t);
				strftime_lang(timebuf, sizeof(timebuf), u,
						STRFTIME_SHORT_DATE_FORMAT, tm);
				notice_lang(s_GlobalNoticer, u, msg, timebuf, n->text());
				break;
			}
		}	
	}
	else 
	{
		News *n;
		int i, count = 0;	/* Number we're going to show--not more than 3 */
		
		for (i = news.array().size() - 1; i >= 0; i--) 
		{
			if (count >= 3)
				break;
			
			n = (News *)news.array()[i];	
			if (!n)
				continue;
			
			if (n->type() == type)
				count++;
		}

		while (++i < news.array().size()) 
		{
			n = (News *)news.array()[i];	
			if (!n)
				continue;
			
			if (n->type() == type)
			{
				struct tm *tm;
				char timebuf[64];
				time_t t = n->since();
		
				tm = localtime(&t);
				strftime_lang(timebuf, sizeof(timebuf), u,
					STRFTIME_SHORT_DATE_FORMAT, tm);
				notice_lang(s_GlobalNoticer, u, msg, timebuf, n->text());
			}
		}
	}
}

/*************************************************************************/
/***************************** News editing ******************************/
/*************************************************************************/

int NewsCmd::add(User *u, container_type &ctn, char const *sub)
{
	if (news.size() == ctn.max_size())
	{
		notice_lang(m_service, u, OPER_NEWS_TOO_MANY, context(u));
		return crStop;
	}
	
	char *text = strtok(NULL, "");
	
	if (!text)
	{
		syntax_error(m_service, u, m_name, OPER_NEWS_ADD_SYNTAX);
		return crStop;
	}
	
	/* Validate the text */
	
	if (strlen(text) > News::text_len)
	{
		notice_lang(m_service, u, OPER_NEWS_TEXT_TOO_LONG,
				News::text_len);
		return crStop;
	}
	
	/* All data is correct, proceed to the creation of the news */
	
	News *n = (News *)news.create();
	
	n->type(m_context);
	n->text(text);
	n->creator(u->nick);
	
	/* Try to add it to the list */
	
	if (!news.add(n, 0))
	{
		notice_lang(m_service, u, m_msgs[cmDisabled]);
		delete n;
	} else {
		notice_lang(m_service, u, OPER_NEWS_ADDED, context(u));
	}

	return crStop;
}

/*************************************************************************/

bool NewsCmd::allow_dispatch(User *u, container_type &ctn,
		char const *sub)
{
	if (!stricmp(sub, "LIST"))
		return is_oper(u);
	else
		return is_services_admin(u);
}

/*************************************************************************/

bool NewsCmd::changeable(User *u, container_type &ctn,
		char const *sub) const
{
	return news.db()->writable();
}

/*************************************************************************/

NewsCmd::container_type *NewsCmd::container(User *u)
{
	m_context = 0;
	
	char const *context = strtok(NULL, " ");
	
	if (!context)
		syntax_error(m_service, u, m_name, m_msgs[cmSyntax]);
	else if (!stricmp(context, "LOGON"))
		m_context = News::ntLogon;
	else if (!stricmp(context, "OPER"))
		m_context = News::ntOper;
	else if (!stricmp(context, "RANDOM"))
		m_context = News::ntRandom;
	else
		syntax_error(m_service, u, m_name, m_msgs[cmSyntax]);
	
	if (!m_context)
		return 0;
	else
		return &news.array();
}

/*************************************************************************/

char const *NewsCmd::context(User *u) const
{
	if (m_context == News::ntLogon)
		return "LOGON";
	else if (m_context == News::ntOper)
		return "OPER";
	else if (m_context == News::ntRandom)
		return "RANDOM";
	else
		return "";
}

/*************************************************************************/

bool NewsCmd::del_record(User *u, container_type &ctn, char const *sub,
		databases::Record *rec)
{
	return news.remove(rec);
}

/*************************************************************************/

bool NewsCmd::is_member(User *u, container_type &ctn,
				char const *sub, databases::Record *rec) const
{
	News *n = (News *)rec;
	return (n->type() == m_context);
}

/*************************************************************************/

bool NewsCmd::list_record(User *u, container_type &ctn, char const *sub,
		uint16_t num, databases::Record const *rec, uint16_t count)
{
	News *n = (News *)rec;
	
	if (count == 0)
		notice_lang(m_service, u, OPER_NEWS_LIST_COLHEAD);
	notice_lang(m_service, u, OPER_NEWS_LIST_FORMAT, num, n->text());
	
	return true;
}

/*************************************************************************/

void do_news(User *u)
{
	static NewsCmd cmd;
	cmd.execute(u);
}

/*************************************************************************/

void do_oldnews(User *u)
{
	notice_lang(s_OperServ, u, OBSOLETE_COMMAND, "NEWS");
}

/*************************************************************************/
