/* NickServ header file.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#ifndef NICKSERV_H
#define NICKSERV_H

/*************************************************************************/

#include <containers/vectorsp.h>
#include <databases/collection.h>
#include <databases/record.h>
#include <log/levels.h>
#include <misc/string.h>

#include "memoserv.h"
#include "services.h"

/*************************************************************************/

namespace modules { class Module; }
namespace password { class Encrypter; }

class NickAlias;
class NickGroup;

/*************************************************************************/

/// Vector for nickname groups.
/** Vector for nickname groups.
  *
  * This just adds two methods (add() and remove()) that adds and removes
  * NickGroup record pointers to the vector. */

class VectorNG : public containers::VectorSP<databases::Record *, uint16_t>
{
	public:
		/// Adds a nickname group to the vector.
		/** Adds a nickname group record to the vector.
		  *
		  * \param[in] na The pointer to the nickname group record. */
		void add(NickGroup *ng);
		
		/// Removes a nickname group from the vector.
		/** Removes a nickname group from the vector. This does not
		  * resize the vector; rather, the pointer in the vector is
		  * just set to NULL in its current emplacement, thus
		  * creating hole in the list.
		  *
		  * \param[in] na The pointer to the nickname group record.
		  * \return true if the record was found in the vector, false
		  * otherwise. */
		bool remove(NickGroup *ng);
};

/*************************************************************************/

/// Sorted vector for nickname aliases.
/** Sorted vector for nickname aliases.
  *
  * This just adds two methods (add() and remove()) that adds and removes
  * NickGroup record pointers to the vector. */

class VectorNA : public containers::VectorSP<databases::Record *, uint16_t>
{
	public:
		/// Adds a nickname alias to the vector.
		/** Adds a nickname alias record to the vector, in sorted order.
		  *
		  * \param[in] na The pointer to the nickname alias record. */
		void add(NickAlias *na);
		
		/// Removes a nickname alias from the vector.
		/** Removes a nickname alias from the vector.
		  *
		  * \param[in] na The pointer to the nickname alias record.
		  * \return true if the record was found in the vector, false
		  * otherwise. */
		bool remove(NickAlias *na);
};

/*************************************************************************/

class NickGroup : public databases::NameRecord
{		
	public:
		/* Classes */
		
		class Access : public databases::NameRecord
		{
			public:
				/* Data */
				static databases::Field *f_owner;
				static databases::Field *f_mask;
				
				/* Consts */
				
				/// Maximum length of the mask.
				static size_t const mask_len = 127;
				
				/* Functions */
				
				/// Constructor.
				Access();
				/// Constructor setting the record ID.
				Access(identifier_type id);
				
				/// Overrides Record::get_s().
				virtual char const *get_s(databases::Field const *field) const;
				/// Overrides Record::set_s().
				virtual char const *set_s(databases::Field const *field,
						char const *value);
				/// Overrides Record::get_ri().
				virtual identifier_type get_ri(databases::Field const *field) const;
				/// Overrides Record::set_ri().
				virtual identifier_type set_ri(databases::Field const *field,
						identifier_type value);
				
				/// Implements NameRecord::name().
				virtual char const *name() const { return mask(); }
				
				/// Overrides Record::memory_usage().
				virtual size_t memory_usage() const;
				
				/// Returns the owner group of the access.
				NickGroup *owner() const { return m_owner; }
				/// Sets the owner group of the access.
				NickGroup *owner(NickGroup *owner);
				/// Sets the owner group ID of the access (for databases only!)
				identifier_type owner_id(identifier_type owner_id);
				
				/// Returns the access mask.
				char const *mask() const { return m_mask; }
				/// Sets the access mask.
				char const *mask(char const *mask);
				
			private:
				NickGroup *m_owner;
				
				misc::cstring m_mask;
		};
		
		/* Typedefs */
		
		typedef containers::VectorSP<databases::Record *, uint16_t> accesses_type;
		
		/* Data */
		static databases::Field *f_name;
		static databases::Field *f_enc;
		static databases::Field *f_pass;
		static databases::Field *f_created;
		static databases::Field *f_flags;
		static databases::Field *f_email;
		static databases::Field *f_auth_code;
		static databases::Field *f_url;
		static databases::Field *f_im_type;
		static databases::Field *f_im_id;
		static databases::Field *f_greet;
		static databases::Field *f_language;
		static databases::Field *f_chost;
		static databases::Field *f_chost_changed;
		static databases::Field *f_chost_setby;
		static databases::Field *f_maxmemos;
		
		MemoInfo memos;
		
		/* Constants */
		
		/// Maximum length of a password.
		static size_t const pass_len = 192;
		/// Real maximum length of a password.
		static size_t const pass_rlen = 31;
		/// Maximum length of the e-mail address.
		static size_t const email_len = 127;
		/// Maximum length of the URL.
		static size_t const url_len = 255;
		/// Maximum length of the IM type.
		static size_t const im_type_len = 31;
		/// Maximum length of the IM identifier.
		static size_t const im_id_len = email_len;
		/// Maximum length of the greet.
		static size_t const greet_len = 300;
		/// Maximum length of the custom hostname.
		static size_t const chost_len = 63;
		/// Maximum length of the custom hostname person
		static size_t const chost_setby_len = 63;
		
		/* Flags */
		
		/// Protect from others taking this nick
		static uint32_t const gfKill          = 0x00000001;
		/// Don't recognize unless IDENTIFY'd
		static uint32_t const gfSecure        = 0x00000002;
		/// Don't allow user to change memo limit
		static uint32_t const gfMemoHardMax   = 0x00000008;
		/// Notify of new memos at signon and un-away
		static uint32_t const gfMemoSignon    = 0x00000010;
		/// Notify of new memos when sent
		static uint32_t const gfMemoReceive   = 0x00000020;
		/// Don't show in LIST to non-services admins
		static uint32_t const gfPrivate       = 0x00000040;
		/// Don't show e-mail in INFO
		static uint32_t const gfHideEmail     = 0x00000080;
		/// Don't show last seen address in INFO
		static uint32_t const gfHideMask      = 0x00000100;
		/// Don't show last quit message in INFO
		static uint32_t const gfHideQuit      = 0x00000200;
		/// Kill in 20 seconds instead of 60
		static uint32_t const gfKillQuick     = 0x00000400;
		/// Kill immediately instead of in 60 secs
		static uint32_t const gfKillImmed     = 0x00000800;
		/// User is a services operator
		static uint32_t const gfServicesOper  = 0x00001000;
		/// User is a services admin
		static uint32_t const gfServicesAdmin = 0x00002000;
		/// User is a services root
		static uint32_t const gfServicesRoot  = 0x00008000;
		/// Enable custom hostname
		static uint32_t const gfCustomHost    = 0x00010000;
		/// E-mail address is authentified
		static uint32_t const gfAuth          = 0x00020000;
		/// Forward new memos to e-mail address
		static uint32_t const gfMemoForward   = 0x00040000;
		
		/// Set the flags to the conf default flags
		static uint32_t const gfDefaults      = 0x80000000;
		
		/// Unused flags
		static uint32_t const gfUnused        = 0x7FF84004;
		/// Default flags for config file
		static uint32_t const gfDefaultConf   = (gfSecure | gfMemoSignon |
				gfMemoReceive | gfHideEmail | gfHideMask);
		
		/* Functions */
		
		/// Constructor.
		NickGroup();
		/// Constructor setting record identifier.
		NickGroup(Record::identifier_type id);
		/// Destructor.
		~NickGroup();
		
		/// Overrides NameRecord::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides NameRecord::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides NameRecord::get_i().
		virtual int32_t get_i(databases::Field const *field) const;
		/// Overrides NameRecord::set_i().
		virtual int32_t set_i(databases::Field const *field, int32_t value);
		/// Overrides NameRecord::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides NameRecord::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides NameRecord::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides NameRecord::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		/// Overrides NameRecord::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides NameRecord::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Overrides NameRecord::end_update().
		virtual bool end_update();
		
		/// Changes the alias that gives the group its name.
		NickAlias *change_name(NickAlias *name);
		
		/// Changes the password.
		bool change_password(char const *pass);
		/// Checks the password.
		bool check_password(char const *pass);
		/// Generates a new password.
		bool generate_password(char *passbuf, size_t size);
		
		/// Begins an authentication process
		/** Begins an authentication process. It generates an
		  * auth code but does not actually send it to the user.
		  * This function REQUIRES that the gfAuth flag is not
		  * set (no checks are done).
		  *
		  * \return true if successful, false otherwise. */ 
		bool begin_auth();
		/// Mails the authentication code
		/** Mails the authentication code to the e-mail address
		  * associated to the nick (which is thus assumed to be
		  * non-NULL). 
		  *
		  * \param[in] nick THe nick of the group this mail should
		  * be targeted to.
		  * \param[in] origin The origin of the mail.
		  * \param[in] force true if the mail must be sent regardless
		  * of quota, false otherwise.
		  * \return true if successful, false otherwise. */
		bool mail_auth(char const *nick, char const *origin,
				bool force) const;
		/// Combines begin_auth() and mail_auth().
		/** Combines begin_auth() and mail_auth(). This should be called
		  * from commands who would generate a new authentication message
		  * when the authentication process is required.
		  *
		  * The function checks whether there is an e-mail address set,
		  * that the gfAuth flag is not set and that there is no
		  * authentication process in progress. If any of those is not
		  * true, no process is started, but the function still returns
		  * true. The caller is responsible to check whether the
		  * authentication system is enabled and required.
		  *
		  * \param[in] nick THe nick of the group this mail should
		  * be targeted to.
		  * \param[in] origin The origin of the mail.
		  * \param[in] force true if the mail must be sent regardless
		  * of quota, false otherwise.
		  * \return true if successful, false otherwise. */
		bool require_auth(User *u, bool force);
		/// Ends an authentication process
		/** Ends an authentication process.
		  *
		  * \param[in] success true if the authentication is successful
		  * (the gfAuth flag will be set), false otherwise.
		  * \return true if successful (the gfAuth flag is guaranteed to
		  * be set if success was true), false otherwise. */
		bool end_auth(bool success); 
		/// Resets authentication status
		bool reset_auth();
		
		/// Changes instant messager info
		bool change_im(char const *im_type, char const *im_id);
		
		/// Changes the custom hostname.
		bool change_chost(char const *chost, time_t changed,
				char const *setby);
		
		/// Is the given user on the group access list?
		bool on_access(User *u) const;
		
		/// Does stuff after successful alias registration
		void post_registered(NickAlias *na);
		/// Does stuff after successful identification
		void post_identified(NickAlias *na);
		/// Does stuff after de-identification
		void post_deidentified(NickAlias *na);
		/// Does stuff after alias unregistration
		void post_unregistered(NickAlias *na);
		
		/// Overrides NameRecord::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Overrides NameRecord::name().
		char const *name() const;
		
		/// Returns the alias that gives the group its name
		NickAlias *name_a() const { return m_name; }
		/// Sets the alias that gives the group its name (doesn't update database)
		NickAlias *name_a(NickAlias *name_a);
		
		/// Returns the ID of the alias that should give the group its name
		identifier_type name_id() const { return m_name_id; }
		/// Sets the ID of the alias that should give the group its name
		/** Sets the ID of the alias that should give the group its name.
		  * This must be set through database modules only, and is
		  * automatically set to zero when the name has been been
		  * resolved (after aliases have been imported).
		  *
		  * \param[in] name_id The ID of the name alias.
		  * \return The new ID of the name alias. */
		identifier_type name_id(identifier_type name_id);
					
		/// Returns the password encoding scheme
		char const *enc() const;
		/// Sets the password encoding scheme
		char const *enc(char const *enc);
		
		/// Returns the password, encrypted with the enc() encoding
		char const *pass() const { return m_pass; }
		/// Sets the password, encrypted with the enc() encoding
		char const *pass(char const *pass);
		
		/// Returns the creation time of the group.
		time_t created() const { return m_created; }
		/// Sets the creation time of the group.
		time_t created(time_t created);
		
		/// Returns the group flags (see the gf* consts).
		uint32_t flags() const { return m_flags; }
		/// Sets the group flags (database only).
		uint32_t flags(uint32_t flags);
		/// Adds flags to the current flags.
		uint32_t flags_add(uint32_t flags);
		/// Removes flags from the current flags.
		uint32_t flags_remove(uint32_t flags);
		
		/// Returns the e-mail address associated to the nick
		char const *email() const { return m_email; }
		/// Sets the e-mail address associated to the nick
		char const *email(char const *email);
		
		/// Returns the auth code
		uint32_t auth_code() const { return m_auth_code; }
		/// Sets the auth code
		uint32_t auth_code(uint32_t auth_code);
		
		/// Returns the URL associated to the nick
		char const *url() const { return m_url; }
		/// Sets the URL associated to the nick
		char const *url(char const *url);
		
		/// Returns the instant messaging type
		char const *im_type() const { return m_im_type; }
		/// Sets the instant messaging type
		char const *im_type(char const *im_type);
		
		/// Returns the instant messaging ID associated to the nick
		char const *im_id() const { return m_im_id; }
		/// Sets the instant messaging ID associated to the nick
		char const *im_id(char const *im_id);
		
		/// Returns the greet associated to the nick
		char const *greet() const { return m_greet; }
		/// Sets the greet associated to the nick
		char const *greet(char const *greet);
		
		/// Returns the language selected by nickname owner (LANG_*)
		uint16_t language() const { return m_language; }
		/// Sets the language (LANG_*)
		uint16_t language(uint16_t language);
		
		/// Returns the custom hostname associated to the nick
		char const *chost() const { return m_chost; }
		/// Sets the custom hostname associated to the nick (by databases only)
		char const *chost(char const *chost);
		
		/// Returns the last time the custom hostname was changed
		time_t chost_changed() const { return m_chost_changed; }
		/// Sets the last time the custom hostname was changed (by databases only)
		time_t chost_changed(time_t chost_changed);
		
		/// Returns the nickname of the person who set the custom hostname
		char const *chost_setby() const { return m_chost_setby; }
		/// Sets the nickname of the person who set the custom hostname (by databases only)
		char const *chost_setby(char const *chost_setby);
		
		/// Returns the maximum number of memos allowed.
		int16_t maxmemos() const { return m_maxmemos; }
		/// Sets the maximum number of memos allowed.
		int16_t maxmemos(int16_t maxmemos);
		
		/// Returns the list of aliases belonging to this group (const).
		VectorNA const &aliases() const { return m_aliases; }
		/// Returns the list of aliases belonging to this group.
		VectorNA &aliases() { return m_aliases; }
		
		/// Returns the access list.
		accesses_type &accesses() { return m_accesses; }
		
		/// Returns the number of bad auths
		uint8_t bad_auths() const { return m_bad_auths; }
		/// Sets the number of bad auths
		uint8_t bad_auths(uint8_t bad_auths)
				{ return (m_bad_auths = bad_auths); }
		/// Increases the number of bad auths by one
		uint8_t bad_auths_inc() { return ++m_bad_auths; }
		
		/// Returns the number of registered channels.
		uint16_t channelcount() const { return m_channelcount; }
		/// Sets the number of registered channels.
		uint16_t channelcount(uint16_t channelcount) 
				{ return (m_channelcount = channelcount); }
		
		/// Group-specific status flags.
		enum
		{
			/// Group is being removed.
			dbsRemoving = 0x8000
		};
	private:
		NickAlias *m_name;
		identifier_type m_name_id;
		
		password::Encrypter *m_enc;
		misc::cstring m_pass;
		
		time_t m_created;
		
		uint32_t m_flags;
		
		misc::cstring m_email;
		uint32_t m_auth_code;
		
		misc::cstring m_url;
		misc::cstring m_im_type;
		misc::cstring m_im_id;
		misc::cstring m_greet;
		
		uint16_t m_language;
		
		misc::cstring m_chost;
		time_t m_chost_changed;
		misc::cstring m_chost_setby;
		
		int16_t m_maxmemos;
		
		/* Unsaved data */
		
		/// List of aliases
		VectorNA m_aliases;
		
		/// List of access entries
		accesses_type m_accesses;
		
		/// Number of bad authentications
		uint8_t m_bad_auths;
		
		/// Number of channels currently registered
		uint16_t m_channelcount;
		
		/// Apply the custom hostname for all aliases of a group.
		void chost_apply();
		/// Unapply the custom hostname for all aliases of a group.
		void chost_unapply();
};

/*************************************************************************/

class Groups : public databases::Collection
{
	public:
		/* Classes */
		
		class Accesses : public databases::Collection
		{
			public:
				/* Consts */
				
				/// Custom error codes
				enum
				{
					/// Access entry with the same name already exists.
					ceName = ceCustom + 0
				};
				
				/* Functions */
				
				/// Implements Collection::create().
				virtual databases::Record *
						create(databases::Record::identifier_type id = 0) const;
				/// Overrides Collection::add().
				virtual bool add(databases::Record *record, int *err = 0);
				
				/// Overrides Collection::clear().
				virtual bool clear(int *err = 0) { return false; }
				
				/// Configures owner collection.
				void setup_owner();
				
				/// Sets the database up.
				/** Sets the database up. This is called by the Module::notify()
				  * function when the database module is loaded. */
				void setup_database(databases::Database *db);
				
				/// Overrides Collection::owner_field().
				virtual databases::Field const
						*owner_field(Collection const *coll) const
						{ return NickGroup::Access::f_owner; }
			protected:
				/// Overrides Collection::add_list().
				virtual bool add_list(databases::Record *record);
				/// Overrides Collection::remove_list().
				virtual void remove_list(databases::Record *record);
				
				/// Overrides Collection::remove_owned_db().
				virtual bool remove_owned_db(databases::Collection const *coll,
						databases::Record const *record, int *err = 0);
				/// Overrides Collection::remove_owned_list().
				virtual bool remove_owned_list(databases::Collection const *coll,
						databases::Record const *record, int *err = 0);
		};
		
		/* Functions */
		
		/// Constructor.
		Groups() : databases::Collection(), m_encrypter(0), m_is_new(false) { }
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides Collection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		/// Overrides Collection::remove().
		virtual bool remove(databases::Record *record, int *err = 0);
		/// Overrides Collection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
	
		/// Loads a password encrypter.
		password::Encrypter *load_encrypter(char const *name);
		/// Decrements the use of the password encrypter by one.
		void release_encrypter(password::Encrypter *enc);
		
		/// Returns the access collection.
		Accesses &accesses() { return m_accesses; }
		
		/// Returns the default password encrypter.
		password::Encrypter *encrypter() { return m_encrypter; }
		/// Sets the default password encrypter.
		password::Encrypter *encrypter(password::Encrypter *enc)
				{ return (m_encrypter = enc); }
		
		/// Are there new groups?
		bool is_new() const { return m_is_new; }
		/// Resets the is_new() result.
		void reset_is_new() { m_is_new = false; }
	private:
		Accesses m_accesses;
		
		password::Encrypter *m_encrypter;
		
		bool m_is_new;
		
		Groups(Groups const &);
		Groups &operator=(Groups const &); 
};

/*************************************************************************/

extern Groups groups;

/*************************************************************************/

class NickAlias : public databases::NameRecord
{			
	public:
		/* Data */
		
		static databases::Field *f_nick;
		static databases::Field *f_group;
		static databases::Field *f_created;
		static databases::Field *f_last_seen;
		static databases::Field *f_last_usermask;
		static databases::Field *f_last_realname;
		static databases::Field *f_last_quit;
		static databases::Field *f_status;
			
		/* Constants */
		
		/// Maximum nickname length.
		static size_t const nick_len = 63;
		/// Maximum usermask length
		static size_t const last_usermask_len = 127;
		/// Maximum realname length
		static size_t const last_realname_len = 127;
		/// Maximum QUIT message length
		static size_t const last_quit_len = 400;
		
		/* Functions */
		
		/// Constructor.
		NickAlias();
		/// Constructor setting the record ID.
		NickAlias(uint32_t id);
		
		/// Overrides NameRecord::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides NameRecord::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides NameRecord::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides NameRecord::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides NameRecord::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides NameRecord::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		/// Overrides NameRecord::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides NameRecord::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Overrides NameRecord::end_update().
		virtual bool end_update();
		
		/// Does stuff after successful registration
		void post_registered(bool validate);
		/// Does stuff after successful recognition
		void post_recognized();
		/// Does stuff after successful identification
		void post_identified();
		/// Does stuff after de-identification
		void post_deidentified();
		/// Does stuff after signoff
		void post_signoff(char const *quit);
		/// Does stuff after unregistration
		void post_unregistered();
		
		/// Overrides NameRecord::name().
		virtual char const *name() const { return m_nick; }
		
		/// Overrides NameRecord::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the nickname.
		char const *nick() const { return m_nick; }
		/// Sets the nickname.
		char const *nick(char const *nick);
		
		/// Returns the group this nick belongs to.
		NickGroup *ng() const { return m_ng; }
		/// Sets the group this nick belongs to.
		NickGroup *ng(NickGroup *ng);
		/// Sets the group this nick belongs to (database only).
		identifier_type ng_id(identifier_type ng_id);
		
		/// Returns the registration time.
		time_t created() const { return m_created; }
		/// Sets the registration time.
		time_t created(time_t created);
		
		/// Returns the last time the nick was seen online.
		time_t last_seen() const { return m_last_seen; }
		/// Sets the last time the nick was seen online.
		time_t last_seen(time_t last_seen);
		
		/// Returns the last usermask.
		char const *last_usermask() const { return m_last_usermask; }
		/// Sets the last usermask.
		char const *last_usermask(char const *last_usermask);
		/// Sets the last usermask through a username and hostname.
		char const *last_usermask(char const *user, char const *host);
		
		/// Returns the last realname.
		char const *last_realname() const { return m_last_realname; }
		/// Sets the last realname.
		char const *last_realname(char const *last_realname);
		
		/// Returns the last quit message.
		char const *last_quit() const { return m_last_quit; }
		/// Sets the last quit message.
		char const *last_quit(char const *last_quit);
		
		/* Status flags */
		
		/// Pending expiration
		static uint16_t const asPendingExpire = 0x0001;
		/// Nick never expires
		static uint16_t const asNoExpire      = 0x0004;

		/// User has IDENTIFY'd
		static uint16_t const asIdentified    = 0x8000;
		/// asOnAccess is set and the Group::gfSecure flag is not set
		static uint16_t const asRecognized    = 0x4000;
		/// User comes from a known address
		static uint16_t const asOnAccess      = 0x2000;
		/// Nick is being held after a kill or nick change
		static uint16_t const asKillHeld      = 0x1000;
		/// Nick change has been issued, still waiting for confirmation
		static uint16_t const asGuested       = 0x0800;

		/// Status flags that can be copied to a nick of the same group during nick change
		static uint16_t const asTransGroup    = 0xC000;
		/// Temporary status flags
		static uint16_t const asTemporary     = 0xF800;
		/// Unused status flags
		static uint16_t const asUnused        = 0x07FA;
		
		/// Returns the nick status (see as* status flags).
		uint16_t status() const { return m_status; }
		/// Sets the nick status -- databases only.
		uint16_t status(uint16_t status);
		/// Adds flags to the nick status.
		uint16_t status_add(uint16_t status);
		/// Removes flags from the nick status.
		uint16_t status_remove(uint16_t status);
		
		/* Not saved */
		User *u;						/* Current online user that has me */
	private:
		misc::cstring m_nick;
		
		NickGroup *m_ng;
		
		time_t m_created;
		time_t m_last_seen;
		
		misc::cstring m_last_usermask;
		misc::cstring m_last_realname;
		misc::cstring m_last_quit;
		
		uint16_t m_status;
		
		void set_reg_mode();
		void unset_reg_mode();
};

/*************************************************************************/

class Aliases : public databases::NameCollection
{
	public:
		/// Constructor.
		Aliases();
		
		/// Overrides NameCollection::end_initial().
		virtual bool end_initial();
		/// Overrides NameCollection::end_import().
		virtual bool end_import(uint16_t type);
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides NameCollection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		/// Overrides NameCollection::remove().
		virtual bool remove(databases::Record *record, int *err = 0);
		/// Overrides NameCollection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Removes expired nicknames.
		void expire();
		
		/// Finds a record by ID.
		virtual NickAlias *ifind(databases::Record::identifier_type id)
				{ return (NickAlias *)NameCollection::find(id); }
		/// Finds a record by name.
		virtual NickAlias *find(char const *name) const
				{ return (NickAlias *)NameCollection::nfind(name); }
		
		/// Configures owner collection.
		void setup_owner();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Overrides NameCollection::owner_field().
		virtual databases::Field const
				*owner_field(Collection const *coll) const;
	protected:
		/// Overrides NameCollection::remove_owned_db().
		virtual bool remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		
		/// Overrides NameCollection::remove_owned_list().
		virtual bool remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
	private:
		Aliases(Aliases const &);
		Aliases &operator=(Aliases const &);
		
		void cleanup_groups();
};

/*************************************************************************/

extern Aliases aliases;

/*************************************************************************/

LOGLEVELH(DebugNickServ, xlog::LogDebug, "Debug::NickServ", "debug: nickserv: ");

LOGLEVELH(InfoNickServ, xlog::LogInfo, "Info::NickServ", "nickserv: ");
LOGLEVELH(InfoNickServList, InfoNickServ, "Info::NickServ::List", "nickserv list: ");
LOGLEVELH(InfoNickServRights, InfoNickServ, "Info::NickServ::Rights", "nickserv rights: ");
LOGLEVELH(InfoNickServSettings, InfoNickServ, "Info::NickServ::Settings", "nickserv settings: ");

LOGLEVELH(ErrorNickServ, xlog::LogError, "Error::NickServ", "Error: nickserv: ");
LOGLEVELH(ErrorNickServList, xlog::LogError, "Error::NickServ::List", "Error: nickserv list: ");

/*************************************************************************/

extern void listnicks(int count_only, const char *nick);

extern void ns_init();

extern bool nickserv_load();
extern void nickserv_notify(modules::Module &sender, int msg, void *arg);
extern void nickserv_start();

extern void nickserv(User *u, char *buf);
extern void load_ns_dbase();
extern void save_ns_dbase();
extern int validate_user(User *u);
extern void cancel_user(User *u);
extern int nick_identified(User *u);
extern int nick_recognized(User *u);
#ifdef HAS_CHGHOST
extern void custom_host_set(User *u);
extern void custom_host_unset(User *u);
#endif

extern void clean_ns_timeouts(NickAlias *na);

/*************************************************************************/

#endif /* NICKSERV_H */
